/* Duotone 1.00 - image processing plug-in for the Gimp 1.0 API
 *
 * Copyright (C) 2002 Alastair Robinson (blackfive@fakenhamweb.co.uk)
 * Based on code by Andrew Kieschnick
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>


/* Declare local functions.
 */
static void      query  (void);
static void      run    (gchar     *name,
			 gint       nparams,
			 GimpParam    *param,
			 gint      *nreturn_vals,
			 GimpParam   **return_vals);

static void      duotone(GimpDrawable  *drawable);

static gint      duotone_dialog (void);

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

static gint DuotoneChannel = 0;

MAIN ()

static void
query (void)
{
  static GimpParamDef args[] =
  {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE, "image", "Input image (unused)" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
    { GIMP_PDB_INT32, "evenodd", "0 = keep odd, 1 = keep even" }
  };
  static gint nargs = sizeof (args) / sizeof (args[0]);

  printf("Attempting to install Duotone\n");

  gimp_install_procedure ("plug_in_duotone",
			  "Duotone",
			  "Duotone limits the colours of an RGB image such that"
				"only those shades achievable through the use of red"
				"and black ink are used.",
			  "Alastair Robinson",
			  "Alastair Robinson",
			  "2002",
			  "<Image>/Image/Colors/Duotone",
			  "RGB*",
			  GIMP_PLUGIN,
			  nargs, 0,
			  args, NULL);
}

static void
run (gchar   *name,
     gint     nparams,
     GimpParam  *param,
     gint    *nreturn_vals,
     GimpParam **return_vals)
{
  static GimpParam values[1];
  GimpDrawable *drawable;
  GimpRunModeType run_mode;
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;

  printf("Attempting to run Duotone\n");

  run_mode = param[0].data.d_int32;

  /*  Get the specified drawable  */
  drawable = gimp_drawable_get (param[2].data.d_drawable);
  printf("Drawable: %ld, %d\n",param[2].data.d_drawable,(long)drawable);

  printf("Run mode: %d\n",run_mode);

  switch (run_mode)
    {
    case GIMP_RUN_INTERACTIVE:
      printf("Running interactively\n");
      gimp_get_data ("plug_in_duotone", &DuotoneChannel);
      if (! duotone_dialog ())
	status = GIMP_PDB_EXECUTION_ERROR;
      break;

    case GIMP_RUN_NONINTERACTIVE:
      printf("Running non-interactively\n");
      if (nparams != 4)
	  status = GIMP_PDB_CALLING_ERROR;
      if (status == GIMP_PDB_SUCCESS)
        ;
      DuotoneChannel = param[3].data.d_int32;
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      printf("Running with last values\n");
      gimp_get_data ("plug_in_duotone", &DuotoneChannel);
      break;

    default:
      printf("Run mode unknown\n");
      break;
    }

  if (status == GIMP_PDB_SUCCESS)
    {
      /*  Make sure that the drawable is gray or RGB color  */
      printf("Drawable ID: %d\n",drawable->id);
      printf("Is RGB? %d\n",gimp_drawable_is_rgb(drawable->id));

      if (gimp_drawable_is_rgb (drawable->id))
	{
        printf("RGB Drawable found\n");
	  gimp_progress_init ("Duotone...");
	  gimp_tile_cache_ntiles (2 * (drawable->width /
				       gimp_tile_width () + 1));
	  duotone(drawable);

	  if (run_mode != GIMP_RUN_NONINTERACTIVE)
	    gimp_displays_flush ();
	  if (run_mode == GIMP_RUN_INTERACTIVE)
	    gimp_set_data ("plug_in_duotone",
			   &DuotoneChannel, sizeof (DuotoneChannel));
	}
      else
	{
        printf("Error:\n");
	  gimp_message ("duotone: cannot operate on indexed color images");
	  status = GIMP_PDB_EXECUTION_ERROR;
	}
    }
  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;

  gimp_drawable_detach (drawable);
}

static void
duotone(GimpDrawable *drawable)
{
  GimpPixelRgn srcPR, destPR;
  gint width, height;
  gint bytes;
  guchar *src;
  gint row, col;
  gint x1, y1, x2, y2;

  /* Get the input area. This is the bounding box of the selection in
   *  the image (or the entire image if there is no selection). Only
   *  operating on the input area is simply an optimization. It doesn't
   *  need to be done for correct operation. (It simply makes it go
   *  faster, since fewer pixels need to be operated on).
   */

  gimp_drawable_mask_bounds (drawable->id, &x1, &y1, &x2, &y2);

  /* Get the size of the input image. (This will/must be the same
   *  as the size of the output image.
   */
  width  = drawable->width;
  height = drawable->height;
  bytes  = drawable->bpp;

	/*  allocate row buffers  */
	src = g_new (guchar, (x2 - x1) * bytes);

	/*  initialize the pixel regions  */
	gimp_pixel_rgn_init (&srcPR, drawable, 0, 0, width, height, FALSE, FALSE);
	gimp_pixel_rgn_init (&destPR, drawable, 0, 0, width, height, TRUE, TRUE);

 	/*  loop through the rows, performing our magic*/
	for (row = y1; row < y2; row++)
	{
		guchar r,g,b;
    int j;
		gimp_pixel_rgn_get_row (&srcPR, src, x1, row, (x2-x1));
		for (col=0; col < ((x2-x1)*bytes); col+=bytes)
		{
			r=src[col]; g=src[col+1]; b=src[col+2];
			j=(g+b)/2;
			if(r>j)
				g=b=j;
			else
				r=g=b=(r+g+b)/3;
			src[col]=r; src[col+1]=g; src[col+2]=b;
		}
		gimp_pixel_rgn_set_row (&destPR, src, x1, row, (x2-x1));
	}

	if ((row % 5) == 0)
		gimp_progress_update ((double) row / (double) (y2 - y1));

  /*  update the duotoned region  */
  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->id, TRUE);
  gimp_drawable_update (drawable->id, x1, y1, (x2 - x1), (y2 - y1));

  g_free (src);
}

gboolean run_flag = FALSE;

static void
duotone_ok_callback (GtkWidget *widget,
			 gpointer   data)
{
  run_flag = TRUE;

  gtk_widget_destroy (GTK_WIDGET (data));
}

static gint
duotone_dialog (void)
{
  GtkWidget *dlg;
  GtkWidget *vbox;
  GtkWidget *frame;

  gimp_ui_init ("duotone", FALSE);

  dlg = gimp_dialog_new ("Duotone", "duotone",
			 gimp_standard_help_func, "filters/duotone.html",
			 GTK_WIN_POS_MOUSE,
			 FALSE, TRUE, FALSE,

			 "OK", duotone_ok_callback,
			 NULL, NULL, NULL, TRUE, FALSE,
			 "Cancel", gtk_widget_destroy,
			 NULL, 1, NULL, FALSE, TRUE,

			 NULL);

  gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
		      GTK_SIGNAL_FUNC (gtk_main_quit),
		      NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), vbox, TRUE, TRUE, 0);
  gtk_widget_show (vbox);

  frame =
    gimp_radio_group_new2 (TRUE, "Mode",
		gimp_radio_button_update,
			   &DuotoneChannel, (gpointer) DuotoneChannel,

			   "Keep Odd Fields", (gpointer) 0, NULL,
			   "Keep Even Fields", (gpointer) 1, NULL,

				 NULL);
  gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);
  gtk_widget_show (frame);

  gtk_widget_show (dlg);

  gtk_main ();
  gdk_flush ();

  return run_flag;
}

