/*
    The Fixer - A GIMP plug-in for fixing borders and applying themes to images
    Copyright (C) 2000  Paul Francis Harrison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <unistd.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

/* Ideas:
    - morphing, take as corpus first and last frame, use to fix intermediate
      frames that got blurred/stretched/etc
      
   TODO:
    - on repeat, is the drawable still valid??
    - make Patch more efficient
    - integrate instructions into GIMP help system
 */

/* A class to find the nearest matching vector from a set of vectors */
#include "patch_classifier.h"

/* Patch class stores a 3x3 patch of pixels and defines various vector style
   operations on them */

inline int wrap(int value,int max) {
  return value < 0 ? value + max : (value >= max ? value - max : value);
}

struct Patch {
  int nData;
  guchar *data;

  Patch() : nData(0), data(0) { }
  ~Patch() { delete[] data; }

  void size(int n) { 
    if (n != nData) {
      delete[] data;
      nData = n;
      data = new guchar[n];
    }
  }

  Patch(const Patch &assign) {
    nData = assign.nData;
    data = new guchar[nData];
    for(int i=0;i<nData;i++) 
      data[i] = assign.data[i];
  }

  void operator = (const Patch &assign) {
    size(assign.nData);
    for(int i=0;i<nData;i++)
      data[i] = assign.data[i];
  }

  /* Exctract a patch of pixels from an image */
  void buildFrom(int channels,int bpp,int width,int height,guchar *image,
                 int x,int y,
                 int a=1,int b=0,int c=0,int d=1) {
    /* abcd form transformation matrix
       [ a b ]
       [ c d ]
       to allow rotation, mirroring of patches */
    size(channels*9);
       
    int index=0;
    for(int yy=y-1;yy<=y+1;yy++)
      for(int xx=x-1;xx<=x+1;xx++)
        for(int i=0;i<channels;i++,index++) {
	  data[index] = image[
	    i+
	    wrap(a*xx+b*yy,width)*bpp+
	    wrap(c*xx+d*yy,height)*bpp*width
	  ];
	}
  }
};

Patch operator + (const Patch &a,const Patch &b) {
  Patch result;
  result.size(a.nData);
  for(int i=0;i<a.nData;i++)
    result.data[i] = a.data[i] + b.data[i];

  return result;
}

Patch operator - (const Patch &a,const Patch &b) {
  Patch result;
  result.size(a.nData);
  for(int i=0;i<a.nData;i++)
    result.data[i] = a.data[i] - b.data[i];

  return result;
}

Patch operator / (const Patch &a,int b) {
  Patch result;
  result.size(a.nData);
  for(int i=0;i<a.nData;i++)
    result.data[i] = a.data[i] / b;

  return result;
}

int _distance(const Patch &a,const Patch &b) {
  int sum = 0;
  for(int i=0;i<a.nData;i++)
    sum += abs(b.data[i]-a.data[i]);

  return sum;
}

/* A structure for holding options about what the user wants
   to do to the image exactly */
struct Parameters {
  int tileVertical, tileHorizontal, dontChangeBorder;
  gint32 corpusId; //-1 for default
  int iterations;  //-1 == forever
  
  int allowRotation, allowVerticalMirror, allowHorizontalMirror;
};

/* Various Gtk handler functions */

void optionToggled(GtkWidget *widget, gpointer data) {
  *((int*)data) = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}

void setToTrue(GtkWidget *widget, gpointer data) {
  *((int *)data) = 1;
}

void buttonPressed(GtkWidget *widget, gpointer data) {
  *((int *)data) = 1;
  gtk_main_quit();
}

gint handleDeleteEvent(GtkWidget *window, GdkEvent *event, gpointer data) {
  *((int *)data) = 1;
  return TRUE; //Will destroy window ourself in due course 
}

gint handleDeleteEventAndQuit(GtkWidget *window, GdkEvent *event, 
                              gpointer data) {
  *((int *)data) = 1;
  gtk_main_quit();
  return TRUE; //Will destroy window ourself in due course 
}

gint menuConstrainRGB(gint32 imageId, gint32 drawableId, gpointer data) {
  return drawableId == -1 || gimp_drawable_is_rgb(drawableId); 
}

gint menuConstrainGray(gint32 imageId, gint32 drawableId, gpointer data) {
  return drawableId == -1 || gimp_drawable_is_gray(drawableId); 
}

void menuCallback(gint32 id, gpointer data) {
  *((gint32*)data) = id;
}

/* Open a dialog, get user to select options */
bool getParametersByAsking(Parameters &param, int defaultDrawable) {
  int iterateForever = 1, okPressed = 0, cancelPressed = 0;
  param.tileVertical = 
  param.tileHorizontal = 1;
  param.dontChangeBorder = 0;
  param.corpusId = -1;
  param.iterations = -1;
  param.allowRotation = 
  param.allowVerticalMirror = 
  param.allowHorizontalMirror = 0;
  param.corpusId = defaultDrawable;

  GtkWidget *window;
  window = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(window), "The Fixer");
  gtk_window_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
                     GTK_SIGNAL_FUNC(handleDeleteEventAndQuit), &cancelPressed);
  
  /* Ok, Cancel */
  
  GtkWidget *hbbox;
  hbbox = gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(hbbox), 4);
  gtk_box_pack_end(GTK_BOX(GTK_DIALOG(window)->action_area),hbbox, 
                   FALSE,FALSE,0);

  GtkWidget *button;
  button = gtk_button_new_with_label("OK");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
                     GTK_SIGNAL_FUNC(buttonPressed), &okPressed);
  gtk_box_pack_start(GTK_BOX(hbbox), button, FALSE,FALSE,0);
  gtk_widget_grab_default(button);
  
  button = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
                     GTK_SIGNAL_FUNC(buttonPressed), &cancelPressed); 
  gtk_box_pack_start(GTK_BOX(hbbox), button, FALSE,FALSE,0);
  
  /* Options */

  GtkWidget *notebook;
  notebook = gtk_notebook_new();
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox),notebook,TRUE,TRUE,4);

  GtkWidget *main;
  main = gtk_vbox_new(FALSE,0);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), main,
                           gtk_label_new("Options"));

  GtkWidget *box;
  box = gtk_hbox_new(FALSE,0);
  gtk_box_pack_start_defaults(GTK_BOX(main), box);
  
  GtkWidget *label;
  label = gtk_label_new("Corpus: ");
  gtk_box_pack_start(GTK_BOX(box), label, FALSE,FALSE,0);

  GtkWidget *menu;
  menu = gimp_drawable_menu_new(
                  gimp_drawable_is_rgb(defaultDrawable) ? 
                    menuConstrainRGB : menuConstrainGray, 
                  menuCallback, 
                  &param.corpusId,defaultDrawable); 

  GtkWidget *optionMenu;
  optionMenu = gtk_option_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(optionMenu), menu);
  gtk_box_pack_start(GTK_BOX(box), optionMenu, FALSE,FALSE,0);
  
  GtkWidget *option;
  option = gtk_check_button_new_with_label("Make vertically tileable");
  gtk_signal_connect(GTK_OBJECT(option), "toggled",
                     GTK_SIGNAL_FUNC(optionToggled), &param.tileVertical);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(option), TRUE);
  gtk_box_pack_start_defaults(GTK_BOX(main), option);

  option = gtk_check_button_new_with_label("Make horizontally tileable");
  gtk_signal_connect(GTK_OBJECT(option), "toggled",
                     GTK_SIGNAL_FUNC(optionToggled), &param.tileHorizontal);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(option), TRUE);
  gtk_box_pack_start_defaults(GTK_BOX(main), option);

  option = gtk_check_button_new_with_label("Don't change border pixels");
  gtk_signal_connect(GTK_OBJECT(option), "toggled",
                     GTK_SIGNAL_FUNC(optionToggled), &param.dontChangeBorder);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(option), FALSE);
  gtk_box_pack_start_defaults(GTK_BOX(main), option);

  GtkWidget *radio;
  radio = gtk_radio_button_new_with_label(NULL,"Iterate until told to stop");
  gtk_signal_connect(GTK_OBJECT(radio), "toggled",
                     GTK_SIGNAL_FUNC(optionToggled), &iterateForever);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), TRUE);
  gtk_box_pack_start_defaults(GTK_BOX(main), radio);
  
  box = gtk_hbox_new(FALSE,0);
  gtk_box_pack_start_defaults(GTK_BOX(main), box);

  radio = gtk_radio_button_new_with_label_from_widget(
	       GTK_RADIO_BUTTON(radio),
               "Stop after this many iterations: "); 
  gtk_box_pack_start(GTK_BOX(box), radio, FALSE,FALSE,0);

  GtkWidget *iterationEntry;
  iterationEntry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(iterationEntry),"50");
  gtk_box_pack_start(GTK_BOX(box), iterationEntry, FALSE,FALSE,0);

  /* Esoterica */

  GtkWidget *esoterica;
  esoterica = gtk_vbox_new(FALSE,0);

  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), esoterica,
                           gtk_label_new("Esoterica"));
  
  option = gtk_check_button_new_with_label("Use rotations of corpus");
  gtk_signal_connect(GTK_OBJECT(option), "toggled",
                     GTK_SIGNAL_FUNC(optionToggled), &param.allowRotation);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(option), FALSE);

  gtk_box_pack_start_defaults(GTK_BOX(esoterica), option);

  option = gtk_check_button_new_with_label("Use vertical mirror of corpus");
  gtk_signal_connect(GTK_OBJECT(option), "toggled",
                     GTK_SIGNAL_FUNC(optionToggled), 
		     &param.allowVerticalMirror);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(option), FALSE);

  gtk_box_pack_start_defaults(GTK_BOX(esoterica), option);

  option = gtk_check_button_new_with_label("Use horizontal mirror of corpus");
  gtk_signal_connect(GTK_OBJECT(option), "toggled",
                     GTK_SIGNAL_FUNC(optionToggled), 
		     &param.allowHorizontalMirror);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(option), FALSE);

  gtk_box_pack_start_defaults(GTK_BOX(esoterica), option);

  gtk_widget_show_all(window);

  gtk_main();

  if (iterateForever)
    param.iterations = -1;
  else {
    gchar *chars = gtk_editable_get_chars(GTK_EDITABLE(iterationEntry),0,-1);

    if (sscanf(chars," %d ",&param.iterations) != 1) {
      param.iterations = -1;
      gimp_message("The number of iterations specified doesn't seem to be "
                   "a number, will iterate until told to stop.");
    }
    
    g_free(chars);
  }
  
  gtk_object_destroy(GTK_OBJECT(window));
  
  /* Actually remove the damn thing. Is there a better way to do this? */
  while(gtk_events_pending())
    gtk_main_iteration();
  
  return cancelPressed;
}

/* Restore the last parameters used from deep in the bowels of
   the main GIMP app */
bool getLastParameters(Parameters &param) {
  gimp_get_data("plug_in_fixer", &param);

  return false;
}

/* Convert argument list into parameters */
bool getParametersFromList(Parameters &param, int nArgs, GParam *args) {
  if (nArgs != 10)
    return true;

  param.tileVertical          = args[3 ].data.d_int32;
  param.tileHorizontal        = args[4 ].data.d_int32;
  param.dontChangeBorder      = args[5 ].data.d_int32;
  param.corpusId              = args[6 ].data.d_int32;
  param.iterations            = args[7 ].data.d_int32;
  param.allowRotation         = args[8 ].data.d_int32;
  param.allowVerticalMirror   = args[9 ].data.d_int32;
  param.allowHorizontalMirror = args[10].data.d_int32;

  return false;
}

/* Class for displaying a preview window */
struct PreviewWindow {
  GtkWidget *window, *preview, *label;
  int okSignalled, cancelSignalled;

  //Size of preview widget
  int width,height;

  PreviewWindow();
};

void handleSizeAllocate(GtkWidget *widget, GtkAllocation *alloc, 
                        gpointer data) {
  ((PreviewWindow*)data)->width = alloc->width;
  ((PreviewWindow*)data)->height = alloc->height;
}

PreviewWindow::PreviewWindow() {
  okSignalled = 0;
  cancelSignalled = 0;
  
  window = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(window),"The Fixer");
  gtk_window_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
                     GTK_SIGNAL_FUNC(handleDeleteEvent), 
		     &cancelSignalled);
  
  gtk_preview_set_gamma(gimp_gamma());
  
  preview = gtk_preview_new(GTK_PREVIEW_COLOR);
  gtk_preview_size(GTK_PREVIEW(preview),150,150);
  gtk_preview_set_expand(GTK_PREVIEW(preview),TRUE);
  gtk_signal_connect(GTK_OBJECT(preview),"size-allocate",
                     GTK_SIGNAL_FUNC(handleSizeAllocate), this);
  
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), preview, 
                     TRUE, TRUE, 4);

  label = gtk_label_new("Constructing classifier tree...");

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), label, FALSE, FALSE, 0);

  GtkWidget *hbbox;
  hbbox = gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(hbbox), 4);

  gtk_box_pack_end(GTK_BOX(GTK_DIALOG(window)->action_area), hbbox, 
                   FALSE,FALSE,0);

  GtkWidget *okButton;
  okButton = gtk_button_new_with_label("OK");
  GTK_WIDGET_SET_FLAGS(okButton, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(okButton),"clicked",GTK_SIGNAL_FUNC(setToTrue),
                     &okSignalled);
  gtk_box_pack_start(GTK_BOX(hbbox), okButton, FALSE,FALSE,0);
  gtk_widget_grab_default(okButton);

  GtkWidget *cancelButton;
  cancelButton = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS(cancelButton, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(cancelButton),"clicked",
                     GTK_SIGNAL_FUNC(setToTrue), 
		     &cancelSignalled);

  gtk_box_pack_start(GTK_BOX(hbbox), cancelButton, FALSE,FALSE,0);
}

/* Do the whole Fixer thang. */
bool fix(int channels /* 3=RGB, 1=Gray */,
         guchar *corpusData, int corpusWidth, int corpusHeight, int corpusBpp,
         bool corpusHasAlpha,
         guchar *data,int width,int height,int bpp,bool hasAlpha,
	 guchar *mask,
         Parameters parameters) {
  /* Check if images have sane sizes */
  if (corpusWidth < 3 || corpusHeight < 3 || width < 3 || height < 3)
    return false;
	 
  gimp_progress_init("Constructing classifier tree...");
  
  Classifier<Patch> classifier;
  int nPatches =
    (corpusWidth-2)*(corpusHeight-2)*
      (1+
       (parameters.allowRotation ? 3 : 0)+
       (parameters.allowVerticalMirror ? 1 : 0)+
       (parameters.allowHorizontalMirror ? 1 : 0)
      );

  Patch *input = new Patch[nPatches];

  /* Macro to check alpha value, transparent regions are omitted from
     corpus */
# define BAD(xx,yy) (corpusData[((xx)+(yy)*corpusWidth+1)*corpusBpp-1] < 128)

  int i=0;
  for(int y=1;y<corpusHeight-1;y++)
    for(int x=1;x<corpusWidth-1;x++) 
      if (!corpusHasAlpha ||
	  !(BAD(x-1,y-1) || BAD(x-1,y) || BAD(x-1,y+1) ||
	    BAD(x  ,y-1) || BAD(x  ,y) || BAD(x  ,y+1) ||
	    BAD(x+1,y-1) || BAD(x+1,y) || BAD(x+1,y+1))) {
	input[i++].buildFrom(channels,corpusBpp,corpusWidth,corpusHeight,
			     corpusData,x,y);

	if (parameters.allowRotation) {
	  input[i++].buildFrom(channels,corpusBpp,corpusWidth,corpusHeight,
			       corpusData,x,y,-1,0,0,-1);
	  input[i++].buildFrom(channels,corpusBpp,corpusWidth,corpusHeight,
			       corpusData,x,y,0,1,-1,0);
	  input[i++].buildFrom(channels,corpusBpp,corpusWidth,corpusHeight,
			       corpusData,x,y,0,-1,1,0);
	}

	if (parameters.allowVerticalMirror)
	  input[i++].buildFrom(channels,corpusBpp,corpusWidth,corpusHeight,
			       corpusData,x,y,1,0,0,-1);

	if (parameters.allowHorizontalMirror)
	  input[i++].buildFrom(channels,corpusBpp,corpusWidth,corpusHeight,
			       corpusData,x,y,-1,0,0,1);
      }

# undef BAD

  if (i == 0)
    return false;

  classifier.build(i,input);
  delete[] input;

  Patch *patches = new Patch[width*height];
  int  *distances = new int[width*height];
  int  *bestDistances = new int[width*height];
  bool *changeable = new bool[width*height];

  if (parameters.dontChangeBorder) {
    if (hasAlpha)
      for(int i=0;i<width*height;i++)
        mask[i] = guint(mask[i])*guint(data[i*bpp+bpp-1])>>8;

#   define BAD(xx,yy) ( \
      (((xx) < 0 || (xx) >= width) && !parameters.tileHorizontal) || \
      (((yy) < 0 || (yy) >= height) && !parameters.tileVertical) || \
      mask[wrap((xx),width)+wrap((yy),height)*width] < 128 \
    )
    
    for(int y=0;y<height;y++)
      for(int x=0;x<width;x++)
        changeable[x+y*width] = 
	  !(BAD(x-1,y-1) || BAD(x-1,y) || BAD(x-1,y+1) ||
	    BAD(x  ,y-1) || BAD(x  ,y) || BAD(x  ,y+1) ||
	    BAD(x+1,y-1) || BAD(x+1,y) || BAD(x+1,y+1));
	    
#   undef BAD
  } else {
    for(int i=0;i<width*height;i++)
      changeable[i] = true;
  }

  /* Will work well if corpus == thing to fix, otherwise, well, gotta use
     something, shrug. */
  for(int y=0;y<height;y++)
    for(int x=0;x<width;x++) {
      int yy = y;
      if (yy < 1) yy = 1;
      if (yy > corpusHeight-2) yy = corpusHeight-2;
      int xx = x;
      if (xx < 1) xx = 1;
      if (xx > corpusWidth-2) xx = corpusWidth-2;

      patches[y*width+x].buildFrom(channels,corpusBpp,corpusWidth,corpusHeight,
                                   corpusData,xx,yy);
    }

  PreviewWindow previewer;

  gimp_progress_init("Iterating...");
  gtk_label_set(GTK_LABEL(previewer.label),"Iterating");
  
  gtk_widget_show_all(previewer.window);

  int vBorder = (parameters.tileVertical   ? 0 : 1);
  int hBorder = (parameters.tileHorizontal ? 0 : 1);

  for(int iterations=1;
      iterations != parameters.iterations+1 && 
      !previewer.okSignalled && 
      !previewer.cancelSignalled;
      iterations++) {
    if (parameters.iterations > 1)
      gimp_progress_update(iterations/(parameters.iterations-1.0));
      
    for(int y=0;y<height;y++)
      for(int x=0;x<width;x++)
	bestDistances[y*width+x] = -1;

    Patch patch;
    for(int y=vBorder;y<height-vBorder && !previewer.okSignalled && 
	!previewer.cancelSignalled;y++) {
      for(int x=hBorder;x<width-hBorder;x++) {
	patch.buildFrom(channels,bpp,width,height,data,x,y);

	patches[y*width+x] = classifier.nearest(patch,patches[y*width+x], distances[y*width+x]);
      }
    
      while(gtk_events_pending())
	gtk_main_iteration();
    }

    if (previewer.okSignalled || previewer.cancelSignalled) break;

    for(int y=vBorder;y<height-vBorder;y++)
      for(int x=hBorder;x<width-hBorder;x++) {
	int i=0;
	for(int yy=-1;yy<=1;yy++)
	  for(int xx=-1;xx<=1;xx++,i+=channels) {
	    int offset = wrap(x+xx,width)+wrap(y+yy,height)*width;
	    if (changeable[offset] &&
	        (bestDistances[offset] < distances[x+y*width] ||
		 (bestDistances[offset] == distances[x+y*width] && (rand()&1))
		)) {
	      bestDistances[offset] = distances[x+y*width];
	      for(int j=0;j<channels;j++)
		data[offset*bpp+j] = patches[x+y*width].data[i+j];
	    }
	  }
      }

    while(gtk_events_pending())
      gtk_main_iteration();
      
    /* This could be made faster */
    guchar *previewRow = new guchar[previewer.width*3];
    int addX = -previewer.width/2 + (parameters.tileHorizontal?0:width/2);
    while(addX < 0) addX += width;
    int addY = -previewer.height/2 + (parameters.tileVertical?0:height/2);
    while(addY < 0) addY += height;
    for(int y=0;y<previewer.height;y++) {
      for(int x=0;x<previewer.width;x++)
	for(int i=0;i<3;i++)
	  previewRow[x*3+i] = data[
	    (channels==1?0:i)+
	    ((x+addX) % width)*bpp+
	    ((y+addY) % height)*bpp*width]; 
      gtk_preview_draw_row(GTK_PREVIEW(previewer.preview),previewRow,0,y,
			   previewer.width);
    }
    delete[] previewRow;
    gtk_widget_draw(previewer.preview,0);

    char text[80];
    sprintf(text,"Iterations: %d",iterations);
    gtk_label_set(GTK_LABEL(previewer.label),text);

    while(gtk_events_pending())
      gtk_main_iteration();
  }
    
  delete[] patches;
  delete[] distances;
  delete[] bestDistances;
  delete[] changeable;

  gtk_object_destroy(GTK_OBJECT(previewer.window));
  
  return !previewer.cancelSignalled;
}

/* Talking to GIMP stuff */

static void query();
static void run(gchar *name,
                gint nparams,
		GParam *param,
		gint *nreturn_vals,
		GParam **return_vals);

GPlugInInfo PLUG_IN_INFO = {
  NULL, /* init_proc */
  NULL, /* quit_proc */
  query, /* query_proc */
  run, /* run_proc */
};

/* Macro to define the usual plugin main function */
MAIN()

/* Add capabilities of this plugin to Procedural DataBase */
static void query() {
  static GParamDef args[] = {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE, "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Input drawable" },

    { PARAM_INT32, "vtile", "Make tilable vertically" },
    { PARAM_INT32, "htile", "Make tilable horizontally" },
    { PARAM_INT32, "noborder", "Dont change border pixels" },
    { PARAM_INT32, "corpus", "Layer to use as corpus, -1 for default" },
    { PARAM_INT32, "iters", "Number of iterations, -1 for user prompted" },
    
    { PARAM_INT32, "rotate", "Include rotations of corpus" },
    { PARAM_INT32, "vmirror", "Include vertical mirror of corpus" },
    { PARAM_INT32, "hmirror", "Include horizontal mirror of corpus" }
  };

  GParamDef *return_vals = NULL;
  gint nargs = sizeof(args)/sizeof(args[0]);
  gint nreturn_vals = 0;

  gimp_install_procedure("plug-in-fixer",
                         "Make tiles, eat unwanted blemishes, "
			 "apply themes to images, etc",
			 "",
			 "Paul Francis Harrison",
			 "Paul Francis Harrison",
			 "2000",
			 "<Image>/Filters/Map/Fixer...",
			 "RGB*, GRAY*",
			 PROC_PLUG_IN,
			 nargs, nreturn_vals,
			 args, return_vals);
}


static void run(gchar *name,
                gint nparams,
		GParam *param,
		gint *nreturn_vals,
		GParam **return_vals) {
  static GParam values[1];
  GDrawable *drawable, *corpusDrawable;
  GPixelRgn region;
  guchar *data, *corpusData, *mask;
  bool hasSelection;
  int x1,y1,x2,y2,xoff,yoff,width,height;
  int corpusWidth,corpusHeight,corpusBpp;
  
  *nreturn_vals = 1;
  *return_vals = values;
  values[0].type = PARAM_STATUS;
  values[0].data.d_status = STATUS_SUCCESS;

  /* Init GTK. Yuck! */
  int argc = 1;
  char **argv = new char *[2];
  argv[0] = "fixer";
  argv[1] = 0;
  gtk_init(&argc, &argv);
  gtk_rc_parse(gimp_gtkrc());
 
  /* Get drawable */
  drawable = gimp_drawable_get(param[2].data.d_drawable);

  if (!gimp_drawable_is_rgb(drawable->id) &&
      !gimp_drawable_is_gray(drawable->id)) {
    gimp_drawable_detach(drawable);
    values[0].data.d_status = STATUS_EXECUTION_ERROR;
    return;
  }

  /* Deal with run mode (parameter zero) */
  Parameters parameters;
  
  bool cancelled = false;
  switch(param[0].data.d_int32) {
    case RUN_INTERACTIVE :
      cancelled = getParametersByAsking(parameters,drawable->id); 
      break;
    case RUN_NONINTERACTIVE :
      cancelled = getLastParameters(parameters); 
      break;
    case RUN_WITH_LAST_VALS :
      cancelled = getParametersFromList(parameters, nparams, param); 
      break;
  }

  if (cancelled) {
    gimp_drawable_detach(drawable);
    values[0].data.d_status = STATUS_EXECUTION_ERROR;
    return;
  }

  if (parameters.corpusId == -1)
    parameters.corpusId = drawable->id;

  /* TODO: Check corpusId is valid */

  /* Fetch the corpus */
  corpusDrawable = gimp_drawable_get(parameters.corpusId);
  
  gimp_drawable_mask_bounds(corpusDrawable->id,&x1,&y1,&x2,&y2);
  corpusWidth = x2-x1;
  corpusHeight = y2-y1;
  corpusBpp = corpusDrawable->bpp;
  corpusData = new guchar[corpusWidth*corpusHeight*corpusBpp];
  
  gimp_pixel_rgn_init(&region,corpusDrawable,x1,y1,corpusWidth,corpusHeight,
                      FALSE,FALSE);
  gimp_pixel_rgn_get_rect(&region, corpusData, x1,y1,corpusWidth,corpusHeight);

  gimp_drawable_detach(corpusDrawable);

  /* Fetch the image to be warped, munged, crunched, twisted... */
  hasSelection = gimp_drawable_mask_bounds(drawable->id,&x1,&y1,&x2,&y2);
  width = x2-x1;
  height = y2-y1;
  
  data = new guchar[width*height*drawable->bpp];

  gimp_pixel_rgn_init(&region,drawable,x1,y1,width,height,FALSE,FALSE);
  gimp_pixel_rgn_get_rect(&region, data, x1,y1,width,height);

  /* Fetch the selection mask if necessary */
  mask = new guchar[width*height];
  for(int i=0;i<width*height;i++)
    mask[i] = 255;
  
  if (hasSelection) {
    int selId = gimp_image_get_selection(gimp_drawable_image_id(drawable->id));
    if (selId != -1) {
      GDrawable *maskDrawable = gimp_drawable_get(selId);

      gimp_drawable_offsets(drawable->id, &xoff, &yoff);

      gimp_pixel_rgn_init(&region,maskDrawable,x1+xoff,y1+yoff,
                          width,height,FALSE,FALSE);
      gimp_pixel_rgn_get_rect(&region,mask,x1+xoff,y1+yoff,width,height);

      gimp_drawable_detach(maskDrawable);
    }
  }

  /* Do it! */
  bool success = fix(gimp_drawable_is_rgb(drawable->id) ? 3 : 1,
                     corpusData,corpusWidth,corpusHeight,corpusBpp,
                     gimp_drawable_has_alpha(parameters.corpusId),
                     data,width,height,drawable->bpp,
                     gimp_drawable_has_alpha(drawable->id), 
		     mask, 
		     parameters);

  if (success) {
    /* Write result to region */
    gimp_pixel_rgn_init(&region,drawable,x1,y1,width,height,TRUE,TRUE);
    gimp_pixel_rgn_set_rect(&region, data, x1,y1,width,height);

    /* Voodoo to update actual image */
    gimp_drawable_flush(drawable);
    gimp_drawable_merge_shadow(drawable->id,TRUE);
    gimp_drawable_update(drawable->id,x1,y1,width,height);

    /* Store parameters deep in the bowels of the GIMP */
    gimp_set_data("plug_in_fixer", &parameters, sizeof(parameters));
  }
  
  /* Unget drawable */
  gimp_drawable_detach(drawable);
  
  delete[] data;
  delete[] corpusData;
  delete[] mask;
  
  gimp_displays_flush();
}

