/* dialog.c: user interface for GIMP Adaptive Contrast Enhancment plug-in */
/* Copyright (C) 1999-2000 Kevin M. Turner <kevint@poboxes.com> */
/* $Id: dialog.c,v 1.28 2000/09/15 02:26:17 acapnotic Exp $ */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.  See the file COPYING for details.
 *
 */

#include <config.h> /* autostuff */

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "plugin-intl.h" /* i18n - gettext */

#include "types.h"
#include "dialog.h"
#include "preview.h"

typedef enum 
{ 
  UPDATE_RECALC, 
  UPDATE_OUTPUT 
} ControlUpdateType;

typedef enum
{
  INTERFACE_SIMPLE,
  INTERFACE_NORMAL,
  INTERFACE_COMPLEX
} InterfaceLevel;

typedef struct {
  /* Perhaps theses should exist as named data members of a GtkObject... */
  GtkAdjustment* strength;
  GtkAdjustment* bradj;
  GtkAdjustment* iterations;
  GtkAdjustment* wsize;
  GtkAdjustment* wsize_x;
  GtkAdjustment* wsize_y;
  GtkAdjustment* smoothing;
  GtkAdjustment* coefftol;

  GtkAdjustment* colormethod;

  gboolean auto_update;

  /* private: */
  guint revision;
  guint out_revision;

  ManyviewMatrix*   manyview_matrix;
  Glace_ClientData* bigview;
  guint bigview_revision;
  guint bigview_out_revision;
} ControlBackend;

typedef struct {
  Glace_ClientData* clientData;
  ControlBackend* control;
} AutoUpdateData;

extern GimpPDBStatusType* const gimp_ace_status;

/* Dialog creations */
static GtkWidget* ace_dialog_new   (GtkSignalFunc     ok_callback, 
				    GtkSignalFunc     close_callback);

static gboolean   dialog_setup     (AutoUpdateData*   update_data);

/* General dialog callbacks */
static void       ok_callback      (GtkWidget*        widget, 
				    GtkWidget*        dlg);

static void       close_callback   (GtkWidget*        unused_widget, 
				    gpointer          unused_data);

/* Preview-related callbacks */
static void       update_preview_cb(GtkWidget*        unused_widget, 
				    AutoUpdateData*   update_data);

static void       cancel_preview_cb(GtkWidget*        unused_widget, 
				    GlaceGimpPreview* preview);

static void       auto_toggle_cb   (GtkWidget*        toggle, 
                                    AutoUpdateData*   update_data);

/* Callbacks for specific controls. */
static void       omenu_callback   (GtkWidget*        menuitem, 
				    GtkAdjustment*    adj);

static void       iter_changed     (GtkAdjustment*    adj, 
			            GtkProgressBar*   progress);

static void       link_toggled     (GimpChainButton*  chainbutton, 
				    GtkObject*        obj);

/* ControlBackend functions */
static ControlBackend* control_backend_new(AceValues* acevals,
					   const GlaceGimpDrawableInfo* drawable_info);

static void      control_updated_callback (GtkAdjustment*    adjustment,
				           ControlBackend*   control_backend);

static void      control_register_manyview(ControlBackend*   control, 
					   ManyviewMatrix*   manyview_matrix);

static void      control_register_preview (ControlBackend*   control,
					   Glace_ClientData* clientData);

/* this one is private to control_backend_new */
static GtkAdjustment* control_new         (ControlBackend*   control_backend, 
					   gchar*            labeltxt, 
					   float             value, 
					   float             min, 
					   float             max,
					   float             small_inc, 
					   float             page_inc, 
					   float             page_size,
					   ControlUpdateType update_type,
					   InterfaceLevel    interface_level);

/* GSourceFunc */
static gboolean  bounce_control_to_manyview(ControlBackend* control);


/* ControlBackend interfaces to other data structures */
static void control_to_manyview(ControlBackend* control,
				ManyviewMatrix* mv_matrix);

static void control_to_cfginfo(const ControlBackend* control, 
			       Glace_CfgInfo*        cfgInfoPtr);

static void control_to_acevals(const ControlBackend* control, 
			       AceValues*            acevals);

/* Padding options and whatnot. */
/* FIXME: someone with an eye for GUI should go through and tune these a bit,
   but having to hardcode things like padding in numbers of pixels seems pretty
   lame to me. */
#define SPACING 4
#define HBOX_SPACING SPACING
#define DIALOG_BORDER_WIDTH 0
#define OUTER_BORDER_WIDTH 6
#define ACTION_AREA_BORDER_WIDTH OUTER_BORDER_WIDTH
#define TOP_AREA_BORDER_WIDTH OUTER_BORDER_WIDTH
#define TABLE_ROW_SPACING SPACING
#define TABLE_COL_SPACING SPACING

#define SPIN_ATTOPT_X GTK_EXPAND | GTK_FILL
#define SPIN_ATTOPT_Y 0

#define SCALE_ATTOPT_X GTK_EXPAND | GTK_FILL
#define SCALE_ATTOPT_Y 0

#define LABEL_ATTOPT 0

/* This has to match the number and ordering of the Glace_ColorMethods,
 * as defined in glace.h */
static const char* const Glace_ColorMethodNames[] =
{
  N_("add white"),
  N_("+correctdark"),
  N_("limit"),
  N_("+correctdark"),
  N_("gray only"),
  N_("0.6.2 white"),
  N_("+correctdark"),
  N_("+correctlight"),
  N_("+correctlight+dark"), 
  NULL                        /* with a NULL to signal the end */
};

/* This is the only public function in this file.  
   Nearly the entire life of the interactive-mode plug-in is spanned by this 
   function.  Both gtk_init and the gtk mainloop are called from here.  */
GimpPDBStatusType 
ace_dialog(AceValues* const acevals,
	   const GlaceGimpDrawableInfo* const drawable_info)
{
  ControlBackend* controls; 
  Glace_ClientData* clientData;
  int i; 
  GlaceGimpPreview* ace_preview;
  ManyviewMatrix* manyview;
  
  GtkWidget *dlg, *pane, *vbox, *hbox, *table, *frame, *align;
  GtkWidget *label, *hseparator;
  GtkWidget *bradj_scale, *strength_scale;
  GtkWidget *link_toggle, *button, *toggle;
  GtkWidget *iter_spin, *size_spin, *smooth_spin;
  GtkWidget *menuitem, *color_menu, *color_omenu;
  GtkWidget *notebook;

  GtkTooltips *tooltips;

  AutoUpdateData* const update_data = g_new0(AutoUpdateData,1);

#ifdef SHOW_THRESHOLD_CONTROL
  GtkWidget *foobarum_spin;
#endif

  /* dlg creation should be first because it does important
     stuff like initalize GTK+ */
  dlg = ace_dialog_new((GtkSignalFunc) ok_callback,
		       (GtkSignalFunc) close_callback);

  controls = control_backend_new(acevals, drawable_info);
  clientData=clientdata_new(drawable_info,NULL);
  control_register_preview(controls, clientData);

  update_data->clientData = clientData;
  update_data->control = controls;

  gtk_object_set_data(GTK_OBJECT(dlg),"acevals",acevals);
  gtk_object_set_data(GTK_OBJECT(dlg),"control_backend",controls);
  gtk_object_set_data(GTK_OBJECT(dlg),"clientData",clientData);

  tooltips=gtk_tooltips_new();

  /* Pane */
  
  pane=gtk_hpaned_new();

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dlg)->vbox), pane,
		      EXPAND, FILL, 0);
  
  /* Left Pane: Preview */

  clientData->wData->preview = ace_preview = 
	  ace_preview_new(clientData->wData);

  vbox = gtk_vbox_new (!HOMOG, 0);

  gtk_paned_pack1(GTK_PANED(pane), vbox, RESIZE, !SHRINK);
  
  hbox = gtk_hbox_new(!HOMOG, SPACING);

  button=gtk_button_new_with_label(_("Cancel"));

  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(cancel_preview_cb), ace_preview);

  gtk_box_pack_end(GTK_BOX(hbox), button, !EXPAND, !FILL, 0);

  toggle = gtk_check_button_new_with_label(_("Auto"));
  button = gtk_button_new_with_label(_("Update"));

  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(update_preview_cb), 
		     update_data);
  
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(toggle),
			       acevals->ui_auto);

  gtk_signal_connect(GTK_OBJECT(toggle),"toggled",
		     GTK_SIGNAL_FUNC(gimp_toggle_button_update),
		     &(controls->auto_update));

  gtk_signal_connect(GTK_OBJECT(toggle),"toggled",
		     GTK_SIGNAL_FUNC(auto_toggle_cb),
		     update_data);

  gtk_box_pack_end(GTK_BOX(hbox), button, !EXPAND, !FILL, 0);
  gtk_box_pack_end(GTK_BOX(hbox), toggle, !EXPAND, !FILL, 0);

  align = gtk_alignment_new(1,0.5,0,0);
  gtk_container_add(GTK_CONTAINER(align), hbox);
  gtk_box_pack_end(GTK_BOX(vbox), align, !EXPAND, !FILL, SPACING);
  
  /* Preview Full/Minis notebook */
  notebook = gtk_notebook_new();

  gtk_box_pack_start(GTK_BOX(vbox), notebook, EXPAND, FILL, 0);

  /* Full preview */

  vbox = gtk_vbox_new (!HOMOG, 0);

  gtk_box_pack_start(GTK_BOX(vbox),GTK_WIDGET(ace_preview->scrollwin),
		     EXPAND, FILL, 0);

  gtk_box_pack_end(GTK_BOX(vbox), GTK_WIDGET(ace_preview->pbar), 
		   !EXPAND, !FILL, 0);

  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox,
			   NULL);

  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook),
				  vbox, _("Full"));  

  /* Manyview */

  manyview = manyview_matrix_new(drawable_info);
  control_register_manyview(controls, manyview);
  table = manyview->tablewidget;

  gtk_container_set_border_width(GTK_CONTAINER(table),2);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook),table,NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook),table,_("Minis"));

  /**********************************/
  /* Right pane: Parameter settings */
  frame = gtk_frame_new(_("Parameter Settings"));

  gtk_paned_pack2(GTK_PANED(pane), frame, !RESIZE, !SHRINK);  

  /* FIXME: connect signals so we know when the user switched tabs. */
  notebook = gtk_notebook_new();

  gtk_container_add(GTK_CONTAINER(frame), notebook);

  /* Right pane: Parameters: Simple */
  vbox = gtk_vbox_new (FALSE, SPACING);

  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, NULL);

  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), vbox,
				  _("Simple"));
	
  /* Table which holds sliders */
  table = gtk_table_new (2, 3, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, !EXPAND, FILL, 0);
  gtk_table_set_row_spacings (GTK_TABLE (table), SPACING);
  gtk_table_set_col_spacings (GTK_TABLE (table), SPACING);

  /* slider: strength */
  strength_scale = gtk_hscale_new (controls->strength);
  gtk_widget_set_name (strength_scale, "strength_scale");
  gtk_table_attach (GTK_TABLE (table), strength_scale, 1, 2, 0, 1,
		    SCALE_ATTOPT_X, SCALE_ATTOPT_Y, 
		    0, 0);
  gtk_scale_set_value_pos (GTK_SCALE (strength_scale), GTK_POS_LEFT);
  gtk_scale_set_digits (GTK_SCALE (strength_scale), 2);
  gtk_range_set_update_policy (GTK_RANGE (strength_scale), 
			       GTK_UPDATE_DELAYED);

  /* slider: Brightness adjust */
  bradj_scale = gtk_hscale_new (controls->bradj);
  gtk_widget_set_name (bradj_scale, "bradj_scale");
  gtk_table_attach (GTK_TABLE (table), bradj_scale, 1, 2, 1, 2,
		    SCALE_ATTOPT_X, SCALE_ATTOPT_Y, 
		    0, 0);
  gtk_scale_set_value_pos (GTK_SCALE (bradj_scale), GTK_POS_LEFT);
  gtk_scale_set_digits (GTK_SCALE (bradj_scale), 2);
  gtk_range_set_update_policy (GTK_RANGE (bradj_scale), 
			       GTK_UPDATE_DELAYED);
  
  /* Link toggle */

  link_toggle = gimp_chain_button_new(GIMP_CHAIN_RIGHT);

  gimp_chain_button_set_active(GIMP_CHAIN_BUTTON(link_toggle),
			       acevals->ui_link);

  gtk_table_attach (GTK_TABLE (table), link_toggle, 2, 3, 0, 2,
		    LABEL_ATTOPT, LABEL_ATTOPT, 0, 0);

  gtk_signal_connect(GTK_OBJECT(link_toggle),"toggled",
		     link_toggled, controls->strength);

  /* Labels for brightness adjust and strength */
  label = gtk_label_new (_("Strength:"));
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
  
  label = gtk_label_new (_("Brightness Adjust:"));
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);	
  
  /* Seperator */
  hseparator = gtk_hseparator_new ();
  gtk_box_pack_start (GTK_BOX (vbox), hseparator, !EXPAND, !FILL, 0);

  /* Table which holds spinbuttons */
  /* (iterations, size, threshold) */

  table = gtk_table_new (2, 4, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, !EXPAND, !FILL, 0);
  gtk_table_set_row_spacings (GTK_TABLE (table), SPACING);
  gtk_table_set_col_spacings (GTK_TABLE (table), SPACING);

  /* Iteration spin */
  iter_spin = gtk_spin_button_new (controls->iterations, 1, 0);
  gtk_widget_set_name (iter_spin, "iteration_spin");
  gtk_table_attach (GTK_TABLE (table), iter_spin, 0, 1, 1, 2,
		    SPIN_ATTOPT_X, SPIN_ATTOPT_Y,
		    0, 0);
  
  gtk_signal_connect (GTK_OBJECT (controls->iterations), "value_changed",
		      GTK_SIGNAL_FUNC (iter_changed), ace_preview->pbar);
  /* Iteration label */
  label = gtk_label_new (_("Iterations:"));
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
  
  /* size spin */
  size_spin = gtk_spin_button_new(controls->wsize, 0.1, 0);
  gtk_widget_set_name (size_spin, "size_spin");
  gtk_table_attach (GTK_TABLE (table), size_spin, 1, 2, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 
		    0, 0);

  /* size label */
  label = gtk_label_new (_("Detail Size:"));
  gtk_table_attach (GTK_TABLE (table), label, 1, 2, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
  
  /* smoothing spin */
  smooth_spin = gtk_spin_button_new (controls->smoothing, 1, 0);
  gtk_widget_set_name (smooth_spin, "smooth_spin");
  gtk_table_attach (GTK_TABLE (table), smooth_spin, 2, 3, 1, 2,
		    SPIN_ATTOPT_X, SPIN_ATTOPT_Y,
		    0, 0);

  /* smoothing label */
  label = gtk_label_new (_("Smoothing:"));
  gtk_table_attach (GTK_TABLE (table), label, 2, 3, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
 
#ifdef SHOW_THRESHOLD_CONTROL 
  /* foobarum spin */
  foobarum_spin = gtk_spin_button_new(controls->coefftol 1, 3);
  gtk_widget_set_name (foobarum_spin, "foobarum_spin");
  gtk_table_attach (GTK_TABLE (table), foobarum_spin, 3, 4, 1, 2,
		    SPIN_ATTOPT_X, SPIN_ATTOPT_Y,
		    0, 0);

  /* foobarum label */
  label = gtk_label_new (_("Threshold:"));
  gtk_table_attach (GTK_TABLE (table), label, 3, 4, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
#endif  

  /** Color method **/
  frame=gtk_frame_new(_("Color Overflow"));
  
  color_menu=gtk_menu_new();

  for(i=0;i<GLACE_COLOR_LAST;i++)
    {
      if(Glace_ColorMethodNames[i] == NULL)
	{
	  g_error("PROGRAMMER ERROR, please inform <" "MAINTAINER_EMAIL" ">:\n"
		  "file %s:%s\n"
		  "Mismatch between Glace_ColorMethod and Glace_ColorMethodNames.\n"
		  "Only %d name strings defined here, but there are %d color methods.",
		  __FILE__, G_GNUC_PRETTY_FUNCTION,
		  i,GLACE_COLOR_LAST);
	}
      menuitem = 
	gtk_menu_item_new_with_label(gettext(Glace_ColorMethodNames[i]));

      gtk_menu_append(GTK_MENU(color_menu), menuitem);
      gtk_object_set_user_data(GTK_OBJECT(menuitem),
			       GINT_TO_POINTER(i));

      gtk_signal_connect(GTK_OBJECT(menuitem), "activate",
			 GTK_SIGNAL_FUNC(omenu_callback),
			 controls->colormethod);
    }

  if(Glace_ColorMethodNames[i] != NULL)
    {
      g_error("PROGRAMMER ERROR, please inform <" "MAINTAINER_EMAIL" ">:\n"
	      "file %s:%s\n"
	      "Mismatch between Glace_ColorMethod and Glace_ColorMethodNames,\n"
	      "there are more name strings than color methods!\n"
	      "(Or the terminating NULL on Glace_ColorMethodNames was removed.)"
	      "The extra string(s) start with \"%s\"",
	      __FILE__, G_GNUC_PRETTY_FUNCTION,
	      Glace_ColorMethodNames[i]);
    }
    
  gtk_widget_show_all(color_menu);
  
  color_omenu = gtk_option_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(color_omenu),
			   color_menu);

  if(acevals->colormethod < GLACE_COLOR_LAST)
    {
      gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 
				  acevals->colormethod);
    }
  else
    {
      /* Handle out-of-range values gracefully here.  */
      gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 0);
    }
  
  gtk_container_add(GTK_CONTAINER(frame), color_omenu);
  gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, TRUE, 0);

  if(!gimp_drawable_is_rgb(clientData->wData->d->drawable->id))
    gtk_widget_set_sensitive(frame, FALSE);
  
  /* Right pane: Parameters: Normal */
  vbox = gtk_vbox_new (FALSE, SPACING);

  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, NULL);

  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), vbox,
				  _("Normal"));

  /* Right pane: Parameters: Complex */
  vbox = gtk_vbox_new (FALSE, SPACING);

  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, NULL);

  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), vbox,
				  _("Complex"));

  /* Done */
  gtk_widget_show_all(dlg);

  gtk_idle_add((GtkFunction)dialog_setup,update_data);
  
#ifdef ACE_THREADS
  gdk_threads_enter();
  gtk_main();
  gdk_threads_leave();
#else
  gtk_main();
#endif
  
  if(*gimp_ace_status != GIMP_PDB_CANCEL)
    {
      /* FIXME: switching the target should be just another update of
	 type OUTPUT for the bigview */
      clientData->wData->target=GLACE_GIMP_PixelRgn;
      clientData->imgArraysPtr->pixelBytePad = 
	clientData->wData->d->has_alpha ? 1 : 0;

      gimp_progress_init (_("Adaptive Contrast Enhancement:"));
      gimp_ace_go(clientData, G_PRIORITY_DEFAULT_IDLE);
    }

  return(*gimp_ace_status);
}

/* Allocates, initalizes, and returns the plug-in's ControlBackend.
   All the controls are specified completely in the array of 
   ControlAdjustmentContructors here.
   FIXME: There is no corresponding destroy/free function. */
static ControlBackend* 
control_backend_new(AceValues* const av,
		    const GlaceGimpDrawableInfo* const di)
{
  ControlBackend* const s_copy = g_new0(ControlBackend,1);
#if defined (__GNUC__) && !defined (__STRICT_ANSI__)
  /* The advantage to initializing ControlBackend this way is that
     we *can't* forget to initialize some parts of it.  Unfortunately,
     not all compilers let you use non-constant initializers, as this
     is not in ANSI C (but probably is in C9X) */
  /* Yes, I was on crack when I wrote this. */
  /* No, it does not prevent you from specifying things in the wrong order. */
# define CB0(X) /* X */
# define CB_ ,
  ControlBackend s = {
#else
# define CB0(X) (s.X) =
# define CB_ ;
  ControlBackend s;
#endif

  CB0(strength) control_new(s_copy,_("Strength"),av->strength,0,1,0.01,0.1,0, 
			    UPDATE_RECALC,INTERFACE_SIMPLE) CB_

  CB0(bradj) control_new(s_copy,_("Brightness"),av->bradj,0, 1, 0.01, 0.1, 0,
			 UPDATE_RECALC, INTERFACE_NORMAL) CB_

  /* FIXME: Add "Quality", a dumbed-down Iterations/Tolerance control. */

  CB0(iterations) control_new(s_copy,_("Iterations"),av->iterations,
			      1,40,1,5,0, 
			      UPDATE_RECALC, INTERFACE_NORMAL) CB_

  /* Note: There are seperate wsize, wsize_x and wsize_y entries instead
       of just pairing _x to _y because the paired adjustment may have a
       lower maximum than the _x adjustment. */
  /* FIXME: Add AceValues for the _x and _y */
  CB0(wsize) control_new(s_copy,_("Detail Size"),av->wsize, 
			 4, MIN(di->w, di->h) - 1, 2, 10, 0, 
			 UPDATE_RECALC, INTERFACE_SIMPLE) CB_
  
  CB0(wsize_x) control_new(s_copy,_("Detail X-size"),av->wsize, 
			   4, di->w - 1, 2, 10, 0, 
			   UPDATE_RECALC, INTERFACE_NORMAL) CB_

  CB0(wsize_y) control_new(s_copy,_("Detail Y-size"),av->wsize, 
			   4, di->h - 1, 2, 10, 0, 
			   UPDATE_RECALC, INTERFACE_NORMAL) CB_

  /* FIXME: Check what the bounds on smoothing should be. */
  CB0(smoothing) control_new(s_copy,_("Smoothing"),av->smoothing,
			     0,100, 2, 10, 0,
			     UPDATE_RECALC, INTERFACE_NORMAL) CB_

  CB0(coefftol) control_new(s_copy,_("Tolerance"),av->coefftol,
			    0,1,0.0005,0.002,0,
			    UPDATE_RECALC, INTERFACE_COMPLEX) CB_

  CB0(colormethod) control_new(s_copy,_("Color method"),av->colormethod, 
			       0, GLACE_COLOR_LAST-1, 1, 1, 1, 
			       UPDATE_OUTPUT, INTERFACE_NORMAL) CB_

  /* FIXME: Add other complex controls:
     (boolean) hamming toggle 
     (list) Glace_OutputMethods
  */

  CB0(auto_update) av->ui_auto CB_
  CB0(revision) 0 CB_
  CB0(out_revision) 0 CB_
  CB0(manyview_matrix) NULL CB_
  CB0(bigview) NULL CB_ 
  CB0(bigview_revision) 0 CB_
  CB0(bigview_out_revision) 0 CB_

#if defined (__GNUC__) && !defined (__STRICT_ANSI__)
  };
#else
  /* no closing bracket needed. */
#endif
#undef CB0
#undef CB_
  
  /* Establish partnerships. */
  gtk_object_set_data(GTK_OBJECT(s.strength), "linked", 
		      GINT_TO_POINTER(av->ui_link));
  gtk_object_set_data(GTK_OBJECT(s.bradj), "linked", 
		      GINT_TO_POINTER(av->ui_link));
  gtk_object_set_data(GTK_OBJECT(s.strength), "partner", s.bradj);
  gtk_object_set_data(GTK_OBJECT(s.bradj), "partner", s.strength);

  gtk_object_set_data(GTK_OBJECT(s.wsize), "linked", GINT_TO_POINTER(TRUE));
  gtk_object_set_data(GTK_OBJECT(s.wsize), "partner", s.wsize_x);
  gtk_object_set_data(GTK_OBJECT(s.wsize), "partner2", s.wsize_y);

  *s_copy=s;

  return s_copy;
}

static GtkAdjustment*
control_new(ControlBackend* control_backend, 
	    gchar* labeltxt, 
	    float value, float min, float max,
	    float small_inc, float page_inc, float page_size,
	    ControlUpdateType update_type,
	    InterfaceLevel interface_level)
{
  GtkObject *adj;
      
  adj = gtk_adjustment_new(value, min, max, 
			   small_inc, page_inc, page_size);
  
  gtk_object_set_data(adj, "label_text", labeltxt);
  
  gtk_object_set_data(adj, "control_backend", control_backend);
  
  gtk_object_set_data(adj, "update_type", GINT_TO_POINTER(update_type));
  
  gtk_object_set_data(adj, "interface_level", 
		      GINT_TO_POINTER(interface_level));
  
  gtk_signal_connect(adj, "value_changed",
		     GTK_SIGNAL_FUNC(control_updated_callback), 
		     control_backend);

  /* FIXME: This is only useful for the outer miniviews, but currently
     it causes all previews to be recalculated (including bigview). */
  gtk_signal_connect(adj, "changed",
		     GTK_SIGNAL_FUNC(control_updated_callback), 
		     control_backend);

  return GTK_ADJUSTMENT(adj);
}

static void 
control_register_manyview(ControlBackend *control, ManyviewMatrix* manyview_matrix)
{
  g_return_if_fail(control);
  control->manyview_matrix = manyview_matrix;

  /* FIXME: Again, it would be nice if we had some safegaurd here (at
     run time, I guess) to make sure we got 'em all. */
  manyview_register_adjustment(manyview_matrix, control->strength);
  manyview_register_adjustment(manyview_matrix, control->bradj);
  manyview_register_adjustment(manyview_matrix, control->iterations);
  manyview_register_adjustment(manyview_matrix, control->wsize);
  manyview_register_adjustment(manyview_matrix, control->wsize_x);
  manyview_register_adjustment(manyview_matrix, control->wsize_y);
  manyview_register_adjustment(manyview_matrix, control->smoothing);
  manyview_register_adjustment(manyview_matrix, control->coefftol);
  manyview_register_adjustment(manyview_matrix, control->colormethod);

  manyview_look_alive(manyview_matrix);
}

static void 
control_register_preview(ControlBackend *control, Glace_ClientData *clientData)
{
  g_return_if_fail(control);
  control->bigview = clientData;
}

/* GtkSignalFunc for GtkAdjustment's "value_changed" */
static void 
control_updated_callback(GtkAdjustment* adjustment, 
			 ControlBackend* control)
{
  GtkAdjustment *partner_adj=NULL;
  gboolean linked=FALSE;
  ControlUpdateType update_type;
  gint old_revision=control->revision;
  gint old_out_revision=control->out_revision;

  g_return_if_fail(control);
  g_return_if_fail(GTK_IS_ADJUSTMENT(adjustment));
  
  update_type = 
    GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(adjustment),"update_type"));

  linked = 
    GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(adjustment),"linked"));

  partner_adj=gtk_object_get_data(GTK_OBJECT(adjustment),"partner");

  if(linked && partner_adj && GTK_IS_ADJUSTMENT(partner_adj))
    {
      gtk_signal_handler_block_by_func(GTK_OBJECT(adjustment),
				       control_updated_callback,
				       control);
      gtk_adjustment_set_value(partner_adj,adjustment->value);

      partner_adj=gtk_object_get_data(GTK_OBJECT(adjustment),"partner2");
      if(partner_adj && GTK_IS_ADJUSTMENT(partner_adj))
	{
	  gtk_adjustment_set_value(partner_adj,adjustment->value);
	}


      gtk_signal_handler_unblock_by_func(GTK_OBJECT(adjustment),
				       control_updated_callback,
				       control);
    }

  switch(update_type)
    {
    case UPDATE_RECALC:
      control->revision = ++old_revision;
      break;
    case UPDATE_OUTPUT:
      control->out_revision = ++old_out_revision;
      break;
    } /* esac update_type */

  if(control->auto_update)
    {
      if(control->manyview_matrix) 
	{
	  /* We can't run control_to_manyview from here, since we
	     don't want to be emitting value-changed signals from the
	     value-changed callback.  (The fact that the value-changed
	     signal is marked with GTK_RUN_NO_RECURSE has something to
	     do with it.)  So we'll add the control_to_manyview as an
	     idle event, which lets this one return before that one
	     starts... */

	  gtk_idle_add_priority(G_PRIORITY_HIGH_IDLE,
				(GSourceFunc)bounce_control_to_manyview,
				control);

	}
      
      if(control->bigview) 
	{
	  control_to_cfginfo(control, control->bigview->cfgInfoPtr);

	  ace_preview_smart_update(control->bigview, 
				   old_revision,
				   old_out_revision);
	}
    }
}


/* This function perhaps gets called more than it should, 
   but fortunately it is cheap and non-destructive. */
static void
control_to_cfginfo(const ControlBackend* const control, 
		   Glace_CfgInfo* const cfgInfoPtr)
{
  g_assert(control); 
  g_assert(cfgInfoPtr);
  g_assert(GTK_IS_ADJUSTMENT(control->strength));

  GLACE_CFG(heFactor) = 1.0 - control->strength->value;
  GLACE_CFG(passthruFactor) = 1.0 - control->bradj->value;
  GLACE_CFG(numTerms) = control->iterations->value;
  GLACE_CFG(windBaseW) = control->wsize_x->value;
  GLACE_CFG(windBaseH) = control->wsize_y->value;
  GLACE_CFG(coeffTol) = control->coefftol->value;
  GLACE_CFG(gaussCwindWidth) = control->smoothing->value;
  GLACE_CFG(colorMethod) = control->colormethod->value;
}

static void
control_to_acevals(const ControlBackend* const control, 
		   AceValues* const acevals)
{
  g_assert(control); 
  g_assert(acevals);
  g_assert(GTK_IS_ADJUSTMENT(control->strength));

  acevals->strength = control->strength->value;
  acevals->bradj = control->bradj->value;
  acevals->iterations = control->iterations->value;
  acevals->wsize = control->wsize->value;
  acevals->coefftol = control->coefftol->value;
  acevals->smoothing = control->smoothing->value;
  acevals->colormethod = control->colormethod->value;
  acevals->ui_auto = control->auto_update;
  acevals->ui_link = 
    GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(control->strength),"linked"));
}

static gboolean /* GSourceFunc */
bounce_control_to_manyview(ControlBackend* control)
{
  g_assert(control);

  control_to_manyview(control, control->manyview_matrix);

	  manyview_smart_update(control->manyview_matrix, 
				control->revision,
				control->out_revision);
  
  return FALSE;
}

/* Ok, so "control" isn't const, but its state is restored by the end of the
   function. */
static void 
control_to_manyview(ControlBackend* control, ManyviewMatrix* mv_matrix)
{
  gboolean auto_update_store=control->auto_update;
  guint revision_store=control->revision;
  guint out_revision_store=control->out_revision;

  int mv_yy,mv_xx,i;
  gfloat delta_x=0, delta_y=0;
  gfloat x,y;

  const int buflen=31;
  char buf[31];

  g_assert(control);
  g_return_if_fail(mv_matrix);
  g_return_if_fail(GTK_IS_ADJUSTMENT(mv_matrix->x_axis));
  g_return_if_fail(GTK_IS_ADJUSTMENT(mv_matrix->y_axis));

  control->auto_update=FALSE;

  delta_x = (mv_matrix->x_axis->page_size == 1) ? 1 : 
    mv_matrix->x_axis->page_increment;
  
  x = mv_matrix->x_axis->value;
  
  delta_y = (mv_matrix->y_axis->page_size == 1) ? 1 : 
    mv_matrix->y_axis->page_increment;
  
  y = mv_matrix->y_axis->value;
  
  for(mv_yy=0;mv_yy<3;mv_yy++)
    {
      /* g_print("%s setting adjustment '%s' to %f for [y%d]\n",
	      G_GNUC_PRETTY_FUNCTION, 
	      (char*)gtk_object_get_data(GTK_OBJECT(mv_matrix->y_axis),
					 "label_text"),
	      y+delta_y * (mv_yy - 1),
	      mv_yy); */
      /* Trust the adjustment to do the proper bounds enforcing. */
      gtk_adjustment_set_value(mv_matrix->y_axis,y + delta_y * (mv_yy - 1));

      mv_matrix->yvalues[mv_yy]=mv_matrix->y_axis->value;
      for(mv_xx=0;mv_xx<3;mv_xx++)
	{
	  /* g_print("%s setting adjustment '%s' to %f for [y%d][x%d]\n",
		  G_GNUC_PRETTY_FUNCTION, 
		  (char*)gtk_object_get_data(GTK_OBJECT(mv_matrix->x_axis),
				      "label_text"),
		  x+delta_x * (mv_xx - 1),
		  mv_yy, mv_xx); */

	  gtk_adjustment_set_value(mv_matrix->x_axis,
				   x + delta_x * (mv_xx - 1));

	  /* So this gets set three times instead of one.
	     "Oh well, it only costs a few cycles." */
	  mv_matrix->xvalues[mv_xx]=mv_matrix->x_axis->value;

	  control_to_cfginfo(control, 
			     mv_matrix->clients[mv_yy][mv_xx]->cfgInfoPtr);

	} /* next xx */
    } /* next yy */

  for(i=0;i<3;i++)
    {
      g_snprintf(buf,buflen,"%g",mv_matrix->xvalues[i]);
      gtk_label_set_text(GTK_LABEL(mv_matrix->xlabels[i]),buf);
      g_snprintf(buf,buflen,"%g",mv_matrix->yvalues[i]);
      gtk_label_set_text(GTK_LABEL(mv_matrix->ylabels[i]),buf);
    }

  /* restore controls */
  /* g_print("%s restoring controls:\n",G_GNUC_PRETTY_FUNCTION);

  g_print("%s restoring adjustment '%s' to %f\n",
		  G_GNUC_PRETTY_FUNCTION, 
		  (char*)gtk_object_get_data(GTK_OBJECT(mv_matrix->x_axis),
				      "label_text"),
				      x); */

  gtk_adjustment_set_value(mv_matrix->x_axis,x);

  /* g_print("%s restoring adjustment '%s' to %f\n",
		  G_GNUC_PRETTY_FUNCTION, 
		  (char*)gtk_object_get_data(GTK_OBJECT(mv_matrix->y_axis),
				      "label_text"),
				      y); */

  gtk_adjustment_set_value(mv_matrix->y_axis,y);

  control->auto_update=auto_update_store;
  control->revision=revision_store;
  control->out_revision=out_revision_store;
}

/* This function got a lot shorter since gimp_ui_init and gimp_dialog_new 
   were introducted to libgimpui.  Perhaps it should be absorbed back 
   into its caller.  */
static GtkWidget*
ace_dialog_new(GtkSignalFunc ok_callback, GtkSignalFunc close_callback)
{
  GtkWidget *dlg;
  gchar *wintitle;
  
#ifdef ACE_THREADS
  g_thread_init();
#endif
  gimp_ui_init("gimp_ace", TRUE);
  
  wintitle = g_strdup_printf(_("Adaptive Contrast Enhancement v %s"), 
			     VERSION);	
  
  dlg = gimp_dialog_new (wintitle,
			 "plug-in-ace",
			 gimp_standard_help_func, "gimp-ace.html",
			 GTK_WIN_POS_MOUSE,
			 FALSE, TRUE, FALSE,
			 NULL);
  g_free(wintitle);
  
  gimp_dialog_create_action_area(GTK_DIALOG(dlg), 
				 _("OK"),
				 ok_callback, dlg,
				 NULL, NULL, /* slot obj, widget ptr */
				 TRUE /* default */,
				 FALSE /* delete */,
				 _("Cancel"),
				 close_callback, NULL,
				 NULL, NULL,
				 FALSE, TRUE,
				 NULL);
  
  gimp_help_init ();

  return dlg;
}

/* The idea behind this function is it does some time-consuming set-up
 * operations, *after* there's been a chance for the dialog to show
 * itself.                                                             */
static gboolean dialog_setup(AutoUpdateData* update_data)
{ 
  static guint8 stage=1;

  switch(stage) 
    {
    case 1:
      update_data->clientData->wData->target=GLACE_GIMP_PREVIEW;
      ace_preview_fill_from_source(update_data->clientData);
      break;
    case 2:
      control_to_cfginfo(update_data->control, update_data->clientData->cfgInfoPtr);
      ace_preview_smart_update(update_data->clientData,
			       update_data->control->revision,
			       update_data->control->out_revision);
      break;
    case 3:
      manyview_smart_update(update_data->control->manyview_matrix,
			    update_data->control->revision,
			    update_data->control->out_revision);
    }
  stage++; 
  return (stage < 4);
}

/****  Preview-area callbacks. */

static void
update_preview_cb(GtkWidget* G_GNUC_UNUSED unused_widget,
		  AutoUpdateData* const update_data)
{
  control_to_cfginfo(update_data->control, 
		     update_data->clientData->cfgInfoPtr);

  ace_preview_update(update_data->clientData);

  if(update_data->control->manyview_matrix) 
    {
      control_to_manyview(update_data->control, 
			  update_data->control->manyview_matrix);
      
      manyview_smart_update(update_data->control->manyview_matrix, 
			    update_data->control->revision,
			    update_data->control->out_revision);
    }
}

static void
cancel_preview_cb(GtkWidget* G_GNUC_UNUSED unused_widget,
		  GlaceGimpPreview* const preview)
{
  ace_preview_cancel(preview);
}

static void
auto_toggle_cb (GtkWidget *toggle, AutoUpdateData* const update_data)
{
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle)))
    {
      control_to_cfginfo(update_data->control, 
			 update_data->clientData->cfgInfoPtr);
      ace_preview_smart_update(update_data->clientData, 
			       update_data->control->revision,
			       update_data->control->out_revision);
    }
}

/****** Parameter setting callbacks. *******/

static void
iter_changed(GtkAdjustment *adj, GtkProgressBar* pbar)
{
  static guint last_iter=0;

  if(last_iter < 25 && (guint)adj->value > 25)
    {
      gtk_progress_bar_set_bar_style(pbar, GTK_PROGRESS_CONTINUOUS);
    }
  else if(last_iter >= 25 && (guint)adj->value <= 25)
    {
      gtk_progress_bar_set_bar_style(pbar, GTK_PROGRESS_DISCRETE);
    }

  if((guint)adj->value <= 25) 
    {
      gtk_progress_bar_set_discrete_blocks(GTK_PROGRESS_BAR(pbar),
					   (guint)adj->value+PROG_OCHUNKS);
    }
  last_iter = (guint)adj->value;
}

static void
omenu_callback (GtkWidget *menuitem, GtkAdjustment* adj)
{
  g_assert(GTK_IS_ADJUSTMENT(adj));
  gtk_adjustment_set_value(adj, 
			   GPOINTER_TO_INT(gtk_object_get_user_data(GTK_OBJECT(menuitem))));
}

/***********************************
 * Generic action-area callbacks.  */

static void 
ok_callback  (GtkWidget G_GNUC_UNUSED *unused_widget, 
	      GtkWidget* dlg)
{
  AceValues* acevals=NULL;
  ControlBackend* control=NULL;
  Glace_ClientData* clientData=NULL;

  g_assert(GTK_IS_DIALOG(dlg));

  acevals = gtk_object_get_data(GTK_OBJECT(dlg),"acevals");
  g_assert(acevals);

  control = gtk_object_get_data(GTK_OBJECT(dlg),"control_backend");
  g_assert(control);

  clientData = gtk_object_get_data(GTK_OBJECT(dlg),"clientData");
  g_assert(clientData);

  control_to_cfginfo(control, clientData->cfgInfoPtr);
  control_to_acevals(control, acevals);

  gtk_widget_destroy(dlg);
  gtk_main_quit();
}

static void 
close_callback  (GtkWidget G_GNUC_UNUSED *unused_widget, 
		 gpointer G_GNUC_UNUSED unused_data)
{
  *gimp_ace_status=GIMP_PDB_CANCEL;
  gimp_quit();
}

static void
link_toggled (GimpChainButton *chainbutton,
	     GtkObject* obj)
{
  gboolean active;

  g_assert(GTK_IS_OBJECT(obj));

  active=gimp_chain_button_get_active(chainbutton);

  gtk_object_set_data(obj, "linked", GINT_TO_POINTER(active));

  /* Make the link go both ways. */
  obj = GTK_OBJECT(gtk_object_get_data(obj, "partner"));
  
  if(GTK_IS_OBJECT(obj))
    {
      gtk_object_set_data(obj, "linked", GINT_TO_POINTER(active));
    }
}

/* For emacs:
Local Variables:
mode:c
c-file-style:"gnu"
End:
*/
