/* $Id: gimp_ace.c,v 1.28 2000/09/14 06:07:16 acapnotic Exp $ */
/* Adaptive Contrast Enhancement plug-in for the GIMP (v1.2)
 * (A windowed histogram equalization using the Fourier-series based
 * fast algorithm.)
 * Based on J. Alex Stark's <stark@niss.org> GLACE system. 
 * GIMP Plug-in by Kevin Turner <kevint@poboxes.com> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.  See the file COPYING for details.
 *
 */

/* :HEADERS: 
 */

#include <config.h>		/* autostuff */

#ifdef DIAGNOSTIC
# include <sys/types.h> /* SIGSTOP */
# include <signal.h>    /* kill() */
# include <unistd.h>    /* getpid() */
#endif

#include <libgimp/gimpintl.h>

#include "preview.h"
#include "gimp_ace.h"		/* defines the AceValues struct */
#include "dialog.h"             /* because we call the dialog. */
#include "color.h"		/* This has the RGB -> Y conversion factors.
				   for the SeperateChannels */
#include "output.h"             /* Process routine calls Output */

/* :*** Prototypes ***: 
 */

static void query (void);
static void run (char*    name,
		 gint     nparams,
		 GimpParam*  param, 
                 gint*    nreturn_vals, 
                 GimpParam** return_vals);

static GlaceGimpDrawableInfo* drawable_info_new(GimpDrawableID id);

static gboolean SeperateChannels (/* in: */ 
				  guchar inbuf[], 
				  gint buflen,
				  gboolean TypeRGB, 
				  gboolean HasAlpha,
				  /* out: */ 
				  guchar** Graybuf, 
				  guint* outbuflen);

static gboolean GlaceGimp_ProcessIter(Glace_ClientData* const clientData);

static void acevals_to_cfginfo(const AceValues* const acevalsPtr, 
			       Glace_CfgInfo* const cfgInfoPtr);


GimpPDBStatusType* gimp_ace_status;

#ifdef DIAGNOSTIC
#if 0
static void fill_diagnostic(GtkPreview* const preview, 
			    const PreviewPixel* const data,
			    const guint width, 
			    const guint height);
#endif
#endif

#define IS_WITHIN(X,L,H) (((X) >= L) && ((X) <= H))

/* :*** Definitions ***:
 */

GimpPlugInInfo PLUG_IN_INFO = {
  NULL,				/* init_proc */
  NULL,				/* quit_proc */
  query,			/* query_proc */
  run,				/* run_proc */
};

static AceValues acevals = {
  1.0,				/* Strength (full) */
  0.0,				/* brightness adjust (none) */
  0.0,				/* foobarm (none) */
  0.0,				/* smoothing (none) */
  10,				/* iterations */
  25,				/* window size */
  GLACE_COLOR_DESATURATE,	/* Color method */
  FALSE,			/* link strength and brightness adj in dlg. */
  TRUE,                         /* Automatically update preview */
  FALSE                         /* using the _cli interface. */
};

static GimpParamDef ace_pdb_params[] = {
  /* Required params. */
  {GIMP_PDB_INT32, "run_mode",       "GimpRunModeType"},
  {GIMP_PDB_IMAGE, "image_id",       "(unused)"},
  {GIMP_PDB_DRAWABLE, "drawable_id", "Drawable to affect"},
  /* Our params. */
  {GIMP_PDB_FLOAT, "strength",   "[0,1] Magnitude of effect"},
  {GIMP_PDB_FLOAT, "bradj",      "[0,1] Magnitude of brightness adjust"},
  {GIMP_PDB_INT32, "iterations", "[1,] No. of iterations"},
  {GIMP_PDB_FLOAT, "fthreshold", "[FIXME;0 if unused] FOOBARUM threshold"},
  {GIMP_PDB_INT32, "size",       "[FIXME,] Detail size"},
  {GIMP_PDB_FLOAT, "smoothing",  "[0,] Smoothing"},
  {GIMP_PDB_INT32, "colormethod",    "0=GLACE_COLOR_Yxy, 1=GLACE_COLOR_LUMIN"}
};
static const int ace_pdb_nparams = sizeof(ace_pdb_params) / sizeof(GimpParamDef);

static GimpParamDef ace_pdb_params2[] = {
  /* Required params. */
  {GIMP_PDB_INT32, "run_mode",       "GimpRunModeType"},
  {GIMP_PDB_IMAGE, "image_id",       "(unused)"},
  {GIMP_PDB_DRAWABLE, "drawable_id", "Drawable to affect"},
  /* Our params. */
  {GIMP_PDB_STRING, "cmdline",   "A GLACE command-line."}
};

static const int ace_pdb_nparams2 = sizeof(ace_pdb_params2) / sizeof(GimpParamDef);

MAIN ()

static void 
query (void)
{
  const int ace_pdb_nretvals=0;
  GimpParamDef* ace_pdb_retvals=NULL;

  gimp_plugin_domain_register (PLUGIN_NAME, LOCALEDIR);

  gimp_plugin_help_register (DATADIR"/help");

  gimp_install_procedure ("plug_in_ace",
			  _("Adaptive Contrast Enhancement"),
			  _("FIXME: Write ACE help."),
			  "Alex Stark, Kevin Turner,",
			  "Alex Stark, Kevin Turner,",
			  "1998-2000",
			  "<Image>/Image/Colors/Adaptive Contrast...", 
			  "GRAY*, RGB*", 
			  GIMP_PLUGIN, 
			  ace_pdb_nparams, 
			  ace_pdb_nretvals,
			  ace_pdb_params, 
			  ace_pdb_retvals);

  gimp_install_procedure ("plug_in_ace_cli",
			  "Adaptive Contrast Enhancement, command line interface",
			  "Provide a pnmace-style interface for scripting use.",
			  "Alex Stark, Kevin Turner,",
			  "Alex Stark, Kevin Turner,",
			  "1998-2000",
			  NULL,
			  "GRAY*, RGB*", 
			  GIMP_PLUGIN, 
			  ace_pdb_nparams2,
			  ace_pdb_nretvals,
			  ace_pdb_params2,
			  ace_pdb_retvals);
}

static void
run (gchar* name, 
     gint  nparams,      GimpParam*  param,
     gint* nreturn_vals, GimpParam** return_vals)
{
  typedef enum { PARAM_INTERFACE, CMDLINE_INTERFACE } GlacePDBInterface;
  GlacePDBInterface interface;
  static GimpParam retvals[1];
  const GimpRunModeType run_mode = param[0].data.d_int32;
  const GimpDrawableID drawable_id = param[2].data.d_drawable;
  GlaceGimpDrawableInfo* const drawable_info = drawable_info_new(drawable_id);
  Glace_ClientData *clientData=NULL;
  char* cmdline=NULL;
  int i;

  *nreturn_vals = 1;
  retvals[0].type = GIMP_PDB_STATUS;  
  gimp_ace_status = (GimpPDBStatusType*) &retvals[0].data.d_status;
  *return_vals = retvals;

  interface = strcmp(name,"plug_in_ace_cli") == 0 ? 
    CMDLINE_INTERFACE : PARAM_INTERFACE;

#ifdef DIAGNOSTIC
#if 0
  g_message ("gimp_ace PID: %d\n", getpid ());
  kill (getpid (), SIGSTOP);
#endif
#endif

  /* Sanity checking on drawable. */
  if(MIN(drawable_info->w,drawable_info->h)-1 < 4) {
    g_message(_("Selection must be\n"
		"  at least 5 by 5 pixels.\n"
		"This one is too small."));
    *gimp_ace_status = GIMP_PDB_CALLING_ERROR;
    return;
  }

  switch (run_mode)
    {
    case GIMP_RUN_NONINTERACTIVE:
      *gimp_ace_status = GIMP_PDB_CALLING_ERROR;
    
      switch (interface)
	{
	case PARAM_INTERFACE:	  
	  /* Check number of parameters. */
	  if(nparams != ace_pdb_nparams) 
	    {
	      g_warning("nparams is %d, should be %d " 
			"in run() with RUN_NONINTERACTIVE\n",
			nparams, ace_pdb_nparams);
	      return;
	    }
	  
	  /* Check parameter types. */
	  for(i=0;i<ace_pdb_nparams;i++) 
	    {
	      if(param[i].type != ace_pdb_params[i].type) 
		{
		  g_warning("parameter %d has type %d, should be %d "
			    "in run() with RUN_NONINTERACTIVE\n",
			    i, param[i].type,ace_pdb_params[i].type);
		  return;
		}
	    }
	  
	  /* Bounds check parameters. */
	  i=3; 
	  if(!(IS_WITHIN(param[i].data.d_float,0,1)))
	    {
	      g_warning("parameter %d (strength) has value %f, "
			"should be in [0,1] "
			"in run() with RUN_NONINTERACTIVE\n",
			i, param[3].data.d_float);
	      return;
	    }
	  i++;
	  if(!(IS_WITHIN(param[i].data.d_float,0,1)))
	    {
	      g_warning("parameter %d (bradj) has value %f, "
			"should be in [0,1] "
			"in run() with RUN_NONINTERACTIVE\n",
			i, param[3].data.d_float);
	      return;
	    }
	  i++;
	  if(!(param[i].data.d_int32 >= 1))
	    {
	      g_warning("parameter %d (iterations) has value %d, "
			"should be in >= 1"
			"in run() with RUN_NONINTERACTIVE\n",
			i, param[3].data.d_int32);
	      return;
	    }
	  i++;
	  if(!(param[i].data.d_float >= 0))
	    {
	      g_warning("parameter %d (fthreshold) has value %f, "
			"should be >= 0 "
			"in run() with RUN_NONINTERACTIVE\n",
			i, param[3].data.d_float);
	      return;
	    }
	  i++;
	  if(!(IS_WITHIN(param[i].data.d_int32,
			 4,MIN(drawable_info->w,drawable_info->h))))
	    {
	      g_warning("parameter %d (size) has value %d, "
			"should be between 4 and smallest selection dimension."
			"in run() with RUN_NONINTERACTIVE\n",
			i, param[3].data.d_int32);
	      return;
	    }
	  i++;
	  if(!(param[i].data.d_float >= 0))
	    {
	      g_warning("parameter %d (smoothing) has value %f, "
			"should be >= 0 "
			"in run() with RUN_NONINTERACTIVE\n",
			i, param[3].data.d_float);
	      return;
	    }
	  if(!(IS_WITHIN(param[i].data.d_int32,0,GLACE_COLOR_LAST-1)))
	    {
	      g_warning("parameter %d (colormethod) has value %d, "
			"should be in [0,%d] "
			"in run() with RUN_NONINTERACTIVE\n",
			i, param[3].data.d_int32, GLACE_COLOR_LAST-1);
	      return;
	    }
	  
	  /* Assign parameters */
	  i=3;
	  acevals.strength     = param[i++].data.d_float;
	  acevals.bradj        = param[i++].data.d_float;
	  acevals.iterations   = param[i++].data.d_int32;
	  acevals.coefftol     = param[i++].data.d_float;
	  acevals.wsize        = param[i++].data.d_int32;
	  acevals.smoothing    = param[i++].data.d_float;
	  acevals.colormethod = param[i++].data.d_int32;
	  break;
	case CMDLINE_INTERFACE:
	  if(nparams != ace_pdb_nparams2)
	    {
	      g_warning("nparams is %d, should be %d " 
			"in run() with RUN_NONINTERACTIVE\n",
			nparams, ace_pdb_nparams2);
	      return;
	    }
	  
	  /* Check parameter types. */
	  for(i=0;i<ace_pdb_nparams2;i++) 
	    {
	      if(param[i].type != ace_pdb_params2[i].type) 
		{
		  g_warning("parameter %d has type %d, should be %d "
			    "in run() with RUN_NONINTERACTIVE\n",
			    i, param[i].type,ace_pdb_params2[i].type);
		  return;
		}
	    }

	  cmdline = g_strdup(g_strstrip(param[i++].data.d_string));
	  acevals.using_cmdline=TRUE;
	} /* esac interface */

      /* status is no longer a calling error. */
      *gimp_ace_status = GIMP_PDB_EXECUTION_ERROR;      
      break;
    case GIMP_RUN_WITH_LAST_VALS:
      *gimp_ace_status = GIMP_PDB_EXECUTION_ERROR;
      gimp_get_data ("plug_in_ace", &acevals);
      break;
    case GIMP_RUN_INTERACTIVE: /* NOTE: This returns when done. */
      *gimp_ace_status = GIMP_PDB_EXECUTION_ERROR;
      gimp_get_data ("plug_in_ace", &acevals);

      ace_dialog(&acevals, drawable_info);

      if(*gimp_ace_status != GIMP_PDB_CANCEL) 
	{
	  *gimp_ace_status = GIMP_PDB_SUCCESS;
	  gimp_set_data ("plug_in_ace", &acevals, sizeof (AceValues));	  
	  gimp_displays_flush();
	}
      return;
    }

  /* Only non-interactive modes have made it here. */

  if(*gimp_ace_status != GIMP_PDB_CANCEL)
    {
      clientData = clientdata_new(drawable_info, cmdline);
      if(!cmdline) acevals_to_cfginfo(&acevals, clientData->cfgInfoPtr);
      clientData->wData->target=GLACE_GIMP_PixelRgn;
      gimp_progress_init (_("Adaptive Contrast Enhancement:"));
      gimp_ace_go(clientData, G_PRIORITY_DEFAULT_IDLE);

      if(run_mode == GIMP_RUN_WITH_LAST_VALS)
	gimp_displays_flush();
      /* If we made it through this without calling gimp_quit(),
	 then we have been sucessful. */
      *gimp_ace_status = GIMP_PDB_SUCCESS;
    }
}

static GlaceGimpDrawableInfo*
drawable_info_new(GimpDrawableID id)
{
  GlaceGimpDrawableInfo* const self = g_new0(GlaceGimpDrawableInfo,1);
  gint x1=0, y1=0, x2=0, y2=0;
  GimpPixelRgn src_rgn;

  self->drawable = gimp_drawable_get(id);

  gimp_drawable_mask_bounds (id, &x1, &y1, &x2, &y2);
  
  self->x0 = x1;
  self->y0 = y1;
  self->w = x2-x1;
  self->h = y2-y1;

  self->has_alpha = gimp_drawable_has_alpha(id);

  self->gimpImgSize = self->w * self->h * self->drawable->bpp;

  self->gimpImgPtr = g_new(GimpPixel, self->gimpImgSize);

  gimp_pixel_rgn_init (&src_rgn, self->drawable,
		       self->x0, self->y0,
		       self->w, self->h,
		       FALSE, FALSE);

  gimp_pixel_rgn_get_rect (&src_rgn, self->gimpImgPtr,
			   self->x0, self->y0,
			   self->w, self->h);

  self->gimpImgRowstride = self->w * self->drawable->bpp;

  /* Extract a channel for processing. */

  SeperateChannels (self->gimpImgPtr, self->gimpImgSize,
		    gimp_drawable_is_rgb (id), 
		    self->has_alpha, 
		    /* out: */ 
		    &self->inImgPtr, &self->inImgSize);

  return self;
}

Glace_ClientData*
clientdata_new(const GlaceGimpDrawableInfo* const drawable_info, 
	       char* const cmdline)
{
  Glace_ClientData *clientData = Glace_AllocClientData ();
  Glace_ImgArrays *imgArraysPtr = clientData->imgArraysPtr;
  Glace_CfgInfo *cfgInfoPtr = clientData->cfgInfoPtr;

  g_assert(drawable_info);

  Glace_CfgInit (GLACE_CDATA (cfgInfoPtr));

  if(cmdline != NULL)
    {
      gboolean old_in_word=FALSE, in_word=FALSE;
      int argc=0, len=0;
      char* ic=cmdline;

      /* Pass 1: count arguments */
      while(*ic != '\0')
	{
	  old_in_word = in_word;
	  in_word = (*ic != ' '  &&
		     *ic != '\n' &&
		     *ic != '\t');

	  if(in_word == FALSE && old_in_word == TRUE)
	    {
	      /* Just exited a word. */
	      argc++;
	    }
	  ic++; len++;
	} /* whend *ic != '\0' */
      if(in_word == TRUE) argc++;
      
      /* Pass 2: set argv, null-terminate arguments. */
      if(argc > 0)
	{
	  char** argv=g_new(char*,argc);
	  int i, argc_i=0;

	  in_word=FALSE;
	  for(i=0, ic=cmdline; i < len; i++, ic++)
	    {
	      old_in_word = in_word;
	      in_word = (*ic != ' '  &&
			 *ic != '\n' &&
			 *ic != '\t');

	      if(in_word == FALSE && old_in_word == TRUE)
		{
		  *ic = '\0';
		  /* Just exited a word. */
		  argc_i++;
		  if(argc_i > argc)
		    g_error("Argument count mismatch.  Please notify "
			    "MAINTAINER_EMAIL" "about the string:\n%s\n",
			    cmdline);
		}	 
	      else if(old_in_word==FALSE && in_word == TRUE)
		{
		  argv[argc_i] = ic;
		}
	    } /* next char */

	  Glace_ParseArgs (GLACE_CDATA (cfgInfoPtr),
			   GLACE_CDATA (imgArraysPtr), argc, argv);
	  g_free(argv);
	} /* endif argc > 0 */
    }
  else /* No cmdline, use default. */
    {
      gchar *ace_argv[] = { 
	"-reset",
	"25",
	"25",
	"10",
	"-factor",
	"0.0",			/* 1.0 is identity, 0.0 is full effect. */
	"-prescale",
	"1.0",
	"-gauss",
	"0",
	"-hamming",
	"-gg",
	"-passthru",
	"1.0",
	"-inoffset",
	"127.5",
	"-tol",
	"0.0"
      };
      gint ace_argc = 18;

      Glace_ParseArgs (GLACE_CDATA (cfgInfoPtr),
		       GLACE_CDATA (imgArraysPtr), ace_argc, ace_argv);
    }

  clientData->wData->d = drawable_info;

  clientData->wData->iteration = 0;

  Glace_DefaultTmpImg (GLACE_CDATA (cfgInfoPtr), GLACE_CDATA (imgArraysPtr));

  /* skip some pixels on output so we can mix the alpha back in. */
  GLACE_IMG (pixelBytePad) = drawable_info->has_alpha ? 1 : 0;

  /* This refers to precision, not number of channels. */
  GLACE_IMG (pixelSize) = 1;

  GLACE_CFG (chrome) = gimp_drawable_is_rgb (drawable_info->drawable->id)
    ? GLACE_CC : GLACE_GG;

  GLACE_IMG(inImgSize) = drawable_info->inImgSize;

  /* From here down is the stuff to change when windowX0,Y0 != 0,0 */
  GLACE_IMG(windowX0) = GLACE_IMG(viewX0) = 0;
  GLACE_IMG(windowY0) = GLACE_IMG(viewY0) = 0;
  GLACE_IMG(cols) = GLACE_IMG(viewCols) = drawable_info->w;
  GLACE_IMG(rows) = GLACE_IMG(viewRows) = drawable_info->h;

  GLACE_IMG(rowstride) = drawable_info->w; 

  GLACE_IMG(inImgPtr) = drawable_info->inImgPtr;

  /* actually, all processing is done from HBPtr */
  GLACE_IMG(refImgHBPtr) = GLACE_IMG(inImgHBPtr) = GLACE_IMG(inImgPtr);

  return clientData;
}

void
gimp_ace_go(Glace_ClientData* clientData, int priority)
{
  Glace_CfgInfo* const cfgInfoPtr= clientData->cfgInfoPtr;
  GlaceGimpData* const wData= clientData->wData;
  Glace_ImgArrays* const imgArraysPtr = clientData->imgArraysPtr;

  g_assert(clientData);

  /* Sanity checks. */
  if(GLACE_CFG(windBaseW) > GLACE_IMG(cols) - 1) {
    g_warning("window width too wide at %d, "
	      "resetting to %d\n", 
	      GLACE_CFG(windBaseW), GLACE_IMG(cols) - 1);
    GLACE_CFG(windBaseW) = GLACE_IMG(cols) - 1;
  }

  if(GLACE_CFG(windBaseH) > GLACE_IMG(rows) - 1) {
    g_warning("window height too tall at %d, "
	      "resetting to %d\n", 
	      GLACE_CFG(windBaseH), GLACE_IMG(rows) - 1);
    GLACE_CFG(windBaseH) = GLACE_IMG(rows) - 1;
  }

  switch(WRAPPER(target))
    {
    case GLACE_GIMP_PixelRgn:
      clientData->wData->iteration = 0;
      while(GlaceGimp_ProcessIter (clientData));
      break;
    case GLACE_GIMP_PREVIEW:
      clientData->wData->iteration = -2;
      clientData->wData->preview->idle_handle = 
	gtk_idle_add_priority(priority,
			      (GtkFunction)GlaceGimp_ProcessIter, 
			      clientData);
      break;
    } /* esac target */
}

/* Returns FALSE if Graybuf is inbuf, otherwise TRUE */
static gboolean
SeperateChannels (/* in: */ guchar inbuf[], const gint buflen,
		  const gboolean TypeRGB, const gboolean HasAlpha,
		  /* out: */ guchar** Graybuf, guint* outbuflen)
{
  int c = 0, ic = 0;
  guchar *Grayfoo;

  if(!TypeRGB)
    {
      if(!HasAlpha) /* if GRAY */
	{
	  *Graybuf = inbuf;
	  *outbuflen = buflen;
	  return FALSE;
	}
      else /* if GRAYA */
	{
	  if ((buflen % 2) != 0) 
	    {
	      g_warning ("GRAYA with odd size %d?", buflen);
	      *gimp_ace_status = GIMP_PDB_CALLING_ERROR;
	      gimp_quit();
	    }

	  if (*Graybuf == NULL)
	    *Graybuf = g_new (guchar, buflen / 2);

	  Grayfoo = *Graybuf;
	  for(ic=0,c=0;ic<buflen;ic+=2,c++)
	    {
	      Grayfoo[c] = inbuf[ic];
	    }
	  *outbuflen = buflen/2;
	  return TRUE; 
	} /* endif GRAYA */
    } /* endif GRAY* */ 
  else
    { /* RGB */
      const int bpp = HasAlpha ? 4 : 3;
      if ((buflen % bpp) != 0)
	{
	  g_warning("RGB: %d Alpha: %d.  I think bpp should be %d, but that"
		    "doesn't make sense with size %d, as R = %d",
		    TypeRGB, HasAlpha, bpp, buflen, buflen % bpp);
	  *gimp_ace_status = GIMP_PDB_CALLING_ERROR;
	  gimp_quit();
	}
      
      if (*Graybuf == NULL)
	*Graybuf = g_new (guchar, buflen / bpp);

      Grayfoo = *Graybuf;
      
      for(ic=0,c=0;ic<buflen;ic+=bpp,c++)
	{
	  Grayfoo[c] = 
	    Y_r * inbuf[ic+0] + 
	    Y_g * inbuf[ic+1] + 
	    Y_b * inbuf[ic+2];
	}			/* whend */
      *outbuflen = buflen / bpp;
      return TRUE;
    } /* endif rgb */
} /* SeperateChannels() */


/* Iterates through one pass of the Glace_Process loop.
 * Returns FALSE when done.
 * Suitable for use as a callback on an idle handler. */
static gboolean
GlaceGimp_ProcessIter(Glace_ClientData* const clientData)
{
  Glace_WData const wData = clientData->wData;
  Glace_CfgInfo* const cfgInfoPtr = clientData->cfgInfoPtr;
  Glace_ImgArrays* const imgArraysPtr = clientData->imgArraysPtr;
  Glace_TableInfo* const tableInfoPtr = clientData->tableInfoPtr;
  GlaceGimpPreview* const preview = wData->preview;

  if(wData->iteration == -2)
    {
      if(preview->timeout_timer == NULL)
	{
	  preview->timeout_timer = g_timer_new();
	}
      g_timer_start(preview->timeout_timer);
      wData->iteration++;
      return TRUE;
    }

  if(wData->iteration < 0)
    {
      gdouble seconds;
      seconds = g_timer_elapsed(preview->timeout_timer,NULL);
      if(seconds >= DELAY_UPDATE_SECONDS)
	{
	  g_timer_stop(preview->timeout_timer);
	  wData->iteration = 0;
	}
      else /* Wait longer. */
	{
	  return TRUE;
	}
    }

  if(wData->iteration == 0)
    {
      Glace_CfgAllocSeriesVectors (GLACE_CDATA (cfgInfoPtr));
      
      Glace_CfgBeginToHeseries (GLACE_CDATA (cfgInfoPtr));
      
      Glace_CfgHeseriesToAddback (GLACE_CDATA (cfgInfoPtr));
      
      Glace_CFuncGen (GLACE_CDATA (cfgInfoPtr));
      
      Glace_CfgAddbackToEnd (GLACE_CDATA (cfgInfoPtr));
      
      Glace_AllocTables (GLACE_CDATA (cfgInfoPtr), 
			 GLACE_CDATA (tableInfoPtr));
      
      Glace_AllocImgArrays (GLACE_CDATA (cfgInfoPtr), 
			    GLACE_CDATA (imgArraysPtr));

      /* try: */
      Glace_WindChk (cfgInfoPtr, GLACE_IMG (cols), GLACE_IMG (rows));

      if(Glace_WIsError(wData))
	{
	  /* WindChk failed -- abort now. */
	  return FALSE;
	}

      Glace_InitAccIm (cfgInfoPtr, imgArraysPtr);
      
      wData->iteration = GLACE_CFG(firstTerm);
    }

  if(wData->iteration <= GLACE_CFG (numTerms))
    {
      Glace_SetForTerm (tableInfoPtr, wData->iteration, GLACE_COS);

#ifdef DIAGNOSTIC
      Glace_TermlyReport (cfgInfoPtr, tableInfoPtr);
      if(GLACE_CFG(verbose)) Glace_WMessage (wData, ";\n");
#endif
      
      Glace_SetTables (cfgInfoPtr, tableInfoPtr);
      Glace_GenGen (cfgInfoPtr, tableInfoPtr, imgArraysPtr);
      
      Glace_SetForTerm (tableInfoPtr, wData->iteration, GLACE_SIN);
      
      Glace_SetTables (cfgInfoPtr, tableInfoPtr);
      Glace_GenGen (cfgInfoPtr, tableInfoPtr, imgArraysPtr);
            
      /* FIXME: Wrapper's _progress_update should handle both cases. */
      switch(WRAPPER(target))
	{
	case GLACE_GIMP_PixelRgn:
	  gimp_progress_update (0.10 + 0.90 * wData->iteration / 
				GLACE_CFG (numTerms));
	  break;
	case GLACE_GIMP_PREVIEW:
	  ace_preview_progress_update(preview, (gfloat) 
				      (PROG_INPUT_CHUNKS + wData->iteration) / 
				      (GLACE_CFG(numTerms) + PROG_OCHUNKS));
	  break;
	} /* esac target */
      wData->iteration++;
      return TRUE;
    }
  else /* Last iteration is done. */
    {
      GlaceGimp_Output (wData, cfgInfoPtr, imgArraysPtr);  
      switch(WRAPPER(target))
	{
	case GLACE_GIMP_PixelRgn:
	  break;
	case GLACE_GIMP_PREVIEW:
	  ace_preview_update_finished(preview);
	  break;
	} /* esac target */

      wData->iteration=0;
      return FALSE;
    } /* endif more iterations */
}

static void
acevals_to_cfginfo(const AceValues* const acevalsPtr, 
		   Glace_CfgInfo* const cfgInfoPtr)
{
  g_assert(acevalsPtr);
  g_assert(cfgInfoPtr);

  cfgInfoPtr->heFactor        = 1.0 - acevalsPtr->strength;
  cfgInfoPtr->passthruFactor  = 1.0 - acevalsPtr->bradj;
  cfgInfoPtr->numTerms        = acevalsPtr->iterations;
  cfgInfoPtr->windBaseW       = acevalsPtr->wsize;
  cfgInfoPtr->windBaseH       = acevalsPtr->wsize;
  cfgInfoPtr->coeffTol        = acevalsPtr->coefftol;
  cfgInfoPtr->gaussCwindWidth = acevalsPtr->smoothing;
  cfgInfoPtr->colorMethod     = acevalsPtr->colormethod;
}

#ifdef DIAGNOSTIC
#if 0
static void 
fill_diagnostic(GtkPreview* const preview, const PreviewPixel* data,
		const guint width, const guint height)
{
  guint yy;
  for(yy=0; yy < height; yy++) {
    gtk_preview_draw_row(preview,data,0,yy,width);
    data+=width;
  }
}
#endif
#endif

/* For emacs:
Local Variables:
mode:c
c-file-style:"gnu"
End:
*/
