/* $Id: preview.c,v 1.15 2000/09/15 02:26:17 acapnotic Exp $ */

#include "types.h"
#include "gimp_ace.h"
#include "preview.h"

static gint ace_preview_events(GtkWidget*   widget, 
			       GdkEvent*    event, 
			       GlaceGimpPreview* preview);

#if 0
static GtkWidget* gtk_arrowbutton_new(int horizontal, int vertical);
#endif

static GtkWidget* gtk_3dotbutton_new (int horizontal, int vertical);

static Glace_ClientData* manyview_new(const GlaceGimpDrawableInfo* drawable_info,
				      ManyviewMatrix* mv_matrix,
				      int priority);

static GtkBox* manyview_axis_box_new(ManyviewMatrix* mv_matrix, 
				     char x_or_y, 
				     GtkMenu** menu, 
				     GtkAdjustment **delta_adj);

static GtkWidget* viewfinder_new(const GlaceGimpDrawableInfo* drawable_info,
				 ManyviewMatrix* mv_matrix);


/* callback for GtkButton "clicked" signal */
static void manyview_axis_menu_callback(GtkMenuItem* omenu, 
					ManyviewMatrix* mv_matrix);

/* callback for GtkAdjustment "value-changed" signal */
static void manyview_axis_delta_changed(GtkAdjustment* delta_adj,
					GtkAdjustment** axis_adj);


/* callback for GtkButton "clicked" signal */
static void manyview_navbutton_clicked(GtkButton* button, 
				       ManyviewMatrix* mv_matrix);

/* callback for GtkWidget "realize" signal */
static void viewfinder_realize        (GtkWidget* widget, 
			               gpointer unused_data);

/* callback for GtkAdjustment "changed" and "value-changed" signals */
static void view_changed(GtkAdjustment* adjustment, 
			 ManyviewMatrix* mv_matrix);

static void viewfinder_draw(GtkWidget* preview, 
			    GdkRectangle* unused,
			    ManyviewMatrix* mv_matrix);

static void viewfinder_draw_rectangle(GtkWidget* preview, 
				      int x, int y, int w, int h,
				      gboolean erase_old);

static gint viewfinder_events(GtkWidget*   widget, 
			      GdkEvent*    event, 
			      ManyviewMatrix* mv_matrix);

static void manyview_size_allocate(GtkWidget* G_GNUC_UNUSED unused_preview, 
				   GtkAllocation* allocation, 
				   ManyviewMatrix* mv_matrix);


static char* dot_xpm[] = {
"4 4 4 1",
"       c None",
".      c #C9C9D4",
"+      c #B8B8C5",
"@      c #9595A7",
".++.",
"+@@+",
"+@@+",
".++."};

#define SPACING 0 /* FIXME */

/* Returns a pointer to a new GlaceGimpPreview.
 * If non-NULL, the provided preview and progress bar widgets will be used.
 * If NULL, new ones will be created. */
GlaceGimpPreview*
ace_preview_new(GlaceGimpData* const wData)
{
  GlaceGimpPreview* const self = g_new0(GlaceGimpPreview, 1);
  const GimpDrawableID drawable_id = wData->d->drawable->id;
  const gboolean is_colour = gimp_drawable_is_rgb(drawable_id);
  GtkPreview* preview;
  GtkProgressBar* pbar;
  GtkScrolledWindow* scrollwin;
  GtkWidget *viewport;

  scrollwin= GTK_SCROLLED_WINDOW(gtk_scrolled_window_new(NULL,NULL));
  gtk_scrolled_window_set_policy(scrollwin, GTK_POLICY_AUTOMATIC, 
				 GTK_POLICY_AUTOMATIC);

  preview = GTK_PREVIEW(gtk_preview_new(is_colour ? GTK_PREVIEW_COLOR : 
					GTK_PREVIEW_GRAYSCALE));

  gtk_preview_size(preview, wData->d->w, wData->d->h);
  gtk_preview_set_expand(preview,FALSE);

  viewport = gtk_viewport_new(NULL,NULL);
  gtk_container_add(GTK_CONTAINER(viewport),GTK_WIDGET(preview));
  gtk_container_add(GTK_CONTAINER(scrollwin),viewport);

  gtk_scrolled_window_set_hadjustment(scrollwin, 
		        gtk_viewport_get_hadjustment(GTK_VIEWPORT(viewport)));
  gtk_scrolled_window_set_vadjustment(scrollwin,
			gtk_viewport_get_vadjustment(GTK_VIEWPORT(viewport)));

  gtk_widget_set_usize(GTK_WIDGET(scrollwin),
		       MIN(PREVIEW_WIDTH, wData->d->w),
		       MIN(PREVIEW_HEIGHT, wData->d->h));

  gtk_widget_set_events(GTK_WIDGET(viewport), 
			GDK_BUTTON_PRESS_MASK |
			GDK_BUTTON_RELEASE_MASK | 
			GDK_BUTTON_MOTION_MASK |
			GDK_POINTER_MOTION_HINT_MASK);
  gtk_signal_connect(GTK_OBJECT(viewport), "event",
		     GTK_SIGNAL_FUNC(ace_preview_events),
		     self);

  pbar = GTK_PROGRESS_BAR(gtk_progress_bar_new());
  gtk_progress_bar_set_bar_style(pbar, GTK_PROGRESS_DISCRETE);

  self->scrollwin = scrollwin;
  self->preview = preview;
  self->pbar = GTK_PROGRESS(pbar);
  self->bpp = is_colour ? 3 : 1;
  self->idle_handle = 0;
  self->idle_priority = 0;
  self->state = PREVIEW_IDLE;

  return self;
}

/* Fills a preview widget with the unprocessed source image.  Used for quickly
   getting something to fill the preview screen. */
void 
ace_preview_fill_from_source(const Glace_ClientData* const clientData)
{
  const GlaceGimpData*   const wData = clientData->wData;
  const Glace_ImgArrays* const imgArraysPtr = clientData->imgArraysPtr;
  GtkPreview* const preview = wData->preview->preview;
  const gboolean has_alpha = wData->d->has_alpha;
  const gint w = imgArraysPtr->viewCols;
  const gint h = imgArraysPtr->viewRows;
  const guint in_bpp = wData->d->drawable->bpp;
  const guint out_bpp = in_bpp - (has_alpha ? 1 : 0);
  const guint rowstride = wData->d->gimpImgRowstride;

  register guint b;
  register gint xx, yy;
  GimpPixel *row_inbuf;
  PreviewPixel *rowbuf;
  register PreviewPixel *outbuf;

  row_inbuf = wData->d->gimpImgPtr 
    + GLACE_IMG(viewY0) * rowstride
    + GLACE_IMG(viewX0) * in_bpp;

  /* If we're fed an RGBA buffer (or GrayA), it must be strained into an RGB
     one before it is put in the preview widget. */
  if(has_alpha)
    { 
      register const GimpPixel *inbuf;

      rowbuf = g_new(PreviewPixel, out_bpp * w);

      for(yy=0; yy < h; yy++ )
	{
	  inbuf = row_inbuf;
	  outbuf=rowbuf;
	  for(xx=w; xx != 0; xx-- )
	    {
	      for(b=out_bpp; b != 0; b--)
		{
		  *outbuf = *inbuf;
		  outbuf++; inbuf++;
		}
	      inbuf++; /* advance over one alpha byte */
	    } /* next xx */
	  
	  gtk_preview_draw_row(preview, rowbuf, 0, yy, w);

	  row_inbuf += rowstride;
	} /* next yy */

      g_free(rowbuf);
    } 
  else /* no alpha byte to skip over */
    {
      for(yy=0; yy < h; yy++ )
	{
	  gtk_preview_draw_row(preview, row_inbuf, 0, yy, w);
	  
	  row_inbuf += rowstride;
	} /* next yy */
    }

  gtk_widget_draw(GTK_WIDGET(preview),NULL);

} /* end preview_reset() */

/* This is a wrapper around ace_preview_update which only calls it if
   the controls have been changed since it was last updated.  Returns
   TRUE if it did anything. */
gboolean
ace_preview_smart_update(Glace_ClientData* const clientData, 
			 const guint revision, const guint out_revision)
{
  gboolean did_update=FALSE;
  GlaceGimpData* wData=clientData->wData;
  
  if(wData->control_revision < revision)
    {
      ace_preview_update(clientData);

      wData->control_revision = MAX(wData->control_revision,
				    revision);
      wData->control_out_revision = MAX(wData->control_out_revision,
					out_revision);

      did_update=TRUE;
    }
  else if (wData->control_out_revision < out_revision)
    {
      /* If there's an update being calculated, don't interrupt it -- the 
	 output will happen when it's done.  */
      if(ace_preview_get_state(clientData->wData->preview) == PREVIEW_IDLE)
	{
	  GlaceGimp_Output(clientData->wData, clientData->cfgInfoPtr,
			   clientData->imgArraysPtr);
	}

      wData->control_out_revision = MAX(wData->control_out_revision,
					out_revision);

      did_update=TRUE;
    }

  return did_update;
}

/* Update the preview.  If calculations are currently in progress,
   cancel them. */
void 
ace_preview_update(Glace_ClientData* const clientData) 
{
  GlaceGimpPreview* const preview = clientData->wData->preview;

  if(ace_preview_get_state(preview) == PREVIEW_WORKING)
    {
      ace_preview_cancel(preview);
    }

  clientData->wData->target=GLACE_GIMP_PREVIEW;
  clientData->imgArraysPtr->pixelBytePad=0;

  ace_preview_progress_update(preview, (float)PROG_INPUT_CHUNKS/
			      (clientData->cfgInfoPtr->numTerms+PROG_OCHUNKS));
  ace_preview_set_state(preview, PREVIEW_WORKING);

  gimp_ace_go(clientData, G_PRIORITY_DEFAULT_IDLE - preview->idle_priority);
}

/* Called when the preview is finished updating.  Resets the progress
   indicator and clears the handler ID. */
void
ace_preview_update_finished(GlaceGimpPreview* const preview)
{
  ace_preview_set_state(preview, PREVIEW_IDLE);
  ace_preview_progress_update(preview, 0.0);
  preview->idle_handle = 0;
}

/* Prevent an in-progress update from continuing. */
void
ace_preview_cancel(GlaceGimpPreview* const preview)
{
  if(ace_preview_get_state(preview) == PREVIEW_WORKING) {
    gtk_idle_remove(preview->idle_handle);
    ace_preview_update_finished(preview);
  }
}

void
ace_preview_progress_update(const GlaceGimpPreview* const preview, 
			    const gfloat percentage)
{
  g_return_if_fail(preview);
  if(preview->pbar) 
    {
      /* FIXME: Get GTK lock */
      gtk_progress_set_percentage(preview->pbar, percentage);
      /* Release GTK lock */
    }
}

/* Simple accesor for getting the preview's state. */
AcePreviewState
ace_preview_get_state(const GlaceGimpPreview* const preview)
{
  return preview->state;
}

/* Simple accesor for setting the preview's state. */
void
ace_preview_set_state(GlaceGimpPreview* const preview, 
		      const AcePreviewState ps)
{
  preview->state=ps;
}


/**********************************************/
/* Private functions:                         */


/* dialog_preview_events for preview scrolling, adopted from
   Jens' and Quartic's bumpmap.c */
static gint
ace_preview_events(GtkWidget*   widget, 
		   GdkEvent*    event, 
		   GlaceGimpPreview* const preview)
{
  /* WARNING: static data in function! */
  static gint mouse_x, mouse_y;
  static gboolean drag_mode;
  
  gint  x,  y;
  gint dx, dy;
  
  GtkAdjustment* xadj =gtk_scrolled_window_get_hadjustment(preview->scrollwin);
  GtkAdjustment* yadj =gtk_scrolled_window_get_vadjustment(preview->scrollwin);

  gtk_widget_get_pointer(widget, &x, &y);
  
  switch (event->type) {
  case GDK_BUTTON_PRESS:
    mouse_x = x;
    mouse_y = y;
	
    drag_mode=TRUE;
    
    gtk_grab_add(widget);
    break;    
  case GDK_BUTTON_RELEASE:
    if (drag_mode != FALSE) {
      gtk_grab_remove(widget);
    }
    drag_mode=FALSE;
    break;    
  case GDK_MOTION_NOTIFY:
    /* this not being a button event, there is no button.button */
    dx = (x - mouse_x);
    dy = (y - mouse_y);
    
    mouse_x = x;
    mouse_y = y;
    
    if ((dx == 0) && (dy == 0))
      break;
    
    gtk_adjustment_set_value(xadj,MIN(xadj->value - dx,
				      xadj->upper-xadj->page_size));
    gtk_adjustment_set_value(yadj,MIN(yadj->value - dy,
				      yadj->upper - yadj->page_size));
    break;
  default:
    break;
  } /* switch */
  return FALSE;
} /* ace_preview_events */



/********************** Many-views *************************/
/********************* (Mini-views) ************************/
/******************** Many-mini-views **********************/

/* Creates (allocates) a new ManyviewMatrix, including all
   the widgets in the table interface and all the clients needed
   to drive the previews.  The one thing it does NOT do is 
   set the axis adjustments. */
ManyviewMatrix*
manyview_matrix_new(const GlaceGimpDrawableInfo* drawable_info)
{
  ManyviewMatrix *self = g_new0(ManyviewMatrix,1);
  GtkTable* table;
  GtkWidget* box;
  GtkWidget* button;
  char buf[31];
  const int buflen=31;
  int xx, yy;
  
  g_assert(drawable_info);

  self->tablewidget = gtk_table_new(9,9,!HOMOG);
  table=GTK_TABLE(self->tablewidget);
  
  /* Axis selection and title. */

  /* axis title: x */
  box=GTK_WIDGET(manyview_axis_box_new(self,'x',&self->xmenu,&self->xdelta));
  gtk_table_attach(table,box,
		   1,9,
		   0,1,
		   /* don't expand y: */ GTK_EXPAND | GTK_FILL, 0,
		   /* padding:      */ 0, 0);
  

  /* axis title: y */
  box=GTK_WIDGET(manyview_axis_box_new(self,'y',&self->ymenu,&self->ydelta));
  gtk_table_attach(table,box,
		   0,1,
		   0,9,
		   /* don't expand x: */ GTK_FILL, GTK_EXPAND | GTK_FILL,
		   /* padding:      */ 0, 0);


  /* Axis labels */
  /* axis labels: x */
  for(xx=0;xx<3;xx++)
    {
      g_snprintf(buf,buflen,"%3f",(float)xx);
      self->xlabels[xx] = gtk_label_new(buf);
      gtk_table_attach(table,self->xlabels[xx],
		       3+xx*2, 3+xx*2+1,
		       1,2,
		       /* don't expand: */ 0, 0,
		       /* padding:      */ 0, 0);
    }
  /* axis labels: y */
  for(yy=0;yy<3;yy++)
    {
      g_snprintf(buf,buflen,"%3f",(float)yy);
      self->ylabels[yy] = gtk_label_new(buf);
      gtk_table_attach(table,self->ylabels[yy],
		       1,2,
		       3+yy*2, 3+yy*2+1,
		       /* don't expand: */ 0, 0,
		       /* padding:      */ 0, 0);
    }

  /* Viewfinder */

  self->view_x = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, drawable_info->w,1,10,
					MIN(PREVIEW_WIDTH,drawable_info->w)));
  self->view_y = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, drawable_info->h,1,10,
				    MIN(PREVIEW_HEIGHT,drawable_info->h)));

  gtk_signal_connect(GTK_OBJECT(self->view_x),"changed", view_changed, self);
  gtk_signal_connect(GTK_OBJECT(self->view_x),"value_changed", 
		     view_changed, self);
  gtk_signal_connect(GTK_OBJECT(self->view_y),"changed", view_changed, self);
  gtk_signal_connect(GTK_OBJECT(self->view_y),"value_changed", 
		     view_changed, self);

  self->viewfinder=viewfinder_new(drawable_info, self);

  gtk_box_pack_start(GTK_BOX(box),self->viewfinder,!EXPAND,!FILL,0);
  gtk_box_reorder_child(GTK_BOX(box),self->viewfinder,0);

  /* Navigation arrows */
#if 0
  /* QUESTION: Clicking on a many-view will re-center using that many-view's
     parameters.  The question is, do we want, in addition to that, outer
     navigation arrows with a meaning of "re-center and *beyond*", which would
     increase the page increment as well as re-centering... ? */

  /* Outer navigaton arrows */
  for(yy=0;yy<3;yy++)
    {
      for(xx=0;xx<3;xx++)
	{
	  if(yy==1 && xx==1) continue;
	  button = gtk_arrowbutton_new(xx-1,-1*(yy-1));

	  gtk_object_set_data(GTK_OBJECT(container),"x_pos",
			      GINT_TO_POINTER((xx-1)*3));
	  gtk_object_set_data(GTK_OBJECT(container),"y_pos",
			      GINT_TO_POINTER((yy-1)*3));

	  gtk_signal_connect(GTK_OBJECT(container),"clicked",
			     manyview_navbutton_clicked,
			     self);

	  gtk_table_attach_defaults(table,button,
				    2+xx*3,2+xx*3+1,
				    2+yy*3,2+yy*3+1);
	} /* next xx */
    } /* next yy */
#endif /* outer navigation arrows. */

  /* Inner navigaton arrows. */
  for(yy=0;yy<3;yy++)
    {
      for(xx=0;xx<3;xx++)
	{
	  if(yy==1 && xx==1) continue;
	  button = gtk_3dotbutton_new(xx-1,-1*(yy-1));

	  gtk_object_set_data(GTK_OBJECT(button),"x_pos",
			      GINT_TO_POINTER((xx-1)*1));
	  gtk_object_set_data(GTK_OBJECT(button),"y_pos",
			      GINT_TO_POINTER((yy-1)*1));

	  gtk_signal_connect(GTK_OBJECT(button),"clicked",
			     manyview_navbutton_clicked,
			     self);

	  gtk_table_attach(table,button,
			   4+xx,4+xx+1,
			   4+yy,4+yy+1,
			   /* don't expand: */ 0, 0,
			   /* padding:      */ 0, 0);
	} /* next xx */
    } /* next yy */

  /* Mini-views */
  for(yy=0;yy<3;yy++)
    {
      for(xx=0;xx<3;xx++)
	{
	  GtkWidget* container;

	  self->clients[yy][xx] = manyview_new(drawable_info,
					       self,
					       (1 + (xx==1) + (yy==1)) *5);

	  manyview_set_geometry(self->clients[yy][xx],
				0,0,
				MIN(PREVIEW_WIDTH,drawable_info->w),
				MIN(PREVIEW_HEIGHT,drawable_info->h));

	  if(yy == 1 && xx == 1)
	    {
	      container=gtk_frame_new(NULL);
	      gtk_frame_set_shadow_type(GTK_FRAME(container),
					GTK_SHADOW_IN);
	    }
	  else
	    {
	      container=gtk_button_new();
	      gtk_button_set_relief(GTK_BUTTON(container),GTK_RELIEF_HALF);

	      gtk_object_set_data(GTK_OBJECT(container),"x_pos",
				  GINT_TO_POINTER((xx-1)*2));
	      gtk_object_set_data(GTK_OBJECT(container),"y_pos",
				  GINT_TO_POINTER((yy-1)*2));

	      gtk_signal_connect(GTK_OBJECT(container),"clicked",
				 manyview_navbutton_clicked,
				 self);
	    }
	  gtk_container_add(GTK_CONTAINER(container),
			    GTK_WIDGET(self->clients[yy][xx]->wData->preview->preview));
	  gtk_table_attach(table,
			   container,
			   3+2*xx,3+2*xx+1,
			   3+2*yy,3+2*yy+1,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   0,0);

	} /* next xx */
    } /* next yy */

  return self;
}

void
manyview_register_adjustment(ManyviewMatrix* mv_matrix, GtkAdjustment* adj)
{
  GtkWidget *menuitem;
  g_assert(mv_matrix);
  g_return_if_fail(GTK_IS_ADJUSTMENT(adj));

  menuitem=
    gtk_menu_item_new_with_label((char*)gtk_object_get_data(GTK_OBJECT(adj),
							    "label_text"));

  gtk_object_set_data(GTK_OBJECT(menuitem),"adjustment",adj);
  gtk_object_set_data(GTK_OBJECT(menuitem),"x_or_y",GINT_TO_POINTER('x'));

  gtk_signal_connect(GTK_OBJECT(menuitem),"select",
		     manyview_axis_menu_callback, mv_matrix);

  gtk_menu_append(mv_matrix->xmenu,menuitem);

  if(!mv_matrix->x_axis)
    gtk_menu_item_select(GTK_MENU_ITEM(menuitem));

  /* Do it again for the Y axis */

  menuitem=
    gtk_menu_item_new_with_label((char*)gtk_object_get_data(GTK_OBJECT(adj),
							    "label_text"));

  gtk_object_set_data(GTK_OBJECT(menuitem),"adjustment",adj);
  gtk_object_set_data(GTK_OBJECT(menuitem),"x_or_y",GINT_TO_POINTER('y'));

  gtk_signal_connect(GTK_OBJECT(menuitem),"select",
		     manyview_axis_menu_callback, mv_matrix);

  gtk_menu_append(mv_matrix->ymenu,menuitem);

  if(!mv_matrix->y_axis && adj != mv_matrix->x_axis)
    gtk_menu_item_select(GTK_MENU_ITEM(menuitem));
}

/* FIXME: Axes should have session management. */
void
manyview_set_axis(ManyviewMatrix* mv_matrix, char x_or_y, GtkAdjustment* adj)
{
  GtkAdjustment** axis_adj_p;
  GtkAdjustment* delta_adj;
  g_assert(mv_matrix);
  g_assert(x_or_y == 'x' || x_or_y == 'X' || 
	   x_or_y == 'y' || x_or_y == 'Y');

  if(adj) g_assert(GTK_IS_ADJUSTMENT(adj));

  switch(x_or_y)
    {
    case 'X':
    case 'x':
      axis_adj_p = &mv_matrix->x_axis;
      delta_adj = mv_matrix->xdelta;
      break;
    case 'Y':
    case 'y':
      axis_adj_p = &mv_matrix->y_axis;
      delta_adj = mv_matrix->ydelta;
      break;
    default:
      g_assert_not_reached();
      axis_adj_p = NULL;
      delta_adj = NULL; /* Silence GCC warnings. */
    }
  
  *axis_adj_p=adj;

  delta_adj->upper=adj->upper/2;
  gtk_adjustment_changed(delta_adj);
  gtk_adjustment_set_value(delta_adj, adj->page_increment);
  /* the set_value on delta_adj should do the job of triggering a 
     recalc/refresh for us. */
}

/* Initalize the axes. */
void
manyview_look_alive(ManyviewMatrix* mv_matrix)
{
  /* FIXME: Is this function still needed? */
}

/* Pass -1 to args which should use default values. */
void
manyview_set_geometry(Glace_ClientData* clientData, 
		      int view_x0, int view_y0,
		      int view_width, int view_height)
{
  Glace_ImgArrays* const imgArraysPtr = clientData->imgArraysPtr;
  const Glace_CfgInfo* const cfgInfoPtr = clientData->cfgInfoPtr;
  GlaceGimpData* const wData = clientData->wData;
  int windowX1, windowY1;
  int xpad, ypad;

  g_assert(clientData);

  /* Set defaults for arguments that weren't supplied. */
  if(view_x0 == -1) view_x0 = GLACE_IMG(viewX0);
  if(view_y0 == -1) view_y0 = GLACE_IMG(viewY0);
  if(view_width == -1) view_width = GLACE_IMG(viewCols);
  if(view_height == -1) view_height = GLACE_IMG(viewRows);

  xpad = GLACE_CFG(windBaseW);
  ypad = GLACE_CFG(windBaseH);

  /** Bounds checking **/
  /* View may not be larger than selection. */
  view_width=CLAMP(view_width,0,WRAPPER(d)->w);
  view_height=CLAMP(view_height,0,WRAPPER(d)->h);

  GLACE_IMG(viewCols)=view_width;
  GLACE_IMG(viewRows)=view_height;

  /* Entire view must fit within selection. */
  view_x0=MIN(view_x0, WRAPPER(d)->w - view_width);
  view_y0=MIN(view_y0, WRAPPER(d)->h - view_height);

  GLACE_IMG(viewX0)=view_x0;
  GLACE_IMG(viewY0)=view_y0;

  /** Set window geometry. **/
  /* coordinates */
  GLACE_IMG(windowX0)=MAX(view_x0-xpad, 0);
  GLACE_IMG(windowY0)=MAX(view_y0-ypad, 0);

  /* size */
  windowX1=MIN(view_x0 + view_width + xpad, WRAPPER(d)->w);
  windowY1=MIN(view_y0 + view_height + ypad, WRAPPER(d)->h);

  GLACE_IMG(cols)=windowX1-GLACE_IMG(windowX0);
  GLACE_IMG(rows)=windowY1-GLACE_IMG(windowY0);

  /* Set starting point in image buffer. */

  GLACE_IMG(inImgPtr) = WRAPPER(d)->inImgPtr 
    + GLACE_IMG(windowY0) * GLACE_IMG(rowstride)
    + GLACE_IMG(windowX0);

  GLACE_IMG(refImgHBPtr) = GLACE_IMG(inImgHBPtr) = GLACE_IMG(inImgPtr);
}

gboolean
manyview_smart_update(ManyviewMatrix* manyview_matrix, 
		      const guint revision, const guint out_revision)
{
  guint xx,yy;
  gboolean did_update=FALSE;

  /* g_print("%s called on revision %d... ",
     G_GNUC_PRETTY_FUNCTION, revision); */

  /* FIXME: This might be where you would want threading action to happen. */

  for(yy=0;yy<3;yy++)
    {
      for(xx=0;xx<3;xx++)
	{
	  did_update |= 
	    ace_preview_smart_update(manyview_matrix->clients[yy][xx],
				     revision, out_revision);
	}
    } /* next yy */
  
  /* g_print("%d finished, %s returning\n",
     revision,G_GNUC_PRETTY_FUNCTION); */

  return did_update;
}

static GtkBox*
manyview_axis_box_new(ManyviewMatrix* mv_matrix, char x_or_y, 
		      GtkMenu** menu_p, GtkAdjustment **delta_adj_p)
{
  GtkWidget *omenu;
  GtkWidget *spin;
  GtkBox *box;
  g_assert(x_or_y == 'x' || x_or_y == 'X' || 
	   x_or_y == 'y' || x_or_y == 'Y');

  *menu_p=GTK_MENU(gtk_menu_new());
  omenu=gtk_option_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(omenu),GTK_WIDGET(*menu_p));

  /* Note: All the default/upper/etc values here are completely bogus
     and will be re-set at the first opportunity. */
  *delta_adj_p=GTK_ADJUSTMENT(gtk_adjustment_new(0.1,0.0,1.0,0.1,0.5,0));
  spin = gtk_spin_button_new(*delta_adj_p,
			     0.1 /* FIXME: climb rate */,
			     2 /* decimal places*/);

  switch(x_or_y)
    {
    case 'X':
    case 'x':
      box=GTK_BOX(gtk_hbox_new(!HOMOG,SPACING));
      gtk_signal_connect(GTK_OBJECT(*delta_adj_p),"value-changed", 
			 manyview_axis_delta_changed,&mv_matrix->x_axis);
      break;
    case 'Y':
    case 'y':
      box=GTK_BOX(gtk_vbox_new(!HOMOG,SPACING));
      gtk_signal_connect(GTK_OBJECT(*delta_adj_p),"value-changed", 
			 manyview_axis_delta_changed,&mv_matrix->y_axis);
      break;
    default:
      g_assert_not_reached();
      box=NULL; /* Silence GCC warnings. */
    }

  gtk_box_pack_start(box,omenu,EXPAND,FILL,/*FIXME: Padding*/0);
  gtk_box_pack_end(box,spin,EXPAND,FILL,/*FIXME: Padding*/0);

  return box;
}

/* This responds to "select" from the menuitem on the axis menu. */
static void
manyview_axis_menu_callback(GtkMenuItem* menuitem, 
			    ManyviewMatrix* mv_matrix)
{
  GtkAdjustment *adj;

  g_assert(GTK_IS_MENU_ITEM(menuitem));
  g_return_if_fail(mv_matrix);

  adj = GTK_ADJUSTMENT(gtk_object_get_data(GTK_OBJECT(menuitem),"adjustment"));

  manyview_set_axis(mv_matrix, 
		    GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(menuitem),
							"x_or_y")),
		    adj);
}

/* Responds to "value-changed" from the axis's delta adjustment. */
static void
manyview_axis_delta_changed(GtkAdjustment* delta_adj,GtkAdjustment** axis_adj)
{
  g_assert(GTK_IS_ADJUSTMENT(delta_adj));
  g_assert(axis_adj);
  g_return_if_fail(GTK_IS_ADJUSTMENT(*axis_adj));

  (*axis_adj)->page_increment=delta_adj->value;
  gtk_adjustment_changed(*axis_adj);

  /* FIXME: Make sure the manyview is updated after this. */
}

/* This responds to "clicked" from a manyview navbutton */
static void
manyview_navbutton_clicked(GtkButton* button, ManyviewMatrix* mv_matrix)
{
  float x_mult, y_mult;
  float newvalue;
  g_assert(GTK_IS_BUTTON(button));
  g_assert(mv_matrix);

  x_mult = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(button),"x_pos"))
    / 2.0;

  y_mult = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(button),"y_pos"))
    / 2.0;

  /* Page size == 1.0 indicates that this is an enumerated type. */
  if(mv_matrix->x_axis->page_size == 1.0) 
    x_mult = (x_mult == 0) ? 0 : ((x_mult < 0) ? -1.0 : 1.0);

  if(mv_matrix->y_axis->page_size == 1.0) 
    y_mult = (y_mult == 0) ? 0 : ((y_mult < 0) ? -1.0 : 1.0);

  g_print("x_mult: %f\ty_mult: %f\n",x_mult,y_mult);

  if(x_mult != 0)
    {
      newvalue = mv_matrix->x_axis->value + 
	x_mult * mv_matrix->x_axis->page_increment;
      gtk_adjustment_set_value(mv_matrix->x_axis,newvalue);
      if(fabs(x_mult) != 1.0)
	{
	  gtk_adjustment_set_value(mv_matrix->xdelta,
				   mv_matrix->xdelta->value * fabs(x_mult));
	}
    }

  if(y_mult != 0)
    {
      newvalue = mv_matrix->y_axis->value + 
	y_mult * mv_matrix->y_axis->page_increment;
      gtk_adjustment_set_value(mv_matrix->y_axis,newvalue);
      if(fabs(y_mult) != 1.0)
	{
	  gtk_adjustment_set_value(mv_matrix->ydelta,
				   mv_matrix->ydelta->value * fabs(y_mult));
	}
    }
}

/* This function is analogous to ace_preview_new().  The fact that
   they both exist means that some re-designing is in order. */
static Glace_ClientData*
manyview_new(const GlaceGimpDrawableInfo* drawable_info, 
	     ManyviewMatrix* mv_matrix, int priority)
{
  Glace_ClientData* const clientData = clientdata_new(drawable_info,NULL);
  GlaceGimpData* const wData = clientData->wData;
  GlaceGimpPreview* const preview = g_new0(GlaceGimpPreview,1);
  const gboolean is_colour = gimp_drawable_is_rgb(drawable_info->drawable->id);

  WRAPPER(preview)=preview;

  preview->bpp = is_colour ? 3 : 1;

  preview->preview=GTK_PREVIEW(gtk_preview_new(is_colour ? GTK_PREVIEW_COLOR :
					       GTK_PREVIEW_GRAYSCALE));

  gtk_preview_size(WRAPPER(preview)->preview, 
		   MIN(PREVIEW_WIDTH,drawable_info->w),
		   MIN(PREVIEW_HEIGHT,drawable_info->h));

  gtk_preview_set_expand(preview->preview,TRUE);

  gtk_signal_connect(GTK_OBJECT(preview->preview),"size-allocate",
		     manyview_size_allocate, mv_matrix);

  preview->idle_handle = 0;
  preview->idle_priority = priority;
  preview->state = PREVIEW_IDLE;

  clientData->cfgInfoPtr->verbose=FALSE;

  return clientData;
}

/* NOTE: This will be called for *each preview in the manyview*,
   however mv_matrix->view_x,y are for the whole manyview. */
static void
manyview_size_allocate(GtkWidget* G_GNUC_UNUSED unused_preview, 
		       GtkAllocation* allocation, 
		       ManyviewMatrix* mv_matrix)
{
  g_assert(allocation);
  g_assert(mv_matrix);

  mv_matrix->view_x->page_size=allocation->width;
  mv_matrix->view_y->page_size=allocation->height;
  gtk_adjustment_changed(mv_matrix->view_x);
  gtk_adjustment_changed(mv_matrix->view_y);
}


/* Includes a quick and dirty integer scaling routine to fit the
   input buffer inside NAV_PREVIEW_SIZE.  Returns the happyfunwidget. */
/* FIXME: audit for robustness when drawable_info->w % src_to_dst_ratio != 0 */
static GtkWidget*
viewfinder_new(const GlaceGimpDrawableInfo* drawable_info, 
	       ManyviewMatrix *mv_matrix)
{
  const GimpPixel* const input = drawable_info->gimpImgPtr;
  GtkPreview *preview;
  PreviewPixel* outbuf;
  PreviewPixel* rowbuf;
  register int out_y;
  register int in_x;
  register int in_y;
  register int xx;
  register int yy;
  register int b;
  const int out_bpp =
    gimp_drawable_is_rgb(drawable_info->drawable->id) ? 3 : 1;
  const int src_to_dst_ratio = MAX(1,RINT(MAX(drawable_info->w,
					      drawable_info->h)/
					  (double)NAV_PREVIEW_SIZE));
  const int out_w= drawable_info->w/src_to_dst_ratio;
  const int in_w= MIN(out_w * src_to_dst_ratio, drawable_info->w);
  const int out_h = drawable_info->h/src_to_dst_ratio;
  const int in_h = MIN(out_h * src_to_dst_ratio, drawable_info->h);

  g_assert(drawable_info);
  g_assert(mv_matrix);
  
  preview=GTK_PREVIEW(gtk_preview_new(out_bpp == 3 ? GTK_PREVIEW_COLOR : 
			  GTK_PREVIEW_GRAYSCALE));


  gtk_preview_size(preview, out_w, out_h);
  gtk_preview_set_expand(preview,FALSE);

  rowbuf=g_new0(PreviewPixel,out_w * out_bpp);

  for(in_y=0, out_y=0; in_y < in_h; in_y += src_to_dst_ratio, out_y++)
    {
      outbuf=rowbuf;
      for(in_x=0; in_x < in_w; in_x += src_to_dst_ratio)
	{
	  long output[4] = {0,0,0,0};

	  /* Find the average value for a block of pixels 
	     (src_to_dst_ratio per side) */
	  for(yy=0;yy<src_to_dst_ratio;yy++)
	    {
	      for(xx=0;xx<src_to_dst_ratio;xx++)
		{
		  for(b=0;b<out_bpp;b++)
		    {
		      output[b]+=input[(in_y+yy)*
				      drawable_info->gimpImgRowstride+
				      (in_x+xx)*
				      drawable_info->drawable->bpp + b];
		    }
		} /* next xx */
	    } /* next yy */
	  
	  for(b=0;b<out_bpp;b++)
	    {
	      *outbuf=output[b]/SQR(src_to_dst_ratio);
	      outbuf++;
	    }
	} /* next destination pixel */

      gtk_preview_draw_row(preview,rowbuf,0,out_y, out_w);
    } /* next destination row */

  g_free(rowbuf);

  gtk_object_set_data(GTK_OBJECT(preview),"src_to_dst_ratio",
		      GINT_TO_POINTER(src_to_dst_ratio));


  /*************************************************/
  /* Position indicator */

  /* gets the GC */
  gtk_signal_connect(GTK_OBJECT(preview),"realize",
		     viewfinder_realize,NULL);

  /* draws the indicator */
  gtk_signal_connect(GTK_OBJECT(preview),"draw",
		     viewfinder_draw, mv_matrix);

  gtk_widget_set_events(GTK_WIDGET(preview), 
			GDK_BUTTON_PRESS_MASK |
			GDK_BUTTON_RELEASE_MASK | 
			GDK_BUTTON_MOTION_MASK |
			GDK_POINTER_MOTION_HINT_MASK);
  gtk_signal_connect(GTK_OBJECT(preview), "event",
		     GTK_SIGNAL_FUNC(viewfinder_events),
		     mv_matrix);

  return GTK_WIDGET(preview);
}

/* GtkWidget "realize" */
static void
viewfinder_realize(GtkWidget* widget, gpointer G_GNUC_UNUSED unused_data)
{
  GdkGCValues gc_vals;
  GdkGC *gc;

  gc_vals.function=GDK_INVERT;
  gc_vals.line_width=BORDER_PEN_WIDTH;
  gc_vals.line_style=GDK_LINE_SOLID;  

  gc = gtk_gc_get(gtk_widget_get_visual(widget)->depth, 
		  gtk_widget_get_colormap(widget), 
		  &gc_vals,
		  GDK_GC_FUNCTION | GDK_GC_LINE_WIDTH | GDK_GC_LINE_STYLE);

  gtk_object_set_data(GTK_OBJECT(widget),"gc",gc);
}

/* GtkWidget "draw" */
static void
viewfinder_draw(GtkWidget* preview, GdkRectangle* G_GNUC_UNUSED unused_area, 
		ManyviewMatrix* mv_matrix)
{
  const int src_to_dst_ratio = 
    GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(preview), 
					"src_to_dst_ratio"));

  viewfinder_draw_rectangle(preview,
			    mv_matrix->view_x->value / src_to_dst_ratio,
			    mv_matrix->view_y->value / src_to_dst_ratio,
			    mv_matrix->view_x->page_size/src_to_dst_ratio,
			    mv_matrix->view_y->page_size/src_to_dst_ratio,
			    FALSE);
}

static void
viewfinder_draw_rectangle(GtkWidget* preview, 
			  int x, int y, int w, int h,
			  gboolean erase_old)
{
  GdkGC *gc=gtk_object_get_data(GTK_OBJECT(preview),"gc");
  static int rect_x, rect_y, rect_w, rect_h;

  g_assert(GTK_IS_WIDGET(preview));

  if(erase_old)
    {
      gdk_draw_rectangle(preview->window, gc,
			 FALSE, 
			 rect_x, rect_y,
			 rect_w, rect_h);
    }

  rect_x=x;
  rect_y=y;
  rect_w=w;
  rect_h=h;

  gdk_draw_rectangle(preview->window, gc,
		     FALSE, 
		     rect_x, rect_y,
		     rect_w, rect_h);
}

/* GtkAdjustment "changed" and "value-changed" signals */
static void 
view_changed(GtkAdjustment* G_GNUC_UNUSED unused_adjustment, 
	     ManyviewMatrix* mv_matrix)
{
  int x,y;
  g_return_if_fail(mv_matrix->clients[1][1]);

  ace_preview_fill_from_source(mv_matrix->clients[1][1]);

  for(y=0;y<3;y++) {
    for(x=0;x<3;x++) {
      manyview_set_geometry(mv_matrix->clients[y][x],
			    mv_matrix->view_x->value,
			    mv_matrix->view_y->value,
			    mv_matrix->view_x->page_size,
			    mv_matrix->view_y->page_size);
      ace_preview_update(mv_matrix->clients[y][x]);
    } /* next x */
  } /* next y */
  
}

static gint
viewfinder_events(GtkWidget*   widget, 
		  GdkEvent*    event, 
		  ManyviewMatrix* mv_matrix)
{
  /* WARNING: static data in function */
  gint  x,  y;
  
  gtk_widget_get_pointer(widget, &x, &y);
  
  switch (event->type) {
  case GDK_BUTTON_RELEASE:
    break;    
  case GDK_BUTTON_PRESS:
  case GDK_MOTION_NOTIFY:
    {
      const int src_to_dst_ratio = 
	GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(widget), 
					    "src_to_dst_ratio"));

      x=MAX(0,x-mv_matrix->view_x->page_size / (2*src_to_dst_ratio));
      y=MAX(0,y-mv_matrix->view_y->page_size / (2*src_to_dst_ratio));

      gtk_adjustment_clamp_page(mv_matrix->view_x,
				x*src_to_dst_ratio,
				mv_matrix->view_x->upper);
      gtk_adjustment_clamp_page(mv_matrix->view_y,
				y*src_to_dst_ratio,
				mv_matrix->view_y->upper);

      viewfinder_draw_rectangle(widget,
				mv_matrix->view_x->value / src_to_dst_ratio,
				mv_matrix->view_y->value / src_to_dst_ratio,
				mv_matrix->view_x->page_size/src_to_dst_ratio,
				mv_matrix->view_y->page_size/src_to_dst_ratio,
				TRUE);
    }
    break;
  default:
    break;
  } /* switch */
  return FALSE;
} /* ace_preview_events */


/********************************************************************/
/* Silly little widgets. 
 */
#if 0 /* arrowbutton not currently used */
#define ARROWBUTTON_SHADOW_TYPE GTK_SHADOW_IN

/* FIXME: These have a bucketfull of ugliness issues.  But please
   think twice before replacing me with something resolution-dependant
   (e.g. pixmap). */
static GtkWidget*
gtk_arrowbutton_new(int horizontal, int vertical)
{
  GtkWidget *button;
  GtkTable *table;
  GtkWidget *arrow;
  
  button=gtk_button_new();
  table=GTK_TABLE(gtk_table_new(3,3,!HOMOG));

  gtk_container_add(GTK_CONTAINER(button),GTK_WIDGET(table));

  if(horizontal > 0)
    {
      arrow = gtk_arrow_new(GTK_ARROW_RIGHT, ARROWBUTTON_SHADOW_TYPE);
      gtk_table_attach_defaults(table,arrow,2,3,1,2);
    }
  else if(horizontal < 0)
    {
      arrow = gtk_arrow_new(GTK_ARROW_LEFT, ARROWBUTTON_SHADOW_TYPE);
      gtk_table_attach_defaults(table,arrow,0,1,1,2);
    }
  if(vertical > 0)
    {
      arrow = gtk_arrow_new(GTK_ARROW_UP, ARROWBUTTON_SHADOW_TYPE);
      gtk_table_attach_defaults(table,arrow,1,2,0,1);
    }
  else if(vertical < 0)
    {
      arrow = gtk_arrow_new(GTK_ARROW_DOWN, ARROWBUTTON_SHADOW_TYPE);
      gtk_table_attach_defaults(table,arrow,1,2,2,3);
    }

  return button;
}
#endif /* arrowbutton not currently used */

/* FIXME: Like the arrowbuttons, this is a pretty ugly hack, in more
   ways than one. */
static GtkWidget*
gtk_3dotbutton_new(int horizontal, int vertical)
{
  GtkWidget *button;
  GtkTable *table;
  GtkWidget *dot[3];
  GdkPixmap *dot_gdk_pixmap;
  int i,xoffset,xmult,yoffset,ymult;
  
  button=gtk_button_new();
  table=GTK_TABLE(gtk_table_new(3,3,HOMOG));

  gtk_button_set_relief(GTK_BUTTON(button),GTK_RELIEF_NONE);
  gtk_container_add(GTK_CONTAINER(button),GTK_WIDGET(table));

  /* FIXME: These may not be visible at all resolutions or GTK themes. */
  dot_gdk_pixmap = 
    gdk_pixmap_colormap_create_from_xpm_d(button->window,
					  /* FIXME: colormap: */ NULL, 
					  /* mask: */ NULL,
					  /* transparent color: */ NULL,
					  dot_xpm);

  for(i=0;i<3;i++)
    {
      dot[i] = gtk_pixmap_new(dot_gdk_pixmap, NULL);
    }

  if((horizontal*vertical)>0)		/* / */
    {
      xoffset=2; xmult= -1;
      yoffset=0; ymult=  1;
    }
  else if((horizontal*vertical)<0)	/* \ */
    {
      xoffset=0; xmult=1;
      yoffset=0; ymult=1;
    }
  else if(horizontal)			/* - */
    {
      xoffset=0; xmult=1;
      yoffset=1; ymult=0;
    }
  else if(vertical)			/* | */
    {
      xoffset=1; xmult=0;
      yoffset=0; ymult=1;
    }
  else 
    {
      g_assert_not_reached();
      xoffset=yoffset=xmult=ymult=0; /* Silence GCC warnings. */
    }

  for(i=0;i<3;i++)
    {
      gtk_table_attach_defaults(table,dot[i],
				xoffset+xmult*i, xoffset+xmult*i+1,
				yoffset+ymult*i, yoffset+ymult*i+1);
    }

  return button;
}

/* For emacs:
Local Variables:
mode:c
c-file-style:"gnu"
End:
*/
