/*
    Homogenizer - A GIMP plugin for removing non-homogeneous lighting in images.
    Copyright (C) 2000  Paul Francis Harrison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <unistd.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

typedef int bool;
const bool true = 1, false = 0;

/* A structure for holding options about what the user wants
   to do to the image exactly */
typedef struct {
  guint xOrder, yOrder;
} Parameters;

void click_func(GtkWidget *widget, gpointer data) {
  if (data)
    *((bool*)data) = true;
  gtk_main_quit();
}

gint delete_func(GtkWidget *widget, GdkEvent *event, gpointer data) {
  if (data)
    *((bool*)data) = true;
  gtk_main_quit();
  return TRUE;
}

/* Open a dialog, get user to select options */
bool getParametersByAsking(Parameters *param, int defaultDrawable) {
  GtkWidget *window, *button_box, *ok, *cancel, *main_box, *frame,
            *table;
  GtkObject *xAdjustment, *yAdjustment;
  GtkWidget *xSpinner, *ySpinner;
  bool cancelled = false;

  window = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(window), "Homogenize");
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
    GTK_SIGNAL_FUNC(delete_func), &cancelled);

  button_box = gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(button_box), 4);
  gtk_box_pack_end(GTK_BOX(GTK_DIALOG(window)->action_area),button_box, 
    FALSE,FALSE,0);

  ok = gtk_button_new_with_label("OK");
  GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(ok), "clicked",
    GTK_SIGNAL_FUNC(click_func), 0);
  gtk_box_pack_start(GTK_BOX(button_box), ok, FALSE,FALSE,0);
  gtk_widget_grab_default(ok);

  cancel = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(cancel), "clicked",
    GTK_SIGNAL_FUNC(click_func), &cancelled);
  gtk_box_pack_start(GTK_BOX(button_box), cancel, FALSE,FALSE,0);

  main_box = GTK_DIALOG(window)->vbox;
  gtk_container_set_border_width (GTK_CONTAINER(main_box), 4);
 
  frame = gtk_frame_new("Order");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame), 6);
  gtk_box_pack_start(GTK_BOX(main_box), frame, TRUE, TRUE, 0); 

  table = gtk_table_new (2, 2, FALSE);
  gtk_table_set_col_spacings(GTK_TABLE(table), 4);
  gtk_table_set_row_spacings(GTK_TABLE(table), 2);
  gtk_container_set_border_width(GTK_CONTAINER (table), 4);
  gtk_container_add(GTK_CONTAINER(frame), table);

  xSpinner = gimp_spin_button_new(&xAdjustment,
     param->xOrder, 1, 32, 1, 1, 1, 1, 0);

  gtk_signal_connect(GTK_OBJECT(xAdjustment), "value_changed",
    GTK_SIGNAL_FUNC(gimp_int_adjustment_update), &param->xOrder);

  gimp_table_attach_aligned(GTK_TABLE(table), 0,0, "Horizontal:", 1.0, 0.5,
    xSpinner, 1, TRUE);
  
  ySpinner = gimp_spin_button_new(&yAdjustment,
     param->yOrder, 1, 32, 1, 1, 1, 1, 0);
  
  gtk_signal_connect(GTK_OBJECT(yAdjustment), "value_changed",
    GTK_SIGNAL_FUNC(gimp_int_adjustment_update), &param->yOrder);

  gimp_table_attach_aligned(GTK_TABLE(table), 0,1, "Vertical:", 1.0, 0.5,
    ySpinner, 1, TRUE);

  gtk_widget_show_all(window);

  gtk_main();

  gtk_widget_destroy(window);

  gdk_flush();
  
  if (cancelled)
    return true;
  
  return false;
}

/* Restore the last parameters used from deep in the bowels of
   the main GIMP app */
bool getLastParameters(Parameters *param) {
  /* Set up defaults in case this is first time */
  param->xOrder = 2;
  param->yOrder = 2;

  gimp_get_data("plug_in_homogenizer", param);

  return false;
}

/* Convert argument list into parameters */
bool getParametersFromList(Parameters *param, int nArgs, GimpParam *args) {
  if (nArgs != 5)
    return true;

  param->xOrder          = args[3 ].data.d_int32;
  param->yOrder          = args[4 ].data.d_int32;

  return false;
}

/* Talking to GIMP stuff */

static void query();
static void run(gchar *name,
                gint nparams,
		GimpParam *param,
		gint *nreturn_vals,
		GimpParam **return_vals);

GimpPlugInInfo PLUG_IN_INFO = {
  NULL, /* init_proc */
  NULL, /* quit_proc */
  query, /* query_proc */
  run, /* run_proc */
};

/* Macro to define the usual plugin main function */
MAIN()

/* Add capabilities of this plugin to Procedural DataBase */
static void query() {
  static GimpParamDef args[] = {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE, "image", "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },

    { GIMP_PDB_INT32, "xorder", "X Order" },
    { GIMP_PDB_INT32, "yorder", "Y Order" }
  };

  GimpParamDef *return_vals = NULL;
  gint nargs = sizeof(args)/sizeof(args[0]);
  gint nreturn_vals = 0;

  gimp_install_procedure("plug-in-homogenizer",
			 "Remove very low frequencies (eg gradients)"
			 " in image",
			 "",
			 "Paul Francis Harrison",
			 "Paul Francis Harrison",
			 "2000",
			 "<Image>/Filters/Enhance/Homogenize...",
			 "RGB*, GRAY*",
			 GIMP_PLUGIN,
			 nargs, nreturn_vals,
			 args, return_vals);
}


static void run(gchar *name,
                gint nparams,
		GimpParam *param,
		gint *nreturn_vals,
		GimpParam **return_vals) {
  guint i,j,k,x,xx,y,yy,color;
  GimpDrawable *drawable;
  Parameters parameters;
  bool cancelled;
  GimpPixelRgn srcRegion, destRegion;
  int x1,y1,x2,y2,width,height;
  gpointer pr;
  double xStep, yStep;
  double *weights;
  double **xCosTable, **yCosTable;
  int maxProgress, progress;

  static GimpParam values[1];
  
  *nreturn_vals = 1;
  *return_vals = values;
  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = GIMP_PDB_SUCCESS;

  /* Init GTK. Yuck! */
  {
    int argc = 1;
    char **argv = g_new(char*,2);
    argv[0] = "homogenizer";
    argv[1] = 0;
    gtk_init(&argc, &argv);
    gtk_rc_parse(gimp_gtkrc());
  }

  /* Get drawable */
  drawable = gimp_drawable_get(param[2].data.d_drawable);

  if (!gimp_drawable_is_rgb(drawable->id) &&
      !gimp_drawable_is_gray(drawable->id)) {
    gimp_drawable_detach(drawable);
    values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    return;
  }

  /* Deal with run mode (parameter zero) */
  
  cancelled = false;
  switch(param[0].data.d_int32) {
    case GIMP_RUN_INTERACTIVE :
      getLastParameters(&parameters); 
      cancelled = getParametersByAsking(&parameters,drawable->id); 
      break;
    case GIMP_RUN_NONINTERACTIVE :
      cancelled = getParametersFromList(&parameters, nparams, param); 
      break;
    case GIMP_RUN_WITH_LAST_VALS :
      cancelled = getLastParameters(&parameters); 
      break;
  }

  if (cancelled) {
    gimp_drawable_detach(drawable);
    values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    return;
  }

  /* Fetch the image to be warped, munged, crunched, twisted... */
  gimp_drawable_mask_bounds(drawable->id,&x1,&y1,&x2,&y2);
  width = x2-x1;
  height = y2-y1;
  
  /* Do it */
  gimp_progress_init("Homogenize");
  maxProgress = width*height*2;
  progress = 0;
  
  gimp_pixel_rgn_init(&srcRegion, drawable,x1,y1,width,height,FALSE,FALSE);

  xStep = 3.14159265 / width;
  yStep = 3.14159265 / height;

  xCosTable = g_new(double*,parameters.xOrder);
  yCosTable = g_new(double*,parameters.yOrder);
  
  for(i=0;i<parameters.xOrder;i++) {
    xCosTable[i] = g_new(double,width);
    for(x=0;x<width;x++)
      xCosTable[i][x] = cos((x+0.5)*i*xStep);  
  }
  
  for(i=0;i<parameters.yOrder;i++) {
    yCosTable[i] = g_new(double,height);
    for(y=0;y<height;y++)
      yCosTable[i][y] = cos((y+0.5)*i*yStep);  
  }

  weights = g_new(double,parameters.xOrder * parameters.yOrder * srcRegion.bpp);
  for(i=0;i<parameters.xOrder*parameters.yOrder*srcRegion.bpp;i++)
    weights[i] = 0;

  for(pr = gimp_pixel_rgns_register(1,&srcRegion);
      pr != NULL;
      pr = gimp_pixel_rgns_process(pr)) {
    for(i=0,y=srcRegion.y;y<srcRegion.y+srcRegion.h;y++) {
      for(x=srcRegion.x;x<srcRegion.x+srcRegion.w;x++) {
        for(color=0;color<srcRegion.bpp;color++,i++) {
	   double value = srcRegion.data[i];

	   for(j=color,yy=0;yy<parameters.yOrder;yy++)
	     for(xx=0;xx<parameters.xOrder;xx++,j+=srcRegion.bpp)
               if (j != color)
	         weights[j] += value * xCosTable[xx][x] * yCosTable[yy][y];
	}
      }
    }

    progress += srcRegion.w*srcRegion.h;
    gimp_progress_update((double)(progress)/maxProgress);
  }
  
  for(j=0,yy=0;yy<parameters.yOrder;yy++)
    for(xx=0;xx<parameters.xOrder;xx++)
      for(k=0;k<srcRegion.bpp;k++,j++)
        weights[j] /= width*height/4.0*
          (xx == 0 ? 2.0 : 1.0) *
	  (yy == 0 ? 2.0 : 1.0);

  gimp_pixel_rgn_init(&srcRegion, drawable,x1,y1,width,height,FALSE,FALSE);
  gimp_pixel_rgn_init(&destRegion,drawable,x1,y1,width,height,TRUE,TRUE);

  for(pr = gimp_pixel_rgns_register(2,&srcRegion,&destRegion);
      pr != NULL;
      pr = gimp_pixel_rgns_process(pr)) {
    for(i=0,y=srcRegion.y;y<srcRegion.y+srcRegion.h;y++) {
      for(x=srcRegion.x;x<srcRegion.x+srcRegion.w;x++) {
        for(color=0;color<srcRegion.bpp;color++,i++) {
	   double value = srcRegion.data[i];

	   for(j=color,yy=0;yy<parameters.yOrder;yy++)
	     for(xx=0;xx<parameters.xOrder;xx++,j+=srcRegion.bpp)
               if (j != color)
	         value -= weights[j] * xCosTable[xx][x] * yCosTable[yy][y];

	   if (value < 0.0)
	     destRegion.data[i] = 0;
	   else if (value > 255.0)
	     destRegion.data[i] = 255;
	   else
	     destRegion.data[i] = (guchar)value;
	}
      }
    }

    progress += srcRegion.w*srcRegion.h;
    gimp_progress_update((double)(progress)/maxProgress);
  }

  g_free(weights);
  
  for(i=0;i<parameters.xOrder;i++)
    g_free(xCosTable[i]);
  g_free(xCosTable);
  
  for(i=0;i<parameters.yOrder;i++)
    g_free(yCosTable[i]);
  g_free(yCosTable);

  /* Voodoo to update actual image */
  gimp_drawable_flush(drawable);
  gimp_drawable_merge_shadow(drawable->id,TRUE);
  gimp_drawable_update(drawable->id,x1,y1,width,height);

  /* Store parameters deep in the bowels of the GIMP */
  gimp_set_data("plug_in_homogenizer", &parameters, sizeof(parameters));
  
  /* Unget drawable */
  gimp_drawable_detach(drawable);
  
  gimp_displays_flush();
}

