/*********************************************************************************/
/* LIC 0.50 -- image filter plug-in for The Gimp program                         */
/* Copyright (C) 2001 Frank Lauterwald                                           */
/*===============================================================================*/
/* E-mail: frankl@gmx.net                                                        */
/* You can contact the original The Gimp authors at gimp@xcf.berkeley.edu        */
/*===============================================================================*/
/* This program is free software; you can redistribute it and/or modify it under */
/* the terms of the GNU General Public License as published by the Free Software */
/* Foundation; either version 2 of the License, or (at your option) any later    */
/* version.                                                                      */
/*===============================================================================*/
/* This program is distributed in the hope that it will be useful, but WITHOUT   */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS */
/* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.*/
/*===============================================================================*/
/* You should have received a copy of the GNU General Public License along with  */
/* this program; if not, write to the Free Software Foundation, Inc., 675 Mass   */
/* Ave, Cambridge, MA 02139, USA.                                                */
/*===============================================================================*/
/* In other words, you can't sue me for whatever happens while using this ;)     */
/*********************************************************************************/
/* CHANGELOG									 */
/* Date		Version		Comment						 */
/* 2001 Oct. 1,    0.30		first "release"					 */
/* 2002 Jan.26,    0.48		fixed bugs in smoothbox and astroid-filters	 */
/* 2002 Jan 26	   0.50		just "tagged" for 0.50beta 			 */
/*********************************************************************************/

/* This File implements two filter-functions for the LIC-Effect */
/* These are not what The Gimp calls a filter. They are just functions needed for the algorithm */



#include "lic_filter.h"
#include "../lic_globals.h"
#include <libgimp/gimpmath.h>

gfloat filter_blockfilter(gfloat x)
{ gdouble temp = fabs(x);
  if (temp>parameters.filterlength) { return 0; }
  return 1;
}

gfloat filter_blockfilter_integral(gfloat x1, gfloat x2)
{ gdouble f = parameters.filterlength;
  if (x1>f) { x1=f; }
  if (x2>f) { x2=f; }
  if (x1<-f) { x1=-f; }
  if (x2<-f) { x2=-f; }
  return x2-x1;
}

/* a box filter in the middle falling close to the edges */
gfloat filter_smoothblockfilter(gfloat x)
{ gdouble temp = fabs(x);
  if (temp>parameters.filterlength) { return 0; }
  if (temp>parameters.filterlength*FILTERFUNCTION_SMOOTHBOX_STARTFALLOFF)
  	{ return (1-((temp/parameters.filterlength)-FILTERFUNCTION_SMOOTHBOX_STARTFALLOFF)/(1-FILTERFUNCTION_SMOOTHBOX_STARTFALLOFF)); }
  return 1;
}

/* computes the integral from 0 to (positive) x */
gfloat filter_smoothblockfilter_integral_helper(gfloat x)
{  if (x <= FILTERFUNCTION_SMOOTHBOX_STARTFALLOFF*parameters.filterlength)
	{ return x; }
   else
   	{ gfloat temp = (x-(FILTERFUNCTION_SMOOTHBOX_STARTFALLOFF*parameters.filterlength));
	  return (FILTERFUNCTION_SMOOTHBOX_STARTFALLOFF*parameters.filterlength) + temp/2; }
}

gfloat filter_smoothblockfilter_integral(gfloat x1, gfloat x2)
{	gfloat result1, result2;
  	gfloat ax1 = fabs(x1);
	gfloat ax2 = fabs(x2);
	if (ax1>parameters.filterlength) { ax1=parameters.filterlength; }
	if (ax2>parameters.filterlength) { ax2=parameters.filterlength; }
	result1 = filter_smoothblockfilter_integral_helper(ax1);
	result2 = filter_smoothblockfilter_integral_helper(ax2);
	if ((x1*x2)>0) /* meaning they have the same sign */
	{ 	return fabs(result2-result1); }
	else
	{ 	return (result2+result1); }
}


gfloat filter_astroidfilter(gfloat x)
{ gdouble temp = fabs(x);
  if (temp>parameters.filterlength) { return 0; }
  return (1-sin(temp/parameters.filterlength*G_PI/2));
}

/* computes the integral from 0 to (positive) x */
gfloat filter_astroidfilter_integral_helper(gfloat x)
{ 	return (x-( cos(x/parameters.filterlength*G_PI/2)*parameters.filterlength*G_PI/2)); }

gfloat filter_astroidfilter_integral(gfloat x1, gfloat x2)
{	gfloat result1, result2;
  	gfloat ax1 = fabs(x1);
	gfloat ax2 = fabs(x2);
	if (ax1>parameters.filterlength) { ax1=parameters.filterlength; }
	if (ax2>parameters.filterlength) { ax2=parameters.filterlength; }
	result1 = filter_astroidfilter_integral_helper(ax1);
	result2 = filter_astroidfilter_integral_helper(ax2);
	if ((x1*x2)>0) /* meaning they have the same sign */
	{ 	return fabs(result2-result1); }
	else
	{ 	return (result2+result1); }
}


gfloat filter_trianglefilter(gfloat x)
{ gdouble temp = fabs(x);
  if (temp>parameters.filterlength) { return 0; }
  return (1-temp/parameters.filterlength);
}

gfloat filter_trianglefilter_integral(gfloat x1, gfloat x2)
{ gdouble f;
  if (x1*x2 < 0) /* this means integration is crossing the middle of the filter -> calculate in two steps */
    { return filter_trianglefilter_integral(x1,0)+filter_trianglefilter_integral(0,x2); }
  f = parameters.filterlength;
  if (x1>f) { x1=f; }
  if (x2>f) { x2=f; }
  if (x1<-f) { x1=-f; }
  if (x2<-f) { x2=-f; }
  return ( (x2-x1)-(x2*x2-x1*x1)/(2*parameters.filterlength) ) ;
}


gfloat filter_otherfilter(gfloat x)
{	gfloat whichpoint;
	gfloat part;
	if (fabs(x)>parameters.filterlength) return 0;
	// which point shall be returned (as a gfloat: e.g. 3.2 means interpolate between point 3 and 4)
	whichpoint = ((x/parameters.filterlength)*FILTERCURVE_NUMBER_OF_POINTS)/2+FILTERCURVE_NUMBER_OF_POINTS/2;
	part=whichpoint-(gint)whichpoint;
	if (part!=0) {
		return  variables.otherfilter_points[(gint)whichpoint]*(1-part)+
			variables.otherfilter_points[(gint)whichpoint+1]*part;
	} else {
		return variables.otherfilter_points[(gint)whichpoint];
	}
	return 0;
}
