/*********************************************************************************/
/* LIC 0.50 -- image filter plug-in for The Gimp program                         */
/* Copyright (C) 2001/02 Frank Lauterwald                                        */
/*===============================================================================*/
/* E-mail: frankl@gmx.net                                                        */
/* You can contact the original The Gimp authors at gimp@xcf.berkeley.edu        */
/*===============================================================================*/
/* This program is free software; you can redistribute it and/or modify it under */
/* the terms of the GNU General Public License as published by the Free Software */
/* Foundation; either version 2 of the License, or (at your option) any later    */
/* version.                                                                      */
/*===============================================================================*/
/* This program is distributed in the hope that it will be useful, but WITHOUT   */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS */
/* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.*/
/*===============================================================================*/
/* You should have received a copy of the GNU General Public License along with  */
/* this program; if not, write to the Free Software Foundation, Inc., 675 Mass   */
/* Ave, Cambridge, MA 02139, USA.                                                */
/*===============================================================================*/
/* In other words, you can't sue me for whatever happens while using this ;)     */
/*********************************************************************************/
/* CHANGELOG									 */
/* Date		Version		Comment						 */
/* 2001 Oct. 1,    0.30		first "release"					 */
/* 2002 Jan. 26,   0.50		release of first beta version			 */
/*********************************************************************************/

/* this file includes color conversion routines
// optimization could be done but I doubt this will seriously increase overall speed
// consider this: as soon as you see the progress bar of the filter all color conversion
// has already been done. This is usually the case after very few seconds		*/

/* all functions have one void pointer as  parameter. They shall interpret this void pointe
//   as whatever they need. Thus if a function can convert rgbpixels to a guchar it shall
//   interpret the parameter as pointer to a rgbpixel
// this makes typechecking more difficult but allows for a trick: the functions are all
//   "instances" of color_conversion_function (see lic_color.h). Thus we can code things
//   like this: color_conversion_function c = { MAYBE A SWITCH() HERE TO DECIDE WHICH
//   FUNCTION TO USE }
//   result = c(&something)
// This is actually done by lic_vectors.c which always passes the address of the current pixel
//   without having to care about if it is greyscale or rgb or whatsoever.			*/

#include "lic_color.h"
#include <libgimp/gimp.h>

const rgbfloatpixel RGBFLOATPIXEL_BLACK={0.0, 0.0, 0.0};
const grayfloatpixel GRAYFLOATPIXEL_BLACK={0.0};

/* functions have the naming scheme: get_xxx_yyy(void* p)
//   where xxx is the input color model e.g. rgb or rgba or grey
//   and yyy is the name of the result channel e.g. hue */

gfloat minimum3(gfloat x,gfloat y,gfloat z)
{ if (x>y)
	{ if (z>y)
		{ return y; }
	  else
	  	{ return z; }
	}
  else
  	{ if (z>x)
		{ return x; }
	  else
	  	{ return z; }
	}
}

gfloat maximum3(gfloat x,gfloat y,gfloat z)
{ if (x>y)
	{ if (x>z)
		{ return x; }
	  else
	  	{ return z; }
	}
  else
  	{ if (y>z)
		{ return y; }
	  else
	  	{ return z; }
	}
}


gfloat get_rgb_red(void *p)
{  return ((rgbpixel*)(p))->r; }

gfloat get_rgb_green(void *p)
{  return ((rgbpixel*)(p))->g; }

gfloat get_rgb_blue(void *p)
{  return ((rgbpixel*)(p))->b; }

gfloat get_rgb_brightness(void *p)
{  rgbpixel col=*((rgbpixel*) p);
   guint min = minimum3(col.r, col.g, col.b);
   guint max = maximum3(col.r, col.g, col.b);
   return (max+min)/2;
}

gfloat get_rgb_saturation(void* p)
{  rgbpixel col=*((rgbpixel*) p);
   guint min = minimum3(col.r, col.g, col.b);
   guint max = maximum3(col.r, col.g, col.b);
   if (min==max) return 0;
   if (max+min>256)
      { return 256*(max-min)/(512-max-min); }
   else
      { return 256*(max-min)/(max+min); }
}

gfloat get_rgb_hue(void* p)
{  gfloat delta, temp;
   rgbpixel col=*((rgbpixel*) p);
   gfloat min = minimum3(col.r, col.g, col.b);
   gfloat max = maximum3(col.r, col.g, col.b);

   if (max==min) temp=-1.0;
    else
      {
        delta=(max-min);
        if (col.r==max)
          temp=(col.g-col.b)/delta;
        else if (col.g==max)
          temp=2.0+(col.b-col.r)/delta;
        else if (col.b==max)
          temp=4.0+(col.r-col.g)/delta;

        temp=temp*60.0;
        if (temp<0.0)
          temp=temp+360.0;
      }
    /* now temp is the hue-angle between 0 and 360 -> map to 0..255 and return	*/
    return (temp*256)/360;
}

gfloat get_gray_brightness(void* v)
{ return ((graypixel*) v)->b; }

gfloat get_gray_saturation(void* v)
{ return 0; }

gfloat get_gray_hue(void* v)
{ return 0; }

