/*********************************************************************************/
/* LIC 0.50 -- image filter plug-in for The Gimp program                         */
/* Copyright (C) 2001/02 Frank Lauterwald                                        */
/*===============================================================================*/
/* E-mail: frankl@gmx.net                                                        */
/* You can contact the original The Gimp authors at gimp@xcf.berkeley.edu        */
/*===============================================================================*/
/* This program is free software; you can redistribute it and/or modify it under */
/* the terms of the GNU General Public License as published by the Free Software */
/* Foundation; either version 2 of the License, or (at your option) any later    */
/* version.                                                                      */
/*===============================================================================*/
/* This program is distributed in the hope that it will be useful, but WITHOUT   */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS */
/* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.*/
/*===============================================================================*/
/* You should have received a copy of the GNU General Public License along with  */
/* this program; if not, write to the Free Software Foundation, Inc., 675 Mass   */
/* Ave, Cambridge, MA 02139, USA.                                                */
/*===============================================================================*/
/* In other words, you can't sue me for whatever happens while using this ;)     */
/*********************************************************************************/
/* CHANGELOG									 */
/* Date		Version		Comment						 */
/* 2001 Oct. 1,    0.30		first "release"					 */
/* 2001 Jan 13	   0.32		removed the #define profiling and debug and	 */
/*				put them into the Makefile			 */
/*		 		Added the ERR, VOUT and TOUT macros		 */
/* 2002 Jan. 26,   0.50		release of first beta version			 */
/*********************************************************************************/

#ifndef __lic_globals__
#define __lic_globals__

/* this file defines global variables, constants and  a few structs */

#define ERR(A) {cerr << "ERROR in " << __FILE__ << " at line " << __LINE__ << ": " << A << endl;}

#if defined(debug) || defined(VERBOSE)
#define VOUT(A) {fprintf(stderr, A);}
#else
#define VOUT(A)
#endif

#ifdef debug
#define TOUT(A) {fprintf(stderr,"%12s(%4d)#> ",__FILE__,__LINE__); fprintf (stderr, A);}
#else
#define TOUT(A)
#endif

//#ifndef M_PI
//#define M_PI 3.141592653587
//#endif

#include "lic_vectors.h"
#include <glib.h>
#include <libgimp/gimp.h>


#define SMALLNUMBER 	0.0005  /* for cabral a "small number" is added to stepsize. This one seems appropriate to me */
#define INFINITY	10e20  /* not really infinity but large enough for our purposes (used in lic_algorithm_include.c, cabral algorithm) */
#define MINSTEPSIZE     0.001   /* for Cabral. If the steps get too small we terminate so we dont have to make millions of steps 	    */

/* note: as the userinterface is written by glade, these constants are not used there,
   so if you make changes here you will have to make them in the ui manually too  */
#define FILTERFUNCTION_BOX 		0
#define FILTERFUNCTION_SMOOTHBOX        1
#define FILTERFUNCTION_TRIANGLE		2
#define FILTERFUNCTION_ASTROID		3
#define FILTERFUNCTION_OTHER		4

#define ALGORITHM_DDAEULER		0
#define ALGORITHM_DDACABRAL		1
#define ALGORITHM_LICEULER		2
#define ALGORITHM_LICCABRAL		3

#define EFFECTCHANNEL_HUE		0
#define EFFECTCHANNEL_SATURATION	1
#define EFFECTCHANNEL_BRIGHTNESS	2
#define EFFECTCHANNEL_RED		3
#define EFFECTCHANNEL_GREEN		4
#define EFFECTCHANNEL_BLUE		5
#define EFFECTCHANNEL_ALPHA		6

#define GETTEXTUREFROM_IMAGE		0
#define GETTEXTUREFROM_NOISE		1

#define C_DEFAULTSTEPSIZE 		0.25
#define C_DEFAULTALGORITHM		ALGORITHM_LICCABRAL
#define C_DEFAULTFILTERLENGTH		8
#define C_DEFAULTEFFECTCHANNEL		EFFECTCHANNEL_BRIGHTNESS
#define C_DEFAULTROTATEVECTORS		TRUE
#define C_DEFAULTFILTERFUNCTION		FILTERFUNCTION_BOX

#define FILTERFUNCTION_SMOOTHBOX_STARTFALLOFF 0.6
#define FILTERCURVE_NUMBER_OF_POINTS 256

#define BLOCKSIZE 	512
#define MAXFILTERLEN     64

#define CLICK_OK	  0
#define CLICK_CANCEL	  1



typedef gfloat(*filterfunction)(gfloat);

typedef gfloat(*filterfunctionintegrate)(gfloat, gfloat);

/** structs with algorithm-specific options */
typedef struct {
	gfloat stepsize;
} lic_parameters_algorithm_dda_euler_struct;

typedef struct {
	gfloat stepsize;
} lic_parameters_algorithm_lic_euler_struct;
/** -- structs with algorithm-specific options */


/** extra unions for algorithms and filterfunctions */
typedef  union {
	lic_parameters_algorithm_dda_euler_struct ddaEuler;
	lic_parameters_algorithm_lic_euler_struct licEuler;
 	void *ddaCabral;
        void *licCabral;
} lic_parameters_algorithm_extra_struct;

typedef union {
	void *nothing;
} lic_parameters_filterfunction_extra_struct;
/** -- extra unions for algorithms and filterfunctions */


/** parts of the main struct */
typedef struct {
	guint16 algorithm;
        lic_parameters_algorithm_extra_struct extras;
} lic_parameters_algorithm_struct;

typedef struct {
	guint16 filterfunction;
        lic_parameters_filterfunction_extra_struct extras;
} lic_parameters_filterfunction_struct;

typedef struct {
	guint16 numx, numy; // the size of the pseudo-vector-field
	gfloat  magnitude, minv, maxv;
	gfloat  dx, dy;
} lic_parameters_noise_struct;

/** -- parts of the main struct */


/** The options-struct */
typedef struct {
        guint16  effectchannel;
        gboolean createNewImage;
 	GimpDrawable *sourceImg, *destImg;
        guint16	getTextureFrom;
        gboolean rotateVectors;
	gint32  textureImageDrawable, vectorImageDrawable, resultImageDrawable;
} lic_parameters_options_struct;
/** -- The options-struct */


/** the main struct itself */
typedef struct {
        gfloat   filterlength;
        lic_parameters_algorithm_struct      algorithm;
        lic_parameters_filterfunction_struct filterfunction;
        lic_parameters_options_struct	     options;
	lic_parameters_noise_struct	     noiseoptions;
} lic_parameters_struct;
/** -- the main struct itself */

extern lic_parameters_struct parameters;

/** for __variables__ as opposed to parameters there is an extra structure */
typedef struct {
	gint32 			  width,height;	/* width and height							*/
	lic_floatvector_field	  floatvectors;	/* holds 64k vectors of length 1 and angle i/64k*2*pi 			*/
	lic_vector_by_index_field *vectors;	/* holds the correct number of indices into the floatvector field	*/

	/* for ui */
	gint ok_or_cancel;  /* if ok or cancel was clicked in the userinterface		*/

	gboolean closeTextureImageAfterCalculation; /* if this is true the texture image will be closed at the on of the calculation		*/
						/* this is mainly intended for the case when the texture image was automatically created	*/
	gint32 textureImageDisplay;		/* the only way I know to delete an image is deleting all if its displays			*/
						/*   thus if we want to delete our texture image we have to store its display somewhere		*/

	/* for reading and writing pixels */
	guchar *inputpixelregiondata;         /* pointer to the block of input pixel data			*/
	guchar *outputpixelregiondata;        /* pointer to the block of output pixel data			*/
	gint32 output_swappedin_x1, output_swappedin_y1, output_swappedin_x2, output_swappedin_y2; /* the coordinates of the currently swapped-in output window		*/
	gint32 input_swappedin_x1, input_swappedin_y1, input_swappedin_x2, input_swappedin_y2; /* the coordinates of the currently swapped-in input window		*/
	GimpPixelRgn output_region;
	GimpPixelRgn input_region;

	/* for the filter-function */
	filterfunction 		filter;
	filterfunctionintegrate	filter_integrate;
	/* information that may be needed for the blockfilter */
	gfloat  otherfilter_points[FILTERCURVE_NUMBER_OF_POINTS];

	/* information about the input-texture block just swapped in */
	gint32 inputblockwidth, inputblockheight;
	gint32 outputblockwidth, outputblockheight;



} lic_variables_struct;


extern lic_variables_struct variables;

#endif
