/*********************************************************************************/
/* LIC 0.50 -- image filter plug-in for The Gimp program                         */
/* Copyright (C) 2001/02 Frank Lauterwald                                        */
/*===============================================================================*/
/* E-mail: frankl@gmx.net                                                        */
/* You can contact the original The Gimp authors at gimp@xcf.berkeley.edu        */
/*===============================================================================*/
/* This program is free software; you can redistribute it and/or modify it under */
/* the terms of the GNU General Public License as published by the Free Software */
/* Foundation; either version 2 of the License, or (at your option) any later    */
/* version.                                                                      */
/*===============================================================================*/
/* This program is distributed in the hope that it will be useful, but WITHOUT   */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS */
/* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.*/
/*===============================================================================*/
/* You should have received a copy of the GNU General Public License along with  */
/* this program; if not, write to the Free Software Foundation, Inc., 675 Mass   */
/* Ave, Cambridge, MA 02139, USA.                                                */
/*===============================================================================*/
/* In other words, you can't sue me for whatever happens while using this ;)     */
/*********************************************************************************/
/* CHANGELOG									 */
/* Date		Version		Comment						 */
/* 2001 Oct. 1,    0.30		first "release"					 */
/* 2002 Jan. 26,   0.50		release of first beta version			 */
/*********************************************************************************/

/*********************************************************/
/* This file contains the functions			 */
/* lic_gimp_interactive(gimp_drawable)			 */
/* lic_compute_image()					 */
/* 							 */
/* These functions are called by run() in lic_plugin.c   */
/* which itself is called by The Gimp			 */
/*********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <libgimp/gimp.h>
#include "userinterface/lic_userinterface.h"
#include "lic_globals.h"
#include "lic_vectors.h"
#include "lic_main.h"
#include "algorithm/lic_algorithm.h"



/* initializes parameters with default values */
void lic_set_default_values()
{	parameters.options.getTextureFrom	= GETTEXTUREFROM_IMAGE;
	parameters.options.effectchannel	= EFFECTCHANNEL_BRIGHTNESS;
	parameters.options.createNewImage	= TRUE;
	parameters.options.rotateVectors	=C_DEFAULTROTATEVECTORS;
	parameters.filterlength			=C_DEFAULTFILTERLENGTH;
	parameters.options.effectchannel	=C_DEFAULTEFFECTCHANNEL;
	parameters.algorithm.extras.licEuler.stepsize 	= C_DEFAULTSTEPSIZE;
	parameters.filterfunction.filterfunction	= C_DEFAULTFILTERFUNCTION;
	parameters.algorithm.algorithm 			= C_DEFAULTALGORITHM;

	parameters.noiseoptions.numx = 40;
	parameters.noiseoptions.numy = 40;
	parameters.noiseoptions.dx   = 2.0;
	parameters.noiseoptions.dy   = 2.0;
	parameters.noiseoptions.minv = -25;
	parameters.noiseoptions.maxv =  25;
	parameters.noiseoptions.magnitude = 2.0;

}

void lic_print_parameter_summary()
{	fprintf(stderr,"== Parameter summary: ==\n");
	fprintf(stderr, "size of images: %i x %i \n",variables.width, variables.height);
	fprintf(stderr, "filterlength:   %f \n", parameters.filterlength);
	fprintf(stderr, "filterfunction: %i \n", parameters.filterfunction.filterfunction);
	fprintf(stderr, "stepsize:   %f \n", parameters.algorithm.extras.ddaEuler.stepsize);
	fprintf(stderr, "algorithm: %i \n", parameters.algorithm.algorithm);
	fprintf(stderr, "rotate vecs: %i \n", parameters.options.rotateVectors);
	fprintf(stderr, "effectchannel: %i \n", parameters.options.effectchannel);
}


void lic_gimp_interactive(GimpDrawable sourceDrawable)
{

	parameters.options.textureImageDrawable = sourceDrawable.id;
	parameters.options.vectorImageDrawable = sourceDrawable.id; /* just to be sure that one is selected */

	lic_set_default_values();
	#ifdef debug
		fprintf(stderr, "Now showing userinterface \n");
	#endif
	show_userinterface();
	if (variables.ok_or_cancel == CLICK_OK)
	{
		variables.width = gimp_drawable_get(parameters.options.vectorImageDrawable)->width;
		variables.height = gimp_drawable_get(parameters.options.vectorImageDrawable)->height;

  		lic_do_work_after_parameters_set();
	}
	else
	{	fprintf(stderr, "Action Cancelled \n"); }
}




/* This functions does everything needed, provided all parameters hav been set correctly before */
/* It does so by calling a few other functions woth certain high-level tasks */
void lic_do_work_after_parameters_set()
{	gint32	 	newdisplay;
	gint32		newlayer;
	gint32		imageType, layerType;

	/* NOTE: PARAMETERS.OPTIONS.VECTORIMAGEID !!MUST!! ALREADY BE SET HERE */

	/* resultImageDrawable defaults to the image that was clicked. Until here this is stored in textureImageDrawable */
	/* textureImageDrawable may however be changed later								 */
	parameters.options.resultImageDrawable=parameters.options.textureImageDrawable;	 /* default. This is changed if "new image" ist set */

	/* firstly get a texture image
		// this includes creating noise if needed, scaling of a selected texture image and more */
	lic_create_texture_image();

	/* secondly create the vectors */
	lic_create_floatvector_field(&variables.floatvectors);
	variables.vectors = lic_create_vectors(parameters.options.vectorImageDrawable);


	/* thirdly create a new result image if this is wanted */
	switch(gimp_drawable_get(parameters.options.textureImageDrawable)->bpp)
	{ 	case 1: imageType = GIMP_GRAY; break;
		case 3: imageType = GIMP_RGB;  break;
		default:fprintf(stderr, "lic_main.c line %i: input image drawable: %i bpp not supported. THIS SHOULDNT HAPPEN. Aborting \n",__LINE__,gimp_drawable_get(parameters.options.textureImageDrawable)->bpp); exit(EXIT_FAILURE); break;
	}

	fprintf(stderr,"size of texture image: %i %i \n",gimp_drawable_get(parameters.options.textureImageDrawable)->width,gimp_drawable_get(parameters.options.textureImageDrawable)->height);

	if (parameters.options.createNewImage)
	{	gint32 image;
		if (!(image=gimp_image_new(variables.width,variables.height,imageType)))
		{ fprintf(stderr,"creating new result image: gimp_image_new() failed\n"); exit(EXIT_FAILURE);
		}
		switch(gimp_drawable_get(parameters.options.textureImageDrawable)->bpp)
		{ 	case 1: layerType = GIMP_GRAY_IMAGE; break;
			case 3: layerType = GIMP_RGB_IMAGE;  break;
			default:fprintf(stderr, "lic_main.c line %i: input image drawable: bpp not supported. THIS SHOULDNT HAPPEN. Aborting \n",__LINE__); exit(EXIT_FAILURE); break;
		}

		if (!(newlayer=gimp_layer_new(image, "Background", variables.width, variables.height, layerType, 100.0, GIMP_NORMAL_MODE)))
		  { fprintf(stderr,"creating new image: gimp_layer_new() failed\n"); exit(EXIT_FAILURE); }
		if (!gimp_image_add_layer(image, newlayer, -1))
		  { fprintf(stderr,"creating new image: gimp_image_add_layer() failed\n"); exit(EXIT_FAILURE); }
		parameters.options.resultImageDrawable = gimp_image_active_drawable(image);

	}
	else
	{	fprintf(stderr, "using old image as result \n");
	}

	/* now everything is ready -> get going*/
	lic_print_parameter_summary();
	lic_compute_image();


	/* get the new picture onto the screen */
	gimp_drawable_flush(gimp_drawable_get(parameters.options.resultImageDrawable));
	gimp_drawable_merge_shadow ((parameters.options.resultImageDrawable), FALSE);
	gimp_drawable_flush(gimp_drawable_get(parameters.options.resultImageDrawable));
	gimp_drawable_update(parameters.options.resultImageDrawable, 0, 0, variables.width, variables.height);
	gimp_displays_flush();
	gimp_drawable_detach(gimp_drawable_get(parameters.options.resultImageDrawable));
	/* a new display is only created if we created a new image. Otherwise it must already be displayed */
	if (parameters.options.createNewImage==TRUE) { newdisplay = gimp_display_new(gimp_drawable_image(parameters.options.resultImageDrawable)); }

	/* eventually close the old texture image */
	if (variables.closeTextureImageAfterCalculation==TRUE)
	{	gimp_display_delete(variables.textureImageDisplay);
	}

}



/* just two small macros */
#define max(a,b) (a>b?a:b)
#define min(a,b) (a<b?a:b)

/***********************************************************/
/* This function takes care of having a texture image of   */
/* the appropriate size					   */
/***********************************************************/
void lic_create_texture_image()
{	gint32 newlayer;
	gint32 newdisplay;
	GimpDrawable *d1, *d2;

	variables.closeTextureImageAfterCalculation=FALSE;
		VOUT("getting texture from image ... checking if texture image is large enough ");
		/* the texture has to be at least as large as the vector field
		//    cabral says (in other words i think): all vectors are remapped to fall into the input texture
		//    This is kind of a virtual "tiling" (in the meaning as in tiled desktop wallpapers)
		//    This remapping takes time (though not much: two % operations for getting a texel) and makes things
		//	a lot more complicated due to the following fact:
		//	We do not "map in" the whole texture but only a block of it. Consider input texture=180x180 large,
		//      size of mapped-in block is 100x100. Consider the currently mapped-in block is (100,0)-(200,100).
		//	When you now try to access e.g. (185,30) this would have to be remapped to (5,30) but thal lies
		//	outside of our mapped-in block. Thus we would have to get a new block just for reading a single pixel!
		//    The solution I took is this: Let the remapping be done by GIMP automatically: we create a new image
		//	which is large enough. Then we take the input texture and copy it into the new image until it is full
		//	of input textures. Then we take the new image as texture for further calculation. Obviously, no
		//	remapping is needed 										*/

		/* size of vector image is already to be found in variables.width, variables.height 	*/
		d1 = gimp_drawable_get(parameters.options.textureImageDrawable);

		if ((variables.width>d1->width) || (variables.height>d1->height))
		{	/* proceed as described */
			gint32 tempImageID;
			gint32 x,y;
			gint32 imageType; /* the colormodel of the new image. e.g. GIMP_RGB */
			GimpPixelRgn rgn1, rgn2;
			gpointer pr;
			/* determine the type of the image to be created */
			switch(d1->bpp)
			{ case 1: imageType = GIMP_GRAY; break;
			  case 3: imageType = GIMP_RGB; break;
			  default: fprintf(stderr, "input image format is not supported. This shouldnt happen. Aborting \n"); exit(EXIT_FAILURE);
			}
			if (!(tempImageID=gimp_image_new(variables.width,variables.height,imageType)))
		  	{ fprintf(stderr,"creating texture image from image: gimp_image_new() failed\n"); exit(EXIT_FAILURE); }
			/* add a new default layer */
			x = d1->bpp;
			switch(x)
			{ case 1: y = GIMP_GRAY_IMAGE; break;
			  case 2: y = GIMP_GRAYA_IMAGE; break;
			  case 3: y = GIMP_RGB_IMAGE; break;
			  case 4: y = GIMP_RGBA_IMAGE; break;
			}
			if (!(newlayer=gimp_layer_new(tempImageID, "Background", variables.width, variables.height, y, 100.0, GIMP_NORMAL_MODE)))
		  	{ fprintf(stderr,"creating texture image from image: gimp_layer_new() failed\n"); exit(EXIT_FAILURE); }
			if (!gimp_image_add_layer(tempImageID, newlayer, -1))
		  	{ fprintf(stderr,"creating texture image from image: gimp_image_add_layer() failed\n"); exit(EXIT_FAILURE); }
			d2 = gimp_drawable_get(gimp_image_active_drawable(tempImageID));
			VOUT("image and layer created \n");
			/* now we have the new image -> make copies of the old one and place them here */
			gimp_pixel_rgn_init(&rgn2, d2, 0, 0, d2->width, d2->height, TRUE, FALSE);
			for (y=0; y<variables.height; y+=d1->height)
				{  for (x=0; x<variables.width; x+=d1->width)
					{  	/* do this the clean way: get a pixel region containing all the source image
						// then iterate over the tiles as in the tuturial		*/
						gimp_pixel_rgn_init(&rgn1, d1, 0, 0, d1->width, d1->height, FALSE, FALSE);
						for (pr = gimp_pixel_rgns_register (1, &rgn1);pr != NULL; pr = gimp_pixel_rgns_process (pr))
						{  /* copy the tile */
							gint32 x1, w1, y1, h1;
							x1=rgn1.x+x; y1=rgn1.y+y; w1=rgn1.w; h1=rgn1.h; /* perform check if / how much of block would fit into destination	*/
							if ((x1+w1<variables.width) && (y1+h1<variables.height)) /* block is completely inside */
							{   gimp_pixel_rgn_set_rect(&rgn2, rgn1.data, x1, y1, w1, h1); }
							else
							{ /* this is more complicated: we have to get exeactly the block that would still fit	*/
							  /* first see if any of this block will be inside					*/
							    if ((x1<variables.width) && (y1<variables.height))
							    {	guchar* buf;
							    	/* a part has to be copied. just calculate w1 and h1 newly according to hwo large it is */
								w1=min(variables.width-x1,w1);
								h1=min(variables.height-y1,h1);
								/* now get exactly the block we want	*/
								if (!(buf=malloc(w1*h1*4))) /* get 4 bytes for pixel - that should be enough		*/
								{	fprintf(stderr, "Cannot allocate %i bytes in lic_create_texture_image \n", w1*h1*4);
									exit(EXIT_FAILURE);
								}
								gimp_pixel_rgn_get_rect(&rgn1, buf,rgn1.x, rgn1.y, w1, h1);
								gimp_pixel_rgn_set_rect(&rgn2, buf, x1, y1, w1, h1);
								free(buf);
							    }
							}
						}
					}
				}
			gimp_drawable_update(d2->id, 0, 0, d2->width, d2->height);
			/*gimp_drawable_merge_shadow (gimp_drawable_image(d2->id), TRUE);*/

			fprintf(stderr, "creating new display \n");
			newdisplay = gimp_display_new(gimp_drawable_image(d2->id));


			gimp_displays_flush();
			gimp_drawable_detach(d1);
			gimp_drawable_detach(d2);
			/* make the new image the texture image for the real calculation */
			parameters.options.textureImageDrawable=newlayer;
			/* mark the texture image for automatically closing at the end   */
			variables.closeTextureImageAfterCalculation=TRUE;
			variables.textureImageDisplay = newdisplay;
		}
}
