/*********************************************************************************/
/* LIC 0.50 -- image filter plug-in for The Gimp program                         */
/* Copyright (C) 2001 Frank Lauterwald                                           */
/*===============================================================================*/
/* E-mail: frankl@gmx.net                                                        */
/* You can contact the original The Gimp authors at gimp@xcf.berkeley.edu        */
/*===============================================================================*/
/* This program is free software; you can redistribute it and/or modify it under */
/* the terms of the GNU General Public License as published by the Free Software */
/* Foundation; either version 2 of the License, or (at your option) any later    */
/* version.                                                                      */
/*===============================================================================*/
/* This program is distributed in the hope that it will be useful, but WITHOUT   */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS */
/* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.*/
/*===============================================================================*/
/* You should have received a copy of the GNU General Public License along with  */
/* this program; if not, write to the Free Software Foundation, Inc., 675 Mass   */
/* Ave, Cambridge, MA 02139, USA.                                                */
/*===============================================================================*/
/* In other words, you can't sue me for whatever happens while using this ;)     */
/*********************************************************************************/
/* CHANGELOG									 */
/* Date		Version		Comment						 */
/* 2001 Oct. 1,    0.30		first "release"					 */
/* 2002 Jan.11,    0.32		fixed a bug in get_vector_index_from_vector      */
/*				that caused vectors of angle 0 to be stored      */
/*				as nullvectors					 */
/* 2002 Jan. 26,   0.50		release of first beta version			 */
/*********************************************************************************/

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimptypes.h>
#include <glib.h>
#include <math.h>

#include <stdio.h>
#include <stdlib.h>

#include "lic_vectors.h"
#include "lic_globals.h"
#include "lic_color.h"
#include <libgimp/gimpmath.h>





/* **************************************************************************************** */
/* create a floatvector field  */
/* vectors are normalized to length 1. Vector[0] = (1,0), vector[65534] = (cos 65534/65535, sin 65534/65535), */
/* vector[65535] = (0/0) */
/* **************************************************************************************** */
void lic_create_floatvector_field(lic_floatvector_field *f)
{ int i;

  f->field = malloc(65536 * sizeof(floatvector));
  if (!f->field) { fprintf(stderr, "in function create_lic_floatvector_field: cannot allocate memory for vectors "); exit(EXIT_FAILURE); }
  for (i=1; i<65536; i++)
    {  f->field[i].x = cos((float)i/65535*2*G_PI);
       f->field[i].y = sin((float)i/65535*2*G_PI);
    }
  f->field[0].x=0;
  f->field[0].y=0;
}
/* **************************************************************************************** */

/* **************************************************************************************** */
/* gets a vector of floats and returns the index to the vector table that is closest to this vector */
/* **************************************************************************************** */
guint16 get_vector_index_from_vector(floatvector f)
{ double angle=0;
  if ((f.x==0) && (f.y==0)) return 0;

  angle = atan2(f.y, f.x)/G_PI/2; /* normalized to -0.5..0.5 i think */
  return angle*65534.0+1;
}
/* **************************************************************************************** */

/* **************************************************************************************** */
/* WARNING: multiple rotations (especially by small angles) may cause some sort of aliasing */
/* **************************************************************************************** */
void rotate_vector_by_index_field(lic_vector_by_index_field *f, guint16 angle_as_int)
{ guint32 index, maxindex;

  fprintf(stderr, "now rotating vectors by %i \n", angle_as_int);

  if(angle_as_int==0) return;

  maxindex = f->width*f->height;
  for (index=0; index<maxindex; index++)
  {  f->data[index]+=angle_as_int;
     if (f->data[index]==angle_as_int) f->data[index]=0;
     else if (f->data[index]==0) f->data[index]=65535;
  }
}

/* **************************************************************************************** */


/* **************************************************************************************** */
/* This function creates a vectorfield from an image and returns a pointer to it	    */
/* ---------------------------------------------------------------------------------------- */
/* In this case vectorfield refers to a field of guint16s that can be translated into real  */
/* vectors (of length 1) via a lic_floatvector_field (see above)			    */
/* this type of field is called lic_vector_by_index_field and defined in lic_vectors.h	    */
/* ---------------------------------------------------------------------------------------- */
lic_vector_by_index_field *lic_create_vectors(gint32 sourcedrawableid)
{ 	lic_vector_by_index_field *field;
	lic_scalarfield		  *scalarfield;
	floatvector		  v;
	int x,y;

	/* variables for temporary values */
	guint32 		width, height;
	guint32			rowindex;

	#ifdef profiling
	/* counters for determining the percentage of nullvectors */
	guint32			number_of_nullvectors;
	guint32			number_of_vectors;
	number_of_vectors = 0; number_of_nullvectors = 0;
	#endif

	if(!(field = malloc(sizeof(field))))
	{ fprintf(stderr, "lic_vector_by_index_field : malloc for field failed \n"); exit(EXIT_FAILURE); }

	/* get the scalar values e.g. brightness depending on what was selected as effect channel */
	scalarfield = lic_create_scalarfield(sourcedrawableid);
	width  = scalarfield->width;
	height = scalarfield->height;
	field->width  = width+2;
	field->height = height+2;


	if (! (field->data=malloc(field->width*field->height*sizeof (guint16))))
        { fprintf(stderr, "lic_vector_by_index_field : malloc for field.data failed \n"); exit(EXIT_FAILURE); }


	/* the actual thing */
	rowindex=0+width;
	for (y=1; y<field->height-1; y++)		/* no vectors are created for the borderline because
							// the field is larger than the image. This is a trick:
							// if there are nullvectors around the image we neednt check
							//  if we hit the image border while calculating the lic:
							// we just do the nullvector-check we would have to do anyway */
	{	rowindex+=width;	/* optimization NOTE: rowindex IS FOR SCALARFIELD INDEX NOT FOR VECTORFIELD INDEX. THEY ARE DIFFERENT ! */
		for (x=1; x<field->width-1; x++)	/* we leave an extra borderline because the usual vector calculation
							// algorithm cannot cope with the borders (e.g. for x=0 we
							// say v.x = image(x+1,y)-image(x-1,y) which could segfault */
		{ /* This is the place to change if you want to make changes: here the vectors for the LIC are created
		  // this block could or even should be put into an extra function at the cost of speed */
			/* calculate vector */
			   v.x = 2*(scalarfield->data[x+1 + rowindex] - scalarfield->data[x-1 + rowindex])+
			   	    scalarfield->data[x+1 + rowindex-width] - scalarfield->data[x-1 + rowindex-width]+
				    scalarfield->data[x+1 + rowindex+width] - scalarfield->data[x-1 + rowindex+width];
			   v.y = 2*(scalarfield->data[x + rowindex+width] - scalarfield->data[x + rowindex-width])+
				    scalarfield->data[x+1 + rowindex+width] - scalarfield->data[x+1 + rowindex-width]+
				    scalarfield->data[x-1 + rowindex+width] - scalarfield->data[x-1 + rowindex-width];
				/* FIXME: needed: some kind of agressively preventing nullvectors */
			/* get it into the field */
			   field->data[x+y*field->width]=get_vector_index_from_vector(v);

			#ifdef profiling
				if ( (v.x==0)&&(v.y==0) )
					{ number_of_nullvectors++; }
			#endif
	/*FIXME: currently the borderline of the image has random vectors */
		}
	}


	#ifdef profiling
		number_of_vectors = (width-2)*(height-2);
		fprintf(stderr, "Percentage of Nullvectors: %f ",(float)number_of_nullvectors/(float)number_of_vectors);
	#endif

	/* now rotate the field if wanted */
	if (parameters.options.rotateVectors)
	{	rotate_vector_by_index_field(field, 16384); }

	/* second pass: initialize the borderline of the field with nullvectors */
	for (x=0; x<field->width; x++)
	{ 	field->data[x+0*field->width]=0;
		field->data[x+(field->height-1)*field->width]=0;
	}
	for (y=0; y<field->height; y++)
	{ 	field->data[		0 + y*field->width]=0;
		field->data[field->width-1  + y*field->width]=0;
	}
	/* --second pass: initialize the borderline of the field with nullvectors */

	/* clean up */
	free(scalarfield);
fprintf(stderr, "vectorfield created \n");
	return field;
}
/* **************************************************************************************** */


/* **************************************************************************************** */
/* This function creates the needed scalarfield						    */
/* **************************************************************************************** */
lic_scalarfield *lic_create_scalarfield(gint32 drawableid)
{	GimpDrawable *drawable;
	GimpPixelRgn region;
	ColorConversionFunction cfunc;
	gpointer pr;
	int  x,y, x0, y0; /* x0 and y0 are increased exactly when x and y are. They however start at 0	*/
	guint w,h, rowstride, index, bpp; /* w and h hold width and height __of the current tile only__	*/
	lic_scalarfield *field;
	drawable = gimp_drawable_get(drawableid);

	/* create field */
	if (!(field=malloc(sizeof(lic_scalarfield))))
	{	fprintf(stderr, "lic_create_scalarfield: malloc for field failed \n"); exit(EXIT_FAILURE); }
	field->width = drawable->width; field->height = drawable->height;
	if (!(field->data=malloc(field->height*field->width*sizeof(gfloat))))
	{	fprintf(stderr, "lic_create_scalarfield: malloc for field->data failed \n"); exit(EXIT_FAILURE); }


	/* WARNING: The Image mode is derived from the number of bytes per pixel
	//		we assume : 1 bpp -> grey
	//			    2 bpp -> exit: unknown
	//			    3 bpp -> rgb
	//			    4 bpp -> rgba ; the FIXME: alpha value is ignored at the moment. Is that what the user would expect?
	// determine which ColorConversionFunction to use	*/
	switch(drawable->bpp)
	{	case 1:
		{
		  switch(parameters.options.effectchannel)
		  {	case EFFECTCHANNEL_HUE:
				{ cfunc = get_gray_hue;
				  break;
				}
			case EFFECTCHANNEL_SATURATION:
				{ cfunc = get_gray_saturation;
			  	break;
				}
			case EFFECTCHANNEL_BRIGHTNESS:
				{ cfunc = get_gray_brightness;
			  	break;
				}
			case EFFECTCHANNEL_RED:
				{ cfunc = get_gray_red;
			  	break;
				}
			case EFFECTCHANNEL_GREEN:
				{ cfunc = get_gray_green;
			  	break;
				}
			case EFFECTCHANNEL_BLUE:
				{ cfunc = get_gray_blue;
			  	break;
				}
			default:
				{ fprintf(stderr, "lic_create_scalarfield: parameters.options.effectchannel has illegal value %i \n",parameters.options.effectchannel); exit(EXIT_FAILURE);
				}
		  }
		  break;
		}
		case 3:
		{
		  switch(parameters.options.effectchannel)
		  {	case EFFECTCHANNEL_HUE:
				{ cfunc = get_rgb_hue;
				  break;
				}
			case EFFECTCHANNEL_SATURATION:
				{ cfunc = get_rgb_saturation;
			  	break;
				}
			case EFFECTCHANNEL_BRIGHTNESS:
				{ cfunc = get_rgb_brightness;
			  	break;
				}
			case EFFECTCHANNEL_RED:
				{ cfunc = get_rgb_red;
			  	break;
				}
			case EFFECTCHANNEL_GREEN:
				{ cfunc = get_rgb_green;
			  	break;
				}
			case EFFECTCHANNEL_BLUE:
				{ cfunc = get_rgb_blue;
			  	break;
				}
			default:
				{ fprintf(stderr, "lic_create_scalarfield: parameters.options.effectchannel has illegal value %i \n",parameters.options.effectchannel); exit(EXIT_FAILURE);
				}
		  }
		  break;
		}
		default:
		{ fprintf(stderr, "currently only 1bpp(grey) and 3bpp(rgb) images are supported \n"); exit(EXIT_FAILURE);
		}
	}

	bpp = drawable->bpp;
	gimp_pixel_rgn_init(&region, drawable, 0, 0, drawable->width, drawable->height, FALSE, FALSE);
	/* loop over all tiles */
	for (pr = gimp_pixel_rgns_register (1, &region);pr != NULL; pr = gimp_pixel_rgns_process (pr))
	{	if (!((region.bpp==3) || (region.bpp==1)))
		{ fprintf(stderr, "at the moment only rgb or grayscale effect images are supported \n"); exit(EXIT_FAILURE); }
		w = region.w; h = region.h; rowstride = region.rowstride;
		y0=0;
		for (y = region.y; y < region.y + h; y++)
		{	x0=0;
			for (x = region.x; x < region.x + w; x++)
			{	index = (x0*bpp)+y0*rowstride;
				field->data[x+y*drawable->width]=cfunc(&(region.data[index]));
			  	x0++;
			}
		  	y0++;
		}
	}
fprintf(stderr, "Scalarfield created \n");
	return field;
}
/* **************************************************************************************** */

/* **************************************************************************************** */
