/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "support2.h"
#include "colorarea.h"
#include "colorpencil.h"
#include "colorpicker.h"
#include "texture.h"
#include "gtkrgbarea.h"
#include "texturewindow.h"
#include "skywindow.h"
#include "gimpsupport.h"
#include "tworld.h"


GtkWidget*
create_color_pencil (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  GtkWidget *pencil;

  pencil = color_pencil_new ();
  gtk_widget_set_extension_events (pencil, GDK_EXTENSION_EVENTS_ALL);

  return pencil;
}


static void
end_path (PencilPath *path)
{
  color_pencil_path_end (NULL, path);
}


static gfloat
pressure_curve (gfloat x)
{
  /* Use x^0.4 for the Graphire USB driver.
     Otherwise use x^1.0 for the unified Graphire/Intuit Serial/USB driver at
     http://www.lepied.com/xfree86/ */

  return pow (x, 1.0);
}


gboolean
on_color_pencil_button_press_event     (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  GtkWidget  *color_area;
  GtkWidget  *window;
  PencilPath *path;
  GtkWidget  *color_pencil;
  gboolean    eraser;
  gfloat      pressure;
  gfloat      red, green, blue;

  if (event->source == GDK_SOURCE_MOUSE &&
      event->button != 1 && event->button != 3)
    return TRUE;

  window = lookup_toplevel (widget);
  color_pencil = lookup_widget (widget, "color_pencil");

  color_area = lookup_widget (widget, "color_area");
  color_area_get_back (color_area, &red, &green, &blue);
  color_pencil_eraser_color (COLOR_PENCIL (color_pencil),
                             red, green, blue);
  color_area_get_fore (color_area, &red, &green, &blue);
  color_pencil_point_color (COLOR_PENCIL (color_pencil),
                            red, green, blue);

  COLOR_PENCIL (color_pencil)->point_radius   = get_float (widget, "radius");
  COLOR_PENCIL (color_pencil)->point_opacity  = get_float (widget, "opacity");
  COLOR_PENCIL (color_pencil)->point_waxiness = get_float (widget, "wax");

  eraser = (event->source == GDK_SOURCE_ERASER) ||
           (event->source == GDK_SOURCE_MOUSE && event->button == 3);
  pressure = (event->source == GDK_SOURCE_MOUSE) ? 0.9 : event->pressure;
  pressure = pressure_curve (pressure);

  path = color_pencil_path_start (COLOR_PENCIL (color_pencil), eraser);
  gtk_object_set_data_full (GTK_OBJECT (window), "data_path", path,
                            (GtkDestroyNotify) end_path);

  color_pencil_path_point (COLOR_PENCIL (color_pencil), path,
                           event->x, event->y, pressure);

  /* Paint queued redraws. */
  gtk_rgb_area_hard_sync (GTK_RGB_AREA (color_pencil));

  return TRUE;
}


gboolean
on_color_pencil_motion_notify_event    (GtkWidget       *widget,
                                        GdkEventMotion  *event,
                                        gpointer         user_data)
{
  PencilPath *path;
  GtkWidget  *color_pencil;
  gfloat      pressure;

  if ((event->state & (256 | 1024)) == 0)
    return TRUE;

  color_pencil = lookup_widget (widget, "color_pencil");
  path = lookup_data (widget, "data_path");

  pressure = (event->source == GDK_SOURCE_MOUSE) ? 0.9 : event->pressure;
  pressure = pressure_curve (pressure);

  if (path != NULL)
    {
      color_pencil_path_point (COLOR_PENCIL (color_pencil), path,
                               event->x, event->y, pressure);

      /* Paint queued redraws. */
      gtk_rgb_area_hard_sync (GTK_RGB_AREA (color_pencil));
    }

  return TRUE;
}


void
on_clear_clicked                       (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *pencil;
  GtkWidget *color_area;
  gfloat     red, green, blue;

  pencil = lookup_widget (GTK_WIDGET (button), "color_pencil");

  /* Set background color. */
  color_area = lookup_widget (GTK_WIDGET (button), "color_area");
  color_area_get_back (color_area, &red, &green, &blue);
  color_pencil_eraser_color (COLOR_PENCIL (pencil),
                             red, green, blue);

  /* Clear and paint queued redraws. */
  color_pencil_clear (COLOR_PENCIL (pencil));
  gtk_rgb_area_hard_sync (GTK_RGB_AREA (pencil));
}


GtkWidget*
create_texture_preview (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  return gtk_rgb_area_new ();
}


void
on_grain_toggled                       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  if (gtk_toggle_button_get_active (togglebutton))
    texture_update (GTK_WIDGET (togglebutton));
}


void
on_texture_ok_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  texture_create (GTK_WIDGET (button));
  gtk_main_quit ();
}


void
on_pencil_ok_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GimpDrawable *drawable;
  GtkWidget    *pencil;
  gint          width, height;
  gint32       *image_id;
  gint          offset_x, offset_y;
  gboolean      is_new;

  drawable = lookup_data (GTK_WIDGET (button), "data_drawable");
  image_id = lookup_data (GTK_WIDGET (button), "data_image_id");
  pencil = lookup_widget (GTK_WIDGET (button), "color_pencil");

  gimp_undo_push_group_start (*image_id);

  /* Save the color information */
  width = gimp_drawable_width (drawable->id);
  height = gimp_drawable_height (drawable->id);
  tiles_to_from_pixmap (GTK_RGB_AREA (pencil)->rgb_data,
                        3, width * 3, width, height,
                        drawable, 0, 0,
                        FALSE);

  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->id, TRUE);
  gimp_drawable_update (drawable->id, 0, 0, width, height);

  gimp_drawable_offsets (drawable->id, &offset_x, &offset_y);

  /* Save the wax layer */
  drawable =
    gimp_drawable_get (find_create_channel (*image_id, "Wax", 0, 255, 0, &is_new));
  if (!is_new)
    copy_to_shadow (drawable);
  tiles_to_from_pixmap (COLOR_PENCIL (pencil)->wax,
                        1, width, width, height,
                        drawable, -offset_x, -offset_y,
                        FALSE);

  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->id, TRUE);
  gimp_drawable_update (drawable->id, 0, 0, width, height);
  gimp_drawable_detach (drawable);

  /* Save the texture layer */
  drawable =
    gimp_drawable_get (find_create_channel (*image_id, "Texture", 0, 0, 255, &is_new));
  if (!is_new)
    copy_to_shadow (drawable);
  tiles_to_from_pixmap (COLOR_PENCIL (pencil)->texture,
                        1, width, width, height,
                        drawable, -offset_x, -offset_y,
                        FALSE);

  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->id, TRUE);
  gimp_drawable_update (drawable->id, 0, 0, width, height);
  gimp_drawable_detach (drawable);

  gimp_undo_push_group_end (*image_id);

  gtk_main_quit ();
}


void
on_color_sel_ok_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *color_sel;
  GtkWidget *area;
  gint       target;
  gpointer   ptr;
  gdouble    colors[3];

  ptr = lookup_data (GTK_WIDGET (button), "target");
  target = GPOINTER_TO_INT (ptr);
  area = lookup_data (GTK_WIDGET (button), "color_area");
  color_sel = GTK_COLOR_SELECTION_DIALOG (
    lookup_toplevel (GTK_WIDGET (button)))->colorsel;
  gtk_color_selection_get_color (GTK_COLOR_SELECTION (color_sel), colors);

  if (target == 0)
    color_area_set_fore (area, colors[0], colors[1], colors[2]);
  else
    color_area_set_back (area, colors[0], colors[1], colors[2]);

  gtk_widget_destroy (lookup_toplevel (GTK_WIDGET (button)));
}


void
on_color_sel_cancel_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy (lookup_toplevel (GTK_WIDGET (button)));
}


GtkWidget*
create_color_area (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  return color_area_new ();
}


void
on_sky_ok_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
  sky_update (lookup_toplevel (GTK_WIDGET (button)), FALSE);
}


void
on_sky_generate_preview_clicked        (GtkButton       *button,
                                        gpointer         user_data)
{
  sky_update (lookup_toplevel (GTK_WIDGET (button)), TRUE);
}


GtkWidget*
create_sky_preview (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  return gtk_rgb_area_new ();
}


void
on_show_sun_toggled                    (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  gboolean  show;
  gchar    *widget_names[] = { "x", "sun_x", "y", "sun_y", NULL };
  gint      i;

  show = get_boolean (GTK_WIDGET (togglebutton), "show_sun");
  for (i = 0; widget_names[i] != NULL; i++)
    gtk_widget_set_sensitive (lookup_widget (GTK_WIDGET (togglebutton),
                                             widget_names[i]), show);
}


GtkWidget*
create_color_picker (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  return color_picker_new ();
}
