#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include "interface.h"
#include "support.h"
#include "gtkrgbarea.h"

static void
fill_rect (GtkRgbArea *area,
           gint x, gint y, gint width, gint height,
           gfloat red, gfloat green, gfloat blue)
{
  gint   i, j;
  guint8 r, g, b;

  r = red * 255.0;
  g = green * 255.0;
  b = blue * 255.0;

  for (i = 0; i < height; i++)
    {
      guint8 *pos;

      pos = &area->rgb_data [((y + i) * area->width + x) * 3];
      for (j = 0; j < width; j++)
        {
          pos[0] = r; pos[1] = g; pos[2] = b;
          pos += 3;
        }
    }
}

static void
draw_rect (GtkRgbArea *area,
           gint x, gint y, gint width, gint height,
           gfloat red, gfloat green, gfloat blue)
{
  fill_rect (area, x, y, width, 1, red, green, blue);
  fill_rect (area, x, y + height - 1, width, 1, red, green, blue);
  fill_rect (area, x, y, 1, height, red, green, blue);
  fill_rect (area, x + width - 1, y, 1, height, red, green, blue);
}

static void
color_area_update (GtkWidget *widget)
{
  GtkRgbArea   *area = GTK_RGB_AREA (widget);
  GdkRectangle *rectangles;
  gdouble      *colors;

  colors = gtk_object_get_data (GTK_OBJECT (widget), "colors");
  rectangles = gtk_object_get_data (GTK_OBJECT (widget), "rectangles");

  fill_rect (area, 0, 0, area->width, area->height, 0.0, 0.0, 0.0);

  /* Draw background area. */
  draw_rect (area, rectangles[1].x, rectangles[1].y,
             rectangles[1].width, rectangles[1].height,
             1.0, 1.0, 1.0);
  fill_rect (area, rectangles[1].x + 2, rectangles[1].y + 2,
             rectangles[1].width - 4, rectangles[1].height - 4,
             colors[3], colors[4], colors[5]);

  /* Draw foreground area. */
  draw_rect (area, rectangles[0].x, rectangles[0].y,
             rectangles[0].width, rectangles[0].height,
             1.0, 1.0, 1.0);
  fill_rect (area, rectangles[0].x + 1, rectangles[0].y + 1,
             rectangles[0].width - 2, rectangles[0].height - 2,
             0.0, 0.0, 0.0);
  fill_rect (area, rectangles[0].x + 2, rectangles[0].y + 2,
             rectangles[0].width - 4, rectangles[0].height - 4,
             colors[0], colors[1], colors[2]);


  gtk_rgb_area_invalidate (area, 0, 0, area->width, area->height);
  gtk_rgb_area_soft_sync (area);
}

static gboolean
point_in (GdkRectangle *rectangle,
          gint          x,
          gint          y)
{
  return x >= rectangle->x &&
         y >= rectangle->y &&
         x < ((gint) rectangle->x + rectangle->width) &&
         y < ((gint) rectangle->y + rectangle->height);
}

static void
color_area_sel_destroy (GtkWidget *area)
{
  gtk_object_remove_data (GTK_OBJECT (area), "color_sel");
  gtk_object_unref (GTK_OBJECT (area));
}

static gint
color_area_button_press_event (GtkWidget      *widget,
                               GdkEventButton *event)
{
  GdkRectangle *rectangles;
  gdouble      *colors;
  gint          target;

  target = -1;
  rectangles = gtk_object_get_data (GTK_OBJECT (widget), "rectangles");
  colors = gtk_object_get_data (GTK_OBJECT (widget), "colors");

  if (point_in (&rectangles[0], event->x, event->y))
    {
      /* Clicked foreground color */
      target = 0;
    }
  else if (point_in (&rectangles[1], event->x, event->y))
    {
      /* Clicked background color */
      target = 1;
    }

  if (target != -1)
    {
      GtkWidget *color_sel_dialog;
      GtkWidget *color_sel;

      color_sel_dialog = gtk_object_get_data (GTK_OBJECT (widget), "color_sel");
      if (color_sel_dialog == NULL)
        {
          color_sel_dialog = create_color_sel_window ();
          gtk_widget_hide (
            GTK_COLOR_SELECTION_DIALOG (color_sel_dialog)->help_button);
          gtk_object_set_data (GTK_OBJECT (widget), "color_sel",
                               color_sel_dialog);

          gtk_object_ref (GTK_OBJECT (widget));
          gtk_object_set_data_full (GTK_OBJECT (color_sel_dialog), "color_area",
                                    widget,
                                    (GtkDestroyNotify) color_area_sel_destroy);
        }

      color_sel = GTK_COLOR_SELECTION_DIALOG (color_sel_dialog)->colorsel;
      gtk_color_selection_set_color (GTK_COLOR_SELECTION (color_sel),
                                     &colors[target * 3]);

      gtk_object_set_data (GTK_OBJECT (color_sel_dialog), "target",
                           GINT_TO_POINTER (target));

      gtk_widget_show (color_sel_dialog);
      gdk_window_raise (color_sel_dialog->window);
    }

  return TRUE;
}

GtkWidget *
color_area_new ()
{
  GtkWidget    *widget;
  GdkRectangle *rectangles;
  gdouble      *colors;

  widget = gtk_rgb_area_new ();
  gtk_widget_set_events (widget, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK);
  gtk_drawing_area_size (GTK_DRAWING_AREA (widget), 50, 40);
  gtk_signal_connect (GTK_OBJECT (widget), "realize",
                      GTK_SIGNAL_FUNC (color_area_update), NULL);
  gtk_signal_connect (GTK_OBJECT (widget), "button_press_event",
                      GTK_SIGNAL_FUNC (color_area_button_press_event), NULL);

  colors = g_new0 (gdouble, 6);
  colors[3] = colors[4] = colors[5] = 1.0;
  gtk_object_set_data_full (GTK_OBJECT (widget), "colors", colors,
                            (GtkDestroyNotify) g_free);

  rectangles = g_new0 (GdkRectangle, 2);
  rectangles[0].x = 0;      rectangles[0].y = 0;
  rectangles[0].width = 35; rectangles[0].height = 25;
  rectangles[1].x = 14;     rectangles[1].y = 14;
  rectangles[1].width = 35; rectangles[1].height = 25;
  gtk_object_set_data_full (GTK_OBJECT (widget), "rectangles", rectangles,
                            (GtkDestroyNotify) g_free);

  return widget;
}

void
color_area_get_fore (GtkWidget *area,
                     gfloat    *red,
                     gfloat    *green,
                     gfloat    *blue)
{
  gdouble *colors;

  colors = gtk_object_get_data (GTK_OBJECT (area), "colors");
  *red   = colors[0];
  *green = colors[1];
  *blue  = colors[2];
}

void
color_area_set_fore (GtkWidget *area,
                     gfloat     red,
                     gfloat     green,
                     gfloat     blue)
{
  gdouble *colors;

  colors = gtk_object_get_data (GTK_OBJECT (area), "colors");
  colors[0] = red;
  colors[1] = green;
  colors[2] = blue;

  color_area_update (area);
}

void
color_area_get_back (GtkWidget *area,
                     gfloat    *red,
                     gfloat    *green,
                     gfloat    *blue)
{
  gdouble *colors;

  colors = gtk_object_get_data (GTK_OBJECT (area), "colors");
  *red   = colors[3];
  *green = colors[4];
  *blue  = colors[5];
}

void
color_area_set_back (GtkWidget *area,
                     gfloat     red,
                     gfloat     green,
                     gfloat     blue)
{
  gdouble *colors;

  colors = gtk_object_get_data (GTK_OBJECT (area), "colors");
  colors[3] = red;
  colors[4] = green;
  colors[5] = blue;

  color_area_update (area);
}
