/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define GTK_NO_CHECK_CASTS

#include <glib.h>
#include <math.h>
#include <libgimp/gimpmath.h>
#include "colorpencil.h"
#include "texture.h"

static GtkRgbAreaClass *parent_class;

static void color_pencil_class_init    (ColorPencilClass *klass);
static void color_pencil_init          (ColorPencil      *pencil);

GtkType
color_pencil_get_type (void)
{
  static GtkType color_pencil_type = 0;
  
  if (!color_pencil_type)
    {
      static const GtkTypeInfo color_pencil_info =
      {
        "ColorPencil",
        sizeof (ColorPencil),
        sizeof (ColorPencilClass),
        (GtkClassInitFunc) color_pencil_class_init,
        (GtkObjectInitFunc) color_pencil_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };
      
      color_pencil_type =
        gtk_type_unique (GTK_TYPE_RGB_AREA, &color_pencil_info);
    }
  
  return color_pencil_type;
}

GtkWidget *
color_pencil_new (void)
{
  ColorPencil *pencil;

  pencil = gtk_type_new (color_pencil_get_type ());

  return GTK_WIDGET (pencil);
}

static void
color_pencil_on_destroy (GtkObject *object)
{
  ColorPencil *pencil = COLOR_PENCIL (object);

  g_free (pencil->wax);
  pencil->wax = NULL;
  g_free (pencil->texture);
  pencil->texture = NULL;

  pencil->point_red      = 0.0;
  pencil->point_green    = 0.0;
  pencil->point_blue     = 0.0;
  pencil->point_radius   = 3.0;
  pencil->point_opacity  = 0.5;
  pencil->point_waxiness = 0.3;

  pencil->eraser_radius  = 10.0;
  pencil->eraser_red     = 1.0;
  pencil->eraser_green   = 1.0;
  pencil->eraser_blue    = 1.0;

  GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

static void
color_pencil_invalidate (ColorPencil *pencil,
                         gint         x,
                         gint         y,
                         gint         width,
                         gint         height)
{
  gtk_rgb_area_invalidate (GTK_RGB_AREA (pencil), x, y, width, height);
}

static gint
color_pencil_on_configure_event (GtkWidget         *widget,
                                 GdkEventConfigure *event)
{
  GtkRgbArea *area = GTK_RGB_AREA (widget);
  ColorPencil *pencil = COLOR_PENCIL (widget);

  if (event->width == area->width && event->height == area->height)
    return TRUE;

  g_free (pencil->wax);
  pencil->wax = g_new0 (guint8, event->width * event->height);
  g_free (pencil->texture);
  pencil->texture = g_new (guint8, event->width * event->height);
  memset (pencil->texture, 0, event->width * event->height);

  color_pencil_invalidate (pencil, 0, 0, event->width, event->height);

  return GTK_WIDGET_CLASS (parent_class)->configure_event (widget, event);
}

static void
color_pencil_class_init (ColorPencilClass *klass)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkType         type;

  type = color_pencil_get_type ();
  object_class = (GtkObjectClass *) klass;
  widget_class = (GtkWidgetClass *) klass;

  parent_class = gtk_type_class (gtk_rgb_area_get_type ());

  object_class->destroy = color_pencil_on_destroy;
  widget_class->configure_event = color_pencil_on_configure_event;
}

static void
color_pencil_init (ColorPencil *pencil)
{
  pencil->wax     = NULL;
  pencil->texture = NULL;

  pencil->point_red      = 0.3;
  pencil->point_green    = 0.3;
  pencil->point_blue     = 0.3;
  pencil->point_radius   = 3.0;
  pencil->point_opacity  = 0.5;
  pencil->point_waxiness = 0.3;

  pencil->eraser_radius  = 10.0;
  pencil->eraser_red     = 1.0;
  pencil->eraser_green   = 1.0;
  pencil->eraser_blue    = 1.0;
}

void
color_pencil_point_color (ColorPencil *pencil,
                          gfloat       red,
                          gfloat       green,
                          gfloat       blue)
{
  pencil->point_red   = red;
  pencil->point_green = green;
  pencil->point_blue  = blue;
}

void
color_pencil_point_properties (ColorPencil *pencil,
                               gfloat       point_radius,
                               gfloat       point_opacity,
                               gfloat       point_waxiness)
{
  if (point_radius >= 0.0)
    pencil->point_radius = point_radius;

  if (point_opacity >= 0.0)
    pencil->point_opacity = point_opacity;

  if (point_waxiness >= 0.0)
    pencil->point_waxiness = point_waxiness;
}

void
color_pencil_eraser_color (ColorPencil *pencil,
                           gfloat       red,
                           gfloat       green,
                           gfloat       blue)
{
  pencil->eraser_red   = red;
  pencil->eraser_green = green;
  pencil->eraser_blue  = blue;
}

static void
erase_pel (ColorPencil *pencil,
           gint         x,
           gint         y,
           gfloat       value,
           gfloat       red,
           gfloat       green,
           gfloat       blue)
{
  GtkRgbArea *area = GTK_RGB_AREA (pencil);
  gint        pos;
  gfloat      r, g, b;

  pos = y * area->width + x;
  pencil->wax[pos] = (gint) (pencil->wax[pos] * value);

  pos *= 3;

  r = ((gfloat) area->rgb_data[pos]) / 255.0F;
  g = ((gfloat) area->rgb_data[pos + 1]) / 255.0F;
  b = ((gfloat) area->rgb_data[pos + 2]) / 255.0F;

  r = r * (1.0 - value) + red * value;
  g = g * (1.0 - value) + green * value;
  b = b * (1.0 - value) + blue * value;

  area->rgb_data[pos] = r * 255.0;
  area->rgb_data[pos + 1] = g * 255.0;
  area->rgb_data[pos + 2] = b * 255.0;
}

static void
paint_pel (ColorPencil *pencil,
           gint         x,
           gint         y,
           gfloat       opacity,
           gfloat       waxiness,
           gfloat       value,
           gfloat       red,
           gfloat       green,
           gfloat       blue)
{
  GtkRgbArea *area = GTK_RGB_AREA (pencil);
  gint        pos;
  gfloat      r, g, b;
  gfloat      r1, g1, b1;
  gfloat      r2, g2, b2;

  pos = y * area->width + x;
  pencil->wax[pos] = (gint)
    ((pencil->wax[pos] * (1.0 - value) + value * 255.0) * waxiness +
     pencil->wax[pos] * (1.0 - waxiness));
  pos *= 3;

  r = ((gfloat) area->rgb_data[pos]) / 255.0F;
  g = ((gfloat) area->rgb_data[pos + 1]) / 255.0F;
  b = ((gfloat) area->rgb_data[pos + 2]) / 255.0F;

  /* (r1, g1, b1) is the alpha composition */
  r1 = r * (1.0 - value) + red * value;
  g1 = g * (1.0 - value) + green * value;
  b1 = b * (1.0 - value) + blue * value;

  red = 1.0 - (1.0 - red) * value;
  green = 1.0 - (1.0 - green) * value;
  blue = 1.0 - (1.0 - blue) * value;

  /* (r2, g2, b2) is the filtered composition */
  r2 = MIN (r, red);
  g2 = MIN (g, green);
  b2 = MIN (b, blue);

  /* Linearly interpolate between the two compositions */
  area->rgb_data[pos] = (r1 * opacity + r2 * (1.0 - opacity)) * 255.0;
  area->rgb_data[pos + 1] = (g1 * opacity + g2 * (1.0 - opacity)) * 255.0;
  area->rgb_data[pos + 2] = (b1 * opacity + b2 * (1.0 - opacity)) * 255.0;
}

static gfloat
threshold (gfloat x, gfloat power, gfloat midpoint)
{
  if (x < midpoint)
    return pow (x * (1.0 / midpoint), power) * 0.5;

  return 1.0 - pow ((1.0 - x) * (1.0 / (1.0 - midpoint)), power) * 0.5;
}

void
color_pencil_point (ColorPencil *pencil,
                    gfloat       center_x,
                    gfloat       center_y,
                    gfloat       pressure)
{
  GtkRgbArea *area = GTK_RGB_AREA (pencil);
  gint        i_radius, i_x, i_y;
  gint        x, y;

  i_x = (gint) center_x;
  i_y = (gint) center_y;
  i_radius = (gint) ceil (pencil->point_radius);

  pressure = MIN (pressure, 1.0);

  for (y = i_y - i_radius - 1; y <= i_y + i_radius + 1; y++)
    {
      for (x = i_x - i_radius - 1; x <= i_x + i_radius + 1; x++)
        {
          if (x >= 0 && y >= 0 && x < area->width && y < area->height)
            {
              gfloat dist;

              dist = sqrt ((x - center_x) * (x - center_x) +
                           (y - center_y) * (y - center_y)) /
                     pencil->point_radius;

              /*
               * The dist value represents a cone:
               *   0.0 at the tip (center_x, center_y),
               *   1.0 at the edges
               */
              if (dist <= 1.0)
                {
                  gfloat texture;
                  gfloat tip_pressure;
                  gint   pos;
                  gfloat alpha;

                  /*
                   * The threshold function provides a reasonable model
                   * of the contact between the pencil tip and the paper
                   * given the tablet's pressure input
                   */
                  dist = threshold (dist, 4.0, pressure);

                  tip_pressure = (1.0 - dist) * pressure;

                  pos = y * area->width + x;

                  /*
                   * Wax covers the paper's grain.
                   *
                   * wax[pos] = 0.0 return texture[pos]
                   * wax[pos] = 1.0 return 0.6
                   *
                   * Linearly interpolate between extremes.
                   */
                  texture = ((gdouble) pencil->wax[pos]) / 256.0 * 0.6 +
                            (1.0 - ((gdouble) pencil->wax[pos]) / 256.0) *
                            (((gdouble) pencil->texture[pos]) / 256.0);

                  alpha = tip_pressure - texture;
                  if (alpha > 1.0)
                    alpha = 1.0;

                  if (alpha > 0.0)
                    paint_pel (pencil, x, y,
                               pencil->point_opacity,
                               pencil->point_waxiness,
                               alpha,
                               pencil->point_red,
                               pencil->point_green,
                               pencil->point_blue);
                }
            }
        }
    }

  color_pencil_invalidate (pencil,
                           i_x - i_radius, i_y - i_radius,
                           i_radius * 2 + 1, i_radius * 2 + 1);
}

void
color_pencil_erase (ColorPencil *pencil,
                    gfloat       center_x,
                    gfloat       center_y,
                    gfloat       pressure)
{
  GtkRgbArea *area = GTK_RGB_AREA (pencil);
  gint i_radius, i_x, i_y;
  gint x, y;

  i_radius = (gint) pencil->eraser_radius;
  i_x = (gint) center_x;
  i_y = (gint) center_y;

  pressure = MIN (pressure, 1.0);

  for (y = i_y - i_radius - 1; y <= i_y + i_radius + 1; y++)
    {
      for (x = i_x - i_radius - 1; x <= i_x + i_radius + 1; x++)
        {
          if (x >= 0 && y >= 0 && x < area->width && y < area->height)
            {
              gfloat dist;

              dist = sqrt ((x - center_x) * (x - center_x) +
                           (y - center_y) * (y - center_y)) /
                     pencil->eraser_radius;

              if (dist <= 1.0)
                {
                  gfloat tip_pressure;
                  gint   pos;
                  gfloat alpha;

                  /* dist is between 0.0 and 1.0 */
                  dist = threshold (dist, 5.0, pressure);

                  tip_pressure = (1.0 - dist) * pressure;

                  pos = y * area->width + x;

                  alpha = (tip_pressure * 256.0 - pencil->texture[pos]) / 256.0;
                  if (alpha > 0.0)
                    erase_pel (pencil, x, y, alpha,
                               pencil->eraser_red,
                               pencil->eraser_green,
                               pencil->eraser_blue);
                }
            }
        }
    }

  color_pencil_invalidate (pencil,
                           i_x - i_radius, i_y - i_radius,
                           i_radius * 2 + 1, i_radius * 2 + 1);
}

void
color_pencil_clear (ColorPencil *pencil)
{
  GtkRgbArea *area = GTK_RGB_AREA (pencil);
  gint        i, size;
  guint8      red, green, blue;

  red = (gint) RINT (pencil->eraser_red * 255.0);
  green = (gint) RINT (pencil->eraser_green * 255.0);
  blue = (gint) RINT (pencil->eraser_blue * 255.0);
  size = area->width * area->height * 3;
  for (i = 0; i < size; i += 3)
    {
      area->rgb_data[i + 0] = red;
      area->rgb_data[i + 1] = green;
      area->rgb_data[i + 2] = blue;
    }
  memset (pencil->wax, 0, area->width * area->height);
  color_pencil_invalidate (pencil, 0, 0, area->width, area->height);
}

PencilPath *
color_pencil_path_start (ColorPencil *pencil,
                         gboolean     eraser)
{
  PencilPath *path;

  path = g_new0 (PencilPath, 1);
  path->eraser = eraser;
  path->no_last_point = TRUE;

  return path;
}

void
color_pencil_path_point (ColorPencil *pencil,
                         PencilPath  *path,
                         gfloat       x,
                         gfloat       y,
                         gfloat       pressure)
{
  if (path->no_last_point)
    {
      if (path->eraser)
        color_pencil_erase (COLOR_PENCIL (pencil), x, y, pressure);
      else
        color_pencil_point (COLOR_PENCIL (pencil), x, y, pressure);

      path->blot_x = x;
      path->blot_y = y;
      path->no_last_point = FALSE;
    }
  else
    {
      gfloat spacing;
      gfloat i, i_distance;

      i_distance = (gint) sqrt ((path->last_x - x) * (path->last_x - x) +
                                (path->last_y - y) * (path->last_y - y));
      if (i_distance == 0)
        i_distance = 1;

      if (path->eraser)
        spacing = pencil->eraser_radius * 0.5;
      else
        spacing = pencil->point_radius * 0.25;

      for (i = 0; i < i_distance; i++)
        {
          gfloat blot_x, blot_y, blot_pressure;
          gfloat blot_spacing;

          blot_x = path->last_x + (x - path->last_x) * i / i_distance;
          blot_y = path->last_y + (y - path->last_y) * i / i_distance;
          blot_pressure = path->last_pressure +
                          (pressure - path->last_pressure) * i / i_distance;

          blot_spacing =
            sqrt ((blot_x - path->blot_x) * (blot_x - path->blot_x) +
                  (blot_y - path->blot_y) * (blot_y - path->blot_y));

          if (blot_spacing >= spacing)
            {
              if (path->eraser)
                color_pencil_erase (pencil, blot_x, blot_y, blot_pressure);
              else
                color_pencil_point (pencil, blot_x, blot_y, blot_pressure);

              path->blot_x = blot_x;
              path->blot_y = blot_y;
            }
        }
    }

  path->last_x = x;
  path->last_y = y;
  path->last_pressure = pressure;
}

void
color_pencil_path_end (ColorPencil *pencil,
                       PencilPath  *path)
{
  g_free (path);
}
