/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

void
copy_to_shadow (GimpDrawable *drawable)
{
  GimpPixelRgn from_region;
  GimpPixelRgn to_region;
  gpointer     from_ptr;
  gpointer     to_ptr;

  gimp_pixel_rgn_init (&from_region, drawable,
                       0, 0,
                       drawable->width, drawable->height,
                       FALSE, FALSE);

  gimp_pixel_rgn_init (&to_region, drawable,
                       0, 0,
                       drawable->width, drawable->height,
                       TRUE, TRUE);

  for (from_ptr = gimp_pixel_rgns_register (1, &from_region),
       to_ptr = gimp_pixel_rgns_register (1, &to_region);
       from_ptr != NULL && to_ptr != NULL;
       from_ptr = gimp_pixel_rgns_process (from_ptr),
       to_ptr = gimp_pixel_rgns_process (to_ptr))
    {
      memcpy (to_region.data, from_region.data, from_region.rowstride * from_region.h);
    }
}

/*
 * Copy/paste data to/from a drawable into an RGB/Grayscale pixbuf.
 */

gboolean
tiles_to_from_pixmap (guint8       *data,
                      gint          bpp,
                      gint          rowstride,
                      gint          width,
                      gint          height,
                      GimpDrawable *drawable,
                      gint          offset_x,
                      gint          offset_y,
                      gboolean      to_pixmap)
{
  GimpPixelRgn region;
  gpointer     ptr;
  gint         clip_x1, clip_y1, clip_x2, clip_y2;

  clip_x1 = MAX (0, offset_x);
  clip_y1 = MAX (0, offset_y);
  clip_x2 = MIN (width - 1, offset_x + drawable->width - 1);
  clip_y2 = MIN (height - 1, offset_y + drawable->height - 1);

  g_return_val_if_fail (bpp == drawable->bpp, FALSE);

  gimp_pixel_rgn_init (&region, drawable,
                       0, 0,
                       drawable->width, drawable->height,
                       !to_pixmap, !to_pixmap);

  for (ptr = gimp_pixel_rgns_register (1, &region);
       ptr != NULL;
       ptr = gimp_pixel_rgns_process (ptr))
    {
      gint y;
      gint length;
      gint region_x, region_y;
      gint pix_x1, pix_x2, pix_y1, pix_y2;

      /* region coordinate converted to pixbuf space */
      region_x = region.x + offset_x;
      region_y = region.y + offset_y;

      pix_x1 = MAX (clip_x1, region_x);
      pix_y1 = MAX (clip_y1, region_y);
      pix_x2 = MIN (clip_x2, (gint) (region_x + region.w - 1));
      pix_y2 = MIN (clip_y2, (gint) (region_y + region.h - 1));

      if (pix_x2 < pix_x1)
        continue;

      length = (pix_x2 - pix_x1 + 1) * bpp;
      for (y = pix_y1; y <= pix_y2; y++)
        {
          gint pix_pos, draw_pos;

          pix_pos = y * rowstride + pix_x1 * bpp;
          draw_pos =
            (y - offset_y - region.y) * region.rowstride +
            (pix_x1 - offset_x - region.x) * bpp;

          if (to_pixmap)
            memcpy (&data[pix_pos], &region.data[draw_pos], length);
          else
            memcpy (&region.data[draw_pos], &data[pix_pos], length);
        }
    }

  return TRUE;
}

gint32
find_channel (gint32  image_id,
              gchar  *name)
{
  gint    i;
  gint32 *channels;
  gint    count;

  channels = gimp_image_get_channels (image_id, &count);
  for (i = 0; i < count; i++)
    {
      gchar *channel_name;
      gint   comparison;

      channel_name = gimp_channel_get_name (channels[i]);
      comparison = strcmp (channel_name, name);
      g_free (channel_name);

      if (comparison == 0)
        return channels[i];
    }

  return -1;
}

static void
clear_channel (gint32 channel_id)
{
  GimpDrawable *drawable;
  GimpPixelRgn  region;
  gpointer      ptr;

  drawable = gimp_drawable_get (channel_id);
  gimp_pixel_rgn_init (&region, drawable,
                       0, 0,
                       drawable->width, drawable->height,
                       TRUE, TRUE);

  for (ptr = gimp_pixel_rgns_register (1, &region);
       ptr != NULL;
       ptr = gimp_pixel_rgns_process (ptr))
    {
      memset (region.data, 0, region.rowstride * region.h);
    }

  gimp_drawable_detach (drawable);
}

gint32
find_create_channel (gint32    image_id,
                     gchar    *name,
                     guint8    red,
                     guint8    green,
                     guint8    blue,
                     gboolean *is_new)
{
  gint32 channel_id;

  channel_id = find_channel (image_id, name);
  if (channel_id == -1)
    {
      guchar color[3];
      gint   width, height;

      color[0] = red;
      color[1] = green;
      color[2] = blue;
      width = gimp_image_width (image_id);
      height = gimp_image_height (image_id);

      channel_id =
        gimp_channel_new (image_id, name, width, height, 0.0, color);
      clear_channel (channel_id);
      gimp_image_add_channel (image_id, channel_id, 0);
      gimp_channel_set_visible (channel_id, FALSE);

      /* Opacity parameter to gimp_channel_new is broken */
      gimp_channel_set_opacity (channel_id, 70.0);

      if (is_new != NULL)
        *is_new = TRUE;
    }
  else if (is_new != NULL)
    *is_new = FALSE;

  return channel_id;
}
