/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "gtkrgbarea.h"

static GtkDrawingAreaClass *parent_class;

static void gtk_rgb_area_class_init    (GtkRgbAreaClass *klass);
static void gtk_rgb_area_init          (GtkRgbArea      *area);

GtkType
gtk_rgb_area_get_type (void)
{
  static GtkType rgb_area_type = 0;
  
  if (!rgb_area_type)
    {
      static const GtkTypeInfo rgb_area_info =
      {
        "GtkRgbArea",
        sizeof (GtkRgbArea),
        sizeof (GtkRgbAreaClass),
        (GtkClassInitFunc) gtk_rgb_area_class_init,
        (GtkObjectInitFunc) gtk_rgb_area_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };
      
      rgb_area_type = gtk_type_unique (GTK_TYPE_DRAWING_AREA, &rgb_area_info);
    }
  
  return rgb_area_type;
}

GtkWidget*
gtk_rgb_area_new (void)
{
  GtkRgbArea *area;

  area = gtk_type_new (GTK_TYPE_RGB_AREA);

  return GTK_WIDGET (area);
}

static gint
gtk_rgb_area_on_expose_event (GtkWidget      *widget,
                              GdkEventExpose *event)
{
  GtkRgbArea *area = GTK_RGB_AREA (widget);
  GdkGC      *gc;
  gint        x1, y1, x2, y2;

  x1 = MIN (MAX (0, event->area.x), area->width - 1);
  y1 = MIN (MAX (0, event->area.y), area->height - 1);
  x2 = MIN (MAX (0, (gint) (event->area.x + event->area.width - 1)),
            (gint) (area->width - 1));
  y2 = MIN (MAX (0, (gint) (event->area.y + event->area.height - 1)),
            (gint) (area->height - 1));

  gc = widget->style->black_gc;
  gdk_draw_rgb_image_dithalign (widget->window, gc, x1, y1,
                                x2 - x1 + 1, y2 - y1 + 1,
                                GDK_RGB_DITHER_NORMAL,
                                &area->rgb_data[(y1 * area->width + x1) * 3],
                                area->width * 3, x1, y1);

  return TRUE;
}

static gint
gtk_rgb_area_on_configure_event (GtkWidget         *widget,
                                 GdkEventConfigure *event)
{
  GtkRgbArea *area = GTK_RGB_AREA (widget);
  gint        size;

  if (event->width == area->width && event->height == area->height)
    return TRUE;

  g_free (area->rgb_data);
  area->width = event->width;
  area->height = event->height;
  size = area->width * area->height * 3;
  area->rgb_data = g_new (guchar, size);
  memset (area->rgb_data, 255, size);

  return TRUE;
}

static void
gtk_rgb_area_on_destroy (GtkObject *object)
{
  GtkRgbArea *area = GTK_RGB_AREA (object);

  area->width = 0;
  area->height = 0;
  g_free (area->rgb_data);
  area->rgb_data = NULL;

  GTK_OBJECT_CLASS (parent_class)->destroy (object);
}


static void
gtk_rgb_area_class_init (GtkRgbAreaClass *klass)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkType         type;

  parent_class = gtk_type_class (gtk_drawing_area_get_type ());

  type = gtk_rgb_area_get_type ();
  object_class = (GtkObjectClass *) klass;
  widget_class = (GtkWidgetClass *) klass;

  object_class->destroy = gtk_rgb_area_on_destroy;
  widget_class->expose_event = gtk_rgb_area_on_expose_event;
  widget_class->configure_event = gtk_rgb_area_on_configure_event;
}

static void
gtk_rgb_area_init (GtkRgbArea *area)
{
  area->width = 0;
  area->height = 0;
  area->rgb_data = NULL;
  area->dirty.x = area->width;
  area->dirty.y = area->height;
  area->dirty.width = 0;
  area->dirty.height = 0;
}

void
gtk_rgb_area_invalidate (GtkRgbArea *area,
                         gint        x,
                         gint        y,
                         gint        width,
                         gint        height)
{
  GdkRectangle rect, out;

  rect.x = x;
  rect.y = y;
  rect.width = width;
  rect.height = height;

  if (area->dirty.width > 0 || area->dirty.height > 0)
    {
      gdk_rectangle_union (&area->dirty, &rect, &out);
      area->dirty = out;
    }
  else
    {
      area->dirty = rect;
    }
}

void
gtk_rgb_area_soft_sync (GtkRgbArea *area)
{
  gtk_widget_queue_draw_area (GTK_WIDGET (area),
                              area->dirty.x, area->dirty.y,
                              area->dirty.width, area->dirty.height);
  area->dirty.width = 0;
  area->dirty.height = 0;
}

void
gtk_rgb_area_hard_sync (GtkRgbArea *area)
{
  if (area->dirty.width > 0 || area->dirty.height > 0)
    {
      GdkEventExpose  event;
      GtkWidgetClass *klass;

      event.area = area->dirty;
      klass = GTK_WIDGET_CLASS (GTK_OBJECT (area)->klass);
      klass->expose_event (GTK_WIDGET (area), &event);
      gdk_flush ();

      area->dirty.width = 0;
      area->dirty.height = 0;
    }
}
