/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include <libgimp/gimpintl.h>
#include <string.h>
#include "pencilwindow.h"
#include "texturewindow.h"
#include "skywindow.h"

#include "interface.h"
#include "support.h"

/* Declare local functions. */
static void query (void);
static void run   (gchar      *name,
                   gint        nparams,
                   GimpParam  *param,
                   gint       *nreturn_vals,
                   GimpParam **return_vals);


GimpPlugInInfo PLUG_IN_INFO = {
  NULL,                        /* init_proc */
  NULL,                        /* quit_proc */
  query,                       /* query_proc */
  run                          /* run_proc */
};

MAIN ();

static void
query ()
{
  static GimpParamDef color_pencils_args[] = {
    { GIMP_PDB_INT32,    "run_mode", "Interactive" },
    { GIMP_PDB_IMAGE,    "image",    "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" }
  };
  static GimpParamDef *color_pencils_return_vals = NULL;
  static int color_pencils_nargs = sizeof (color_pencils_args) / sizeof (color_pencils_args[0]);
  static int color_pencils_nreturn_vals = 0;

  static GimpParamDef texture_args[] = {
    { GIMP_PDB_INT32,    "run_mode", "Interactive" },
    { GIMP_PDB_IMAGE,    "image",    "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" }
  };
  static GimpParamDef *texture_return_vals = NULL;
  static int texture_nargs = sizeof (texture_args) / sizeof (texture_args[0]);
  static int texture_nreturn_vals = 0;

  static GimpParamDef sky_args[] = {
    { GIMP_PDB_INT32,    "run_mode",      "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE,    "image",         "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable",      "Input drawable" },
    { GIMP_PDB_FLOAT,    "tilt",          "Camera tilt angle (0.0 - 90.0)" },
    { GIMP_PDB_FLOAT,    "rotation",    "Camera rotation angle (0.0 - 90.0)" },
    { GIMP_PDB_INT32,    "seed",        "Random seed, -1 to use current time" },
    { GIMP_PDB_INT32,    "sun_show",      "Show sun? (bool)" },
    { GIMP_PDB_FLOAT,    "sun_x",         "Sun's x coordinate (0.0 - 1.0)" },
    { GIMP_PDB_FLOAT,    "sun_y",         "Sun's y coordinate (0.0 - 1.0)" },
    { GIMP_PDB_FLOAT,    "time",          "Time in hours (0.0 - 24.0)" },
    { GIMP_PDB_COLOR,    "horizon_color", "Horizon color" },
    { GIMP_PDB_COLOR,    "sky_color",     "Color at highest point in the sky" },
    { GIMP_PDB_COLOR,    "sun_color",     "Sun color" },
    { GIMP_PDB_COLOR,    "cloud_color" ,  "Cloud color" },
    { GIMP_PDB_COLOR,    "shadow_color",  "Cloud shadow color" }
  };
  static GimpParamDef *sky_return_vals = NULL;
  static int sky_nargs = sizeof (sky_args) / sizeof (sky_args[0]);
  static int sky_nreturn_vals = 0;

  gimp_install_procedure ("plug_in_color_pencils",
                          "Draw using simulated artist's pencils.",
                          "",
                          "David A. Bartold",
                          "David A. Bartold",
                          "2000-2001",
                          "<Image>/Media/Color Pencils...",
                          "RGB",
                          GIMP_PLUGIN,
                          color_pencils_nargs, color_pencils_nreturn_vals,
                          color_pencils_args, color_pencils_return_vals);

  gimp_install_procedure ("plug_in_texture",
                          "Create new paper textures.",
                          "",
                          "David A. Bartold",
                          "David A. Bartold",
                          "2000-2001",
                          "<Image>/Media/Texture...",
                          "RGB",
                          GIMP_PLUGIN,
                          texture_nargs, texture_nreturn_vals,
                          texture_args, texture_return_vals);

  gimp_install_procedure ("plug_in_sky",
                          "Render near photorealistic skies.",
                          "",
                          "David A. Bartold",
                          "David A. Bartold",
                          "2000-2001",
                          "<Image>/Filters/Render/Clouds/Sky...",
                          "RGB*",
                          GIMP_PLUGIN,
                          sky_nargs, sky_nreturn_vals,
                          sky_args, sky_return_vals);
}

static void
assign_color (GimpParamColor *in, GimpRGB *out)
{
  out->r = ((gdouble) (in->red)) / 255.0;
  out->g = ((gdouble) (in->green)) / 255.0;
  out->b = ((gdouble) (in->blue)) / 255.0;
}


static void
run (gchar      *name,
     gint        n_params,
     GimpParam  *param,
     gint       *nreturn_vals,
     GimpParam **return_vals)
{
  static GimpParam   values[1];
  gint32             image_id;
  GimpDrawable      *drawable;
  GimpRunModeType    run_mode;
  GimpPDBStatusType  status = GIMP_PDB_SUCCESS;

  *nreturn_vals = 1;
  *return_vals = values;

  /*  Get the specified drawable  */
  image_id = param[1].data.d_image;
  drawable = gimp_drawable_get (param[2].data.d_drawable);

  run_mode = (GimpRunModeType) param[0].data.d_int32;

  if (strcmp (name, "plug_in_color_pencils") == 0)
    {
      if (run_mode == GIMP_RUN_INTERACTIVE)
        {
          if (pencil_window (image_id, drawable))
            {
              /* The dialog was closed, or something similarly evil happened. */
              status = GIMP_PDB_EXECUTION_ERROR;
            }
        }
    }
  else if (strcmp (name, "plug_in_texture") == 0)
    {
      if (run_mode == GIMP_RUN_INTERACTIVE)
        {
          if (texture_window (image_id, drawable))
            {
              /* The dialog was closed, or something similarly evil happened. */
              status = GIMP_PDB_EXECUTION_ERROR;
            }
        }
    }
  else if (strcmp (name, "plug_in_sky") == 0)
    {
      SkyValues values;

      sky_values_init (&values);
      gimp_get_data ("plug_in_sky", &values);

      switch (run_mode)
        {
        case GIMP_RUN_INTERACTIVE:
          if (sky_window (image_id, drawable, &values))
            {
              /* The dialog was closed, or something similarly evil happened. */
              status = GIMP_PDB_EXECUTION_ERROR;
            }
        break;

        case GIMP_RUN_NONINTERACTIVE:
          if (n_params != 15)
            status = GIMP_PDB_CALLING_ERROR;

          if (status == GIMP_PDB_SUCCESS)
            {
              values.tilt          = param[3].data.d_float;
              values.rotation      = param[4].data.d_float;
              values.seed          = param[5].data.d_int32;
              values.sun_show      = param[6].data.d_int32;
              values.sun_x         = param[7].data.d_float;
              values.sun_y         = param[8].data.d_float;
              values.time          = param[9].data.d_float;
              assign_color (&param[10].data.d_color, &values.horizon_color);
              assign_color (&param[11].data.d_color, &values.sky_color);
              assign_color (&param[12].data.d_color, &values.sun_color);
              assign_color (&param[13].data.d_color, &values.cloud_color);
              assign_color (&param[14].data.d_color, &values.shadow_color);
              sky_values_apply (drawable, &values);
            }
          break;

        case GIMP_RUN_WITH_LAST_VALS:
          sky_values_apply (drawable, &values);
          break;
        }
    }

  if (status == GIMP_PDB_SUCCESS)
    gimp_drawable_detach (drawable);

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;
}
