/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <time.h>

#include "interface.h"
#include "support.h"
#include "support2.h"
#include "colorpicker.h"
#include "tworld.h"
#include "gtkrgbarea.h"

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#define GAMMA 1.5
#define TOSCREEN(x) pow(x, 1.0 / (GAMMA))
#define FROMSCREEN(x) pow(x, GAMMA)


void
sky_values_pack (GtkWidget *window,
                 SkyValues *values)
{
  GtkWidget    *color_picker;

  values->rotation = get_float (window, "rotation_angle");
  values->tilt = get_float (window, "tilt_angle");
  values->seed = get_int (window, "seed");
  values->sun_show = get_boolean (window, "show_sun");
  values->sun_y = get_float (window, "sun_y");
  values->sun_x = get_float (window, "sun_x");
  values->time = get_float (window, "time");

  color_picker = lookup_data (window, "horizon_color");
  color_picker_get_color (color_picker,
                          &values->horizon_color.r,
                          &values->horizon_color.g,
                          &values->horizon_color.b);

  color_picker = lookup_data (window, "sky_color");
  color_picker_get_color (color_picker,
                          &values->sky_color.r,
                          &values->sky_color.g,
                          &values->sky_color.b);

  color_picker = lookup_data (window, "sun_color");
  color_picker_get_color (color_picker,
                          &values->sun_color.r,
                          &values->sun_color.g,
                          &values->sun_color.b);

  color_picker = lookup_data (window, "cloud_color");
  color_picker_get_color (color_picker,
                          &values->cloud_color.r,
                          &values->cloud_color.g,
                          &values->cloud_color.b);

  color_picker = lookup_data (window, "shadow_color");
  color_picker_get_color (color_picker,
                          &values->shadow_color.r,
                          &values->shadow_color.g,
                          &values->shadow_color.b);
}


void
sky_values_unpack (GtkWidget *window,
                   SkyValues *values)
{
  GtkWidget *color_picker;
  gint       seed;

  seed = values->seed;
  if (seed == -1)
    seed = time (NULL);

  set_int (window, "seed", seed);

  set_float (window, "rotation_angle", values->rotation);
  set_float (window, "tilt_angle", values->tilt);
  set_boolean (window, "show_sun", values->sun_show);
  set_float (window, "sun_x", values->sun_x);
  set_float (window, "sun_y", values->sun_y);
  set_float (window, "time", values->time);

  color_picker = lookup_widget (window, "horizon_color");
  color_picker_set_color (color_picker,
                          values->horizon_color.r,
                          values->horizon_color.g,
                          values->horizon_color.b);
  color_picker = lookup_widget (window, "sky_color");
  color_picker_set_color (color_picker,
                          values->sky_color.r,
                          values->sky_color.g,
                          values->sky_color.b);
  color_picker = lookup_widget (window, "sun_color");
  color_picker_set_color (color_picker,
                          values->sun_color.r,
                          values->sun_color.g,
                          values->sun_color.b);
  color_picker = lookup_widget (window, "cloud_color");
  color_picker_set_color (color_picker,
                          values->cloud_color.r,
                          values->cloud_color.g,
                          values->cloud_color.b);
  color_picker = lookup_widget (window, "shadow_color");
  color_picker_set_color (color_picker,
                          values->shadow_color.r,
                          values->shadow_color.g,
                          values->shadow_color.b);
}


void
sky_values_apply (GimpDrawable *drawable,
                  SkyValues    *values)
{
  sky_render (drawable, values);

  gimp_drawable_merge_shadow (drawable->id, TRUE);
  gimp_drawable_update (drawable->id, 0, 0,
                        drawable->width, drawable->height);
  gimp_drawable_flush (drawable);
  gimp_set_data ("plug_in_sky", values, sizeof (SkyValues));
  gimp_displays_flush ();
}


void
sky_update (GtkWidget *window,
            gboolean   preview)
{
  GimpDrawable *drawable;
  SkyValues     values;

  drawable = lookup_data (window, "data_drawable");
  sky_values_pack (window, &values);

  if (preview)
    {
      GtkWidget  *widget;
      GtkRgbArea *preview;

      widget = lookup_widget (window, "preview");
      preview = GTK_RGB_AREA (widget);

      sky_preview (preview->rgb_data, preview->width, preview->height, &values);
      gtk_widget_queue_draw (widget);
    }
  else
    {
      gtk_widget_destroy (window);
      gtk_main_iteration ();

      sky_values_apply (drawable, &values);

      gtk_main_quit ();
    }
}


gint
sky_window (gint32        image_id,
            GimpDrawable *drawable,
            SkyValues    *values)
{
  GtkWidget *window;
  int        argc = 1;
  char      *argv[2] = { "digi_pencil", NULL };
  char     **argv2 = argv;
  GtkWidget *preview;
  gint       width, height;
  gint       max_size;

#ifdef ENABLE_NLS
  bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
  textdomain (PACKAGE);
#endif

  gtk_set_locale ();
  setlocale (LC_NUMERIC, "C");
  gtk_init (&argc, &argv2);
  gdk_rgb_init ();

  gtk_widget_set_default_colormap (gdk_rgb_get_cmap ());
  gtk_widget_set_default_visual (gdk_rgb_get_visual ());

  add_pixmap_directory (g_strconcat (PACKAGE_DATA_DIR, G_DIR_SEPARATOR_S "pixmaps", NULL));

  window = create_sky_window ();
  sky_values_unpack (window, values);
  preview = lookup_widget (window, "preview");
  max_size = MAX (drawable->width, drawable->height);
  width = drawable->width * 100 / max_size;
  height = drawable->height * 100 / max_size;
  gtk_drawing_area_size (GTK_DRAWING_AREA (preview), width, height);

  gtk_object_set_data (GTK_OBJECT (window), "data_image_id", &image_id);
  gtk_object_set_data (GTK_OBJECT (window), "data_drawable", drawable);
  gtk_widget_show_all (window);

  gtk_main ();
  gdk_flush ();

  return 0;
}
