/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <glib.h>
#include <libgimp/gimpmath.h>

static void
blur (gfloat *data,
      gint    width,
      gint    height)
{
  gint x, y, pos;

  /* Perform some more smoothing */
  pos = 0;
  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width; x++)
        {
          gfloat value;
          gint   count;

          count = 1;
          value = data[pos];

          if (x > 0)          value += data[pos - 1], count++;
          if (x < width - 1)  value += data[pos + 1], count++;
          if (y > 0)          value += data[pos - width], count++;
          if (y < height - 1) value += data[pos + width], count++;

          data[pos] = value / count;

          pos++;
        }
    }
}

#if 0
static void
vertical_blur (gfloat *data,
               gint    width,
               gint    height)
{
  gint x, y, pos;

  /* Perform some more smoothing */
  pos = 0;
  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width; x++)
        {
          gfloat value;
          gint   count;

          count = 1;
          value = data[pos];

          if (y > 0)          value += data[pos - width], count++;
          if (y < height - 1) value += data[pos + width], count++;

          data[pos] = value / count;

          pos++;
        }
    }
}

static void
horizontal_blur (gfloat *data,
                 gint    width,
               gint    height)
{
  gint x, y, pos;

  /* Perform some more smoothing */
  pos = 0;
  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width; x++)
        {
          gfloat value;
          gint   count;

          count = 1;
          value = data[pos];

          if (x > 0)          value += data[pos - 1], count++;
          if (x < width - 1)  value += data[pos + 1], count++;

          data[pos] = value / count;

          pos++;
        }
    }
}
#endif

guint8 *
texture_light_new (gint   width,
                   gint   height,
                   gfloat scale,
                   gint   seed)
{
  gint    x, y, pos;
  gfloat  value;
  guint8 *data_8;
  gfloat *data;
  gfloat  min, max;

  if (seed == -1)
    seed = time (NULL);
  srand (seed);

  data = g_new (gfloat, width * height);

  /* Start in the middle of the smoothed random data. */
  value = 0.0;
  for (x = 0; x < width; x++)
    value = value * 0.8 + (rand () & 0xff) / 255.0 * 0.2;

  /* Create x-blurred data */
  pos = 0;
  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width; x++)
        {
          data[pos] = value;

          value = value * 0.8 + (rand () & 0xff) / 255.0 * 0.2;
          pos++;
        }
    }

  /* Radius blur */
  for (x = 0; x < (gint) ceil (scale - 1.0); x++)
    blur (data, width, height);

  /* Get min/max of texture. */
  min = max = data[0];
  for (x = 0; x < width * height; x++)
    {
      if (data[x] < min) min = data[x];
      if (data[x] > max) max = data[x];
    }

  /* Normalize the data. */
  if (max - min > 0.0001)
    {
      for (x = 0; x < width * height; x++)
        data[x] = (data[x] - min) * (1.0 / (max - min));
    }

  /* Calculate integer data. */
  data_8 = g_new (guint8, width * height);
  for (x = 0; x < width * height; x++)
    data_8[x] = RINT (data[x] * 255.0);

  g_free (data);

  return data_8;
}


guint8 *
texture_heavy_new (gint   width,
                   gint   height,
                   gfloat scale,
                   gint   seed)
{
  gint    x, y, pos;
  gfloat  value;
  guint8 *data_8;
  gfloat *data;
  gfloat  min, max;

  if (seed == -1)
    seed = time (NULL);
  srand (seed);

  data = g_new (gfloat, width * height);

  /* Start in the middle of the smoothed random data. */
  value = 0.0;
  for (x = 0; x < width; x++)
    value = value * 0.6 + (rand () & 0xff) / 255.0 * 0.4;

  /* Create x-blurred data */
  pos = 0;
  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width; x++)
        {
          data[pos] = value;

          value = value * 0.6 + (rand () & 0xff) / 255.0 * 0.4;
          pos++;
        }
    }

  /* Radius blur */
  for (x = 0; x < (gint) ceil (scale - 1.0); x++)
    blur (data, width, height);

  /* Normalize the data. */
  min = max = data[0];
  for (x = 0; x < width * height; x++)
    {
      if (data[x] < min) min = data[x];
      if (data[x] > max) max = data[x];
    }

  if (fabs (min - max) > 0.0001)
    {
      for (x = 0; x < width * height; x++)
        data[x] = (data[x] - min) * (1.0 / (max - min));
    }

  /* Calculate integer data. */
  data_8 = g_new (guint8, width * height);
  for (x = 0; x < width * height; x++)
    data_8[x] = RINT (data[x] * 255.0);

  g_free (data);

  return data_8;
}

guint8 *
texture_canvas_new (gint   width,
                    gint   height,
                    gfloat scale,
                    gint   seed)
{
  gint    x, y, pos;
  guint8 *data_8;
  gfloat *data;
  gfloat  min, max;
  gfloat  value;

  if (seed == -1)
    seed = time (NULL);
  srand (seed);

  data = g_new (gfloat, width * height);

  /* Start in the middle of the smoothed random data. */
  value = 0.0;
  for (x = 0; x < width; x++)
    value = value * 0.8 + (rand () & 0xff) / 255.0 * 0.2;

  /* Create x-blurred data */
  pos = 0;
  for (y = 0; y < height; y++)
    {
      for (x = 0; x < width; x++)
        {
          data[pos] = value;

          value = value * 0.9 + (rand () & 0xff) / 255.0 * 0.1;
          pos++;
        }
    }

  /* Create y-blurred data */
  for (x = 0; x < width; x++)
    {
      pos = x;
      for (y = 0; y < height; y++)
        {
          data[pos] += value;

          value = value * 0.9 + (rand () & 0xff) / 255.0 * 0.1;
          pos += width;
        }
    }

  /* Radius blur */
  for (x = 0; x < (gint) ceil (scale - 1.0); x++)
    blur (data, width, height);

  /* Normalize the data. */
  min = max = data[0];
  for (x = 0; x < width * height; x++)
    {
      if (data[x] < min) min = data[x];
      if (data[x] > max) max = data[x];
    }

  if (fabs (min - max) > 0.0001)
    {
      for (x = 0; x < width * height; x++)
        data[x] = (data[x] - min) * (1.0 / (max - min));
    }

  /* Calculate integer data. */
  data_8 = g_new (guint8, width * height);
  for (x = 0; x < width * height; x++)
    data_8[x] = RINT (data[x] * 255.0);

  g_free (data);

  return data_8;
}
