/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "interface.h"
#include "support.h"
#include "support2.h"
#include "texture.h"
#include "perlintexture.h"
#include "gtkrgbarea.h"
#include "gimpsupport.h"


static guint8 *
texture_generate (GtkWidget *widget,
                  gint       width,
                  gint       height,
                  gint       seed)
{
  gchar  *type_list[] = { "light_grain", "light_grain_2", "heavy_grain", "heavy_grain_2", "canvas", NULL };
  guint8 *texture;
  gfloat  scale;
  gint    type;

  type = get_radio_menu (widget, type_list);
  scale = get_float (widget, "scale");

  texture = NULL;
  switch (type)
    {
    case 0: /* light_grain */
      texture = texture_light_new (width, height, scale, seed);
      break;

    case 1: /* light_grain_2 */
      texture = perlin_texture_light_new (width, height, scale, seed);
      break;

    case 2: /* heavy_grain */
      texture = texture_heavy_new (width, height, scale, seed);
      break;

    case 3: /* heavy_grain_2 */
      texture = perlin_texture_heavy_new (width, height, scale, seed);
      break;

    case 4: /* canvas */
      texture = texture_canvas_new (width, height, scale, seed);
      break;
    }

  return texture;
}

void
texture_update (GtkWidget *widget)
{
  gint        i;
  GtkWidget  *preview;
  guint8     *texture;
  guint8     *dest;

  preview = lookup_widget (widget, "preview");

  texture = texture_generate (widget, 100, 100, 0);
  dest = GTK_RGB_AREA (preview)->rgb_data;
  for (i = 0; i < 100 * 100; i++)
    {
      dest[i * 3 + 0] = texture[i];
      dest[i * 3 + 1] = texture[i];
      dest[i * 3 + 2] = texture[i];
    }
  gtk_rgb_area_invalidate (GTK_RGB_AREA (preview), 0, 0, 100, 100);
  gtk_rgb_area_hard_sync (GTK_RGB_AREA (preview));

  g_free (texture);
}

void
texture_create (GtkWidget *widget)
{
  gint32       *image_id;
  GimpDrawable *drawable;
  guint8       *texture;
  gint          width, height;
  gint32        channel_id;

  image_id = lookup_data (widget, "data_image_id");
  drawable = lookup_data (widget, "data_drawable"); /* We don't use this. */
  width = gimp_image_width (*image_id);
  height = gimp_image_height (*image_id);

  channel_id = find_create_channel (*image_id, "Texture", 0, 0, 255, NULL);
  texture = texture_generate (widget, width, height, -1);
  drawable = gimp_drawable_get (channel_id);
  tiles_to_from_pixmap (texture, 1, width,
                        width, height,
                        drawable, 0, 0, FALSE);

  gimp_drawable_detach (drawable);
  gimp_drawable_merge_shadow (channel_id, TRUE);
  gimp_drawable_update (channel_id, 0, 0, width, height);

  g_free (texture);
}

gint
texture_window (gint32        image_id,
                GimpDrawable *drawable)
{
  gchar     *hscales[] = { "scale", NULL };
  GtkWidget *window;
  gint       draw_width, draw_height;
  int        argc = 1;
  char      *argv[2] = { "digi_pencil", NULL };
  char     **argv2 = argv;

#ifdef ENABLE_NLS
  bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
  textdomain (PACKAGE);
#endif

  gtk_set_locale ();
  setlocale (LC_NUMERIC, "C");
  gtk_init (&argc, &argv2);
  gdk_rgb_init ();

  gtk_widget_set_default_colormap (gdk_rgb_get_cmap ());
  gtk_widget_set_default_visual (gdk_rgb_get_visual ());

  add_pixmap_directory (g_strconcat (PACKAGE_DATA_DIR, G_DIR_SEPARATOR_S "pixmaps", NULL));

  draw_width = 800;
  draw_height = 600;

  window = create_texture_window ();
  gtk_object_set_data (GTK_OBJECT (window), "data_image_id", &image_id);
  gtk_object_set_data (GTK_OBJECT (window), "data_drawable", drawable);
  hscale_callbacks (window, hscales, GTK_SIGNAL_FUNC (texture_update), window);
  gtk_widget_show_all (window);

  gtk_widget_realize (window);
  texture_update (window);

  gtk_main ();
  gimp_image_set_active_layer (image_id, drawable->id);
  gimp_displays_flush ();
  gdk_flush ();

  return 0;
}

