/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>              /* For G_PI in GLib 2.0 */
#include <math.h>
#include <libgimp/gimpmath.h>  /* For G_PI otherwise */
#include "tmatrix.h"

#define DEGRAD(x) ((x) * G_PI / 180.0)

void
t_vector_set (TVector *vector,
              gdouble  x,
              gdouble  y,
              gdouble  z)
{
  vector->x = x;
  vector->y = y;
  vector->z = z;
}

gdouble
t_vector_magnitude_squared (TVector *vector)
{
  return vector->x * vector->x + vector->y * vector->y + vector->z * vector->z;
}

gdouble
t_vector_magnitude (TVector *vector)
{
  return sqrt (t_vector_magnitude_squared (vector));
}

void
t_vector_normalize (TVector *inout)
{
  gdouble mag;

  mag = t_vector_magnitude (inout);
  inout->x /= mag;
  inout->y /= mag;
  inout->z /= mag;
}

gdouble
t_vector_dot_product (TVector *in1,
                      TVector *in2)
{
  return in1->x * in2->x + in1->y * in2->y + in1->z * in2->z;
}

/* Create an identity matrix. */
void
t_matrix_identity (TMatrix *matrix)
{
  t_matrix_scale (matrix, 1.0);
}

/* Multiply two matrices together. */
void
t_matrix_multiply (TMatrix *in1,
                   TMatrix *in2,
                   TMatrix *out)
{
  gint i, j, k;

  for (i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
      {
        gdouble sum;

        sum = 0.0;
        for (k = 0; k < 4; k++)
          sum += in1->data[i][k] * in2->data[k][j];

        out->data[i][j] = sum;
      }
}

/*
 * Create a rotation matrix.
 *
 * This rotation matrix code is adapted from the Rotate3D function on
 * page 28 of "Advanced Graphics Programming in C and C++" by
 * Roger T. Stevens and Christopher D. Watkins.
 * 
 * Axes:
 *   x is 1
 *   y is 2
 *   z is 3
 */

void
t_matrix_rotate (TMatrix *matrix,
                 gdouble  angle,
                 gint     axis)
{
  gint   m1, m2;
  gdouble c, s;

  t_matrix_zero (matrix);

  matrix->data[axis - 1][axis - 1] = 1.0;
  matrix->data[3][3] = 1.0;

  m1 = axis % 3;
  m2 = (m1 + 1) % 3;

  c = cos (DEGRAD (angle));
  s = sin (DEGRAD (angle));

  matrix->data[m1][m1] = c;
  matrix->data[m1][m2] = s;
  matrix->data[m2][m2] = c;
  matrix->data[m2][m1] = -s;
}

/* Create a translation matrix. */
void
t_matrix_translate (TMatrix *matrix,
                    gdouble  x,
                    gdouble  y,
                    gdouble  z)
{
  t_matrix_identity (matrix);

  matrix->data[0][3] = x;
  matrix->data[1][3] = y;
  matrix->data[2][3] = z;
}

/* Transform a vector by a matrix. */
void
t_matrix_transform (TMatrix *matrix,
                    TVector *in,
                    TVector *out)
{
  out->x = matrix->data[0][0] * in->x +
           matrix->data[0][1] * in->y +
           matrix->data[0][2] * in->z +
           matrix->data[0][3];

  out->y = matrix->data[1][0] * in->x +
           matrix->data[1][1] * in->y +
           matrix->data[1][2] * in->z +
           matrix->data[1][3];

  out->z = matrix->data[2][0] * in->x +
           matrix->data[2][1] * in->y +
           matrix->data[2][2] * in->z +
           matrix->data[2][3];
}

/* Mathematically combine two matrices. */
void
t_matrix_combine (TMatrix *in1,
                  TMatrix *in2,
                  TMatrix *out)
{
  gint i;

  t_matrix_multiply (in1, in2, out);
  for (i = 0; i < 3; i++)
    out->data[i][3] = in1->data[i][3] + in2->data[i][3];
}

/* Zero a matrix. */
void
t_matrix_zero (TMatrix *matrix)
{
  gint i, j;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      matrix->data[i][j] = 0.0;
}

/* Copy a matrix. */
void
t_matrix_assign (TMatrix *in,
                 TMatrix *out)
{
  gint i, j;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      out->data[i][j] = in->data[i][j];
}

/* Create a scaling matrix. */
void
t_matrix_scale (TMatrix *matrix,
                gdouble  k)
{
  gint i;

  t_matrix_zero (matrix);

  for (i = 0; i < 3; i++)
    matrix->data[i][i] = k;
}
