/*  Physically Modeled Media Plug-In for The GIMP
 *  Copyright (c) 2000-2001 David A. Bartold
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Code taken from Terraform.
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>              /* For G_PI in GLib 2.0 */
#include <math.h>
#include <libgimp/gimpmath.h>  /* For G_PI otherwise */
#include <time.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimpintl.h>

#include "support.h"
#include "tworld.h"
#include "tperlin3d.h"

#define EARTH_RADIUS            6375.0
#define EARTH_CLOUD_HEIGHT         5.0
#define EARTH_ATMOSPHERE_HEIGHT  100.0
#define LENS_ANGLE                70.0

#define SKY_RED      0.01
#define SKY_GREEN    0.04
#define SKY_BLUE     0.18
#define HAZE_RED     0.31
#define HAZE_GREEN   0.35
#define HAZE_BLUE    0.40
#define SUN_RED      0.995
#define SUN_GREEN    0.90
#define SUN_BLUE     0.83
#define CLOUD_RED    1.0
#define CLOUD_GREEN  1.0
#define CLOUD_BLUE   1.0
#define SHADOW_RED   0.0
#define SHADOW_GREEN 0.0
#define SHADOW_BLUE  0.0

#define GAMMA 1.5
#define TOSCREEN(x) pow(x, 1.0 / (GAMMA))
#define FROMSCREEN(x) pow(x, GAMMA)

#define DEGRAD(x) ((x) * G_PI / 180.0)
#define CLIP(x,min,max) MAX (MIN (x, max), min)


typedef struct TWorld TWorld;
struct TWorld
{
  gint       width;
  gint       height;
  TPerlin3D *clouds;
  gfloat     planet_radius;
  gfloat     cloud_height;
  gfloat     camera_distance;
  TMatrix    transform;
  TVector    camera_location;
  gint       sun_x;
  gint       sun_y;
  gboolean   sun_show;
  gfloat     time;
  gdouble    horizon_color[3];
  gdouble    sky_color[3];
  gdouble    sun_color[3];
  gdouble    cloud_color[3];
  gdouble    shadow_color[3];
};

static void
t_world_free (TWorld *world)
{
  t_perlin3d_free (world->clouds);
  g_free (world);
}


static gboolean
sphere_intersect (TWorld  *world,
                  gdouble  radius,
                  TVector *camera_ray,
                  TVector *intersection,
                  gdouble *dist)
{
  TVector point;
  gdouble distance;
  gdouble min_dist;
  gdouble half_chord_sqr;

  /* Sphere is in the center of the world. */
  point.x = 0.0 - world->camera_location.x;
  point.y = 0.0 - world->camera_location.y;
  point.z = 0.0 - world->camera_location.z;

  distance = t_vector_dot_product (camera_ray, &point);
  half_chord_sqr =
    SQR (radius) + SQR (distance) - t_vector_magnitude_squared (&point);
  if (half_chord_sqr < 0.0001)
    return FALSE;

  min_dist = distance + sqrt (half_chord_sqr);
  if (min_dist < 0.0)
    return FALSE;

  if (dist != NULL)
    *dist = min_dist;

  intersection->x = world->camera_location.x + camera_ray->x * min_dist;
  intersection->y = world->camera_location.y + camera_ray->y * min_dist;
  intersection->z = world->camera_location.z + camera_ray->z * min_dist;

  return TRUE;
}

static void
set_fore (gfloat  r,
          gfloat  g,
          gfloat  b,
          guint8 *data)
{
  /*
   * Perform some contrast adjustment and clip the results before applying
   * gamma.  We adjust the contrast because the viewer expects a photograph
   * to be more vivid than the real world.  Without the adjustment, the
   * results won't resemble a photo.  Try removing the contrast adjustment
   * and see for yourself.
   */

  data[0] = pow (CLIP (r * 1.6 - 0.2, 0.0, 1.0), 1.0 / GAMMA) * 255.0;
  data[1] = pow (CLIP (g * 1.6 - 0.2, 0.0, 1.0), 1.0 / GAMMA) * 255.0;
  data[2] = pow (CLIP (b * 1.6 - 0.2, 0.0, 1.0), 1.0 / GAMMA) * 255.0;
}

static void
expose (gfloat *inout_red,
        gfloat *inout_green,
        gfloat *inout_blue,
        gfloat  value,
        gfloat  red,
        gfloat  green,
        gfloat  blue)
{
  *inout_red   = value * red + *inout_red * (1.0 - value);
  *inout_green = value * green + *inout_green * (1.0 - value);
  *inout_blue  = value * blue + *inout_blue * (1.0 - value);
}


static void
interpolate (gfloat *out_red,
             gfloat *out_green,
             gfloat *out_blue,
             gfloat  value,
             gfloat  red1,
             gfloat  green1,
             gfloat  blue1,
             gfloat  red2,
             gfloat  green2,
             gfloat  blue2)
{
  *out_red = value * red2 + (1.0 - value) * red1;
  *out_green = value * green2 + (1.0 - value) * green1;
  *out_blue = value * blue2 + (1.0 - value) * blue1;
}


static void
draw_cloud_point (TWorld *world,
                  gfloat *inout_red,
                  gfloat *inout_green,
                  gfloat *inout_blue,
                  gfloat  value,
                  gint    x,
                  gint    y,
                  gfloat  cloud_x,
                  gfloat  cloud_y,
                  gfloat  shadow_x,
                  gfloat  shadow_y)
{
  gfloat red, green, blue;
  gfloat point1, point2;
  gfloat offset_x, offset_y;

  offset_x = world->time * 0.25;
  offset_y = world->time * 0.33;

  point1 = t_perlin3d_get (world->clouds,
                           cloud_x + offset_x, cloud_y + offset_y,
                           world->time * 2.5);
  point2 = t_perlin3d_get (world->clouds,
                           shadow_x + offset_x, shadow_y + offset_y,
                           world->time * 2.5);

  /* Apply a threshold and exponentiation function */
  if (point1 < 0.525)
    point1 = 0.0;
  else
    point1 = pow ((point1 - 0.525) / (1.0 - 0.525), 0.4);

  if (point2 < 0.56)
    point2 = 0.0;
  else
    point2 = pow ((point2 - 0.56) / (1.0 - 0.56), 0.9);

  /* Apply shadow */
  interpolate (&red, &green, &blue, point2,
               world->cloud_color[0],
               world->cloud_color[1],
               world->cloud_color[2],
               world->shadow_color[0],
               world->shadow_color[1],
               world->shadow_color[2]);

  expose (inout_red, inout_green, inout_blue,
          MIN (value * point1, 1.0),
          red, green, blue);
}


static void
draw_sun (TWorld *world,
          gint    x,
          gint    y,
          gint    sky_height,
          gfloat  sun_size,
          gfloat  sun_red,
          gfloat  sun_green,
          gfloat  sun_blue,
          gfloat *inout_red,
          gfloat *inout_green,
          gfloat *inout_blue)
{
  gdouble dist;
  gdouble distance;

  dist = sqrt (world->width * world->width +
               world->height * world->height);

  distance = sqrt ((world->sun_x - x) * (world->sun_x - x) +
                   (world->sun_y - y) * (world->sun_y - y));

  if (distance < sun_size)
    {
      *inout_red = sun_red;
      *inout_green = sun_green;
      *inout_blue = sun_blue;
    }
  else if (distance < dist)
    {
      gfloat value;

      value = 1.0 - 1.0 /
              pow (200.0, (distance - sun_size) / (dist - sun_size));

      interpolate (inout_red, inout_green, inout_blue,
                   value, sun_red, sun_green, sun_blue,
                   *inout_red, *inout_green, *inout_blue);
   }
}


static void
world_render_sky (TWorld       *world,
                  GimpPixelRgn *region)
{
  gint    x, y;
  TVector ray, world_ray;
  TVector intersection;
  TVector intersection2;
  gfloat  scale;
  gfloat  offset;

  scale = 1.0 / 100.0;

  ray.z = world->camera_distance;
  for (y = region->y; y < region->y + region->h; y++)
    {
      guint8 *data;

      ray.y = (world->height - 1) - y;
      for (x = region->x; x < region->x + region->w; x++)
        {
          gboolean result1, result2;
          gfloat   red, green, blue;
          gdouble  sky_distance;
          gdouble  value;

          ray.x = x - world->width * 0.5;
          t_matrix_transform (&world->transform, &ray, &world_ray);
          t_vector_normalize (&world_ray);

          result1 =
            sphere_intersect (world, world->planet_radius +
                                     world->cloud_height,
                              &world_ray, &intersection, &sky_distance);

          result2 =
            sphere_intersect (world, world->planet_radius +
                                     world->cloud_height * 0.9,
                              &world_ray, &intersection2, NULL);

          value = (intersection.y - world->camera_location.y) /
                  (world->planet_radius + world->cloud_height -
                   world->camera_location.y);

          interpolate (&red, &green, &blue, pow (value, 200.0),
                       world->horizon_color[0],
                       world->horizon_color[1],
                       world->horizon_color[2],
                       world->sky_color[0],
                       world->sky_color[1],
                       world->sky_color[2]);

          if (world->sun_show)
            draw_sun (world, x, y,
                      world->height,
                      world->width / 35,
                      world->sun_color[0],
                      world->sun_color[1],
                      world->sun_color[2],
                      &red, &green, &blue);

          if (result1 == TRUE && result2 == TRUE)
            {
              /*
               * Add an offset to the x and z coordinates because
               *  Perlin noise does not accept negative numbers
               */
              offset = world->planet_radius + world->cloud_height;

              draw_cloud_point (world, &red, &green, &blue,
                pow (value, 100.0), x, y,
                (intersection.x + offset) * scale,
                (intersection.z + offset) * scale,
                (intersection2.x + offset) * scale,
                (intersection2.z + offset) * scale);
            }

          data = &region->data[(y - region->y) * region->rowstride +
                               (x - region->x) * region->bpp];
          set_fore (red, green, blue, data);

          if (region->bpp == 4)
            data[3] = 255;
        }
    }
}


static void
from_screen (gdouble *out,
             GimpRGB *rgb)
{
  out[0] = FROMSCREEN (rgb->r);
  out[1] = FROMSCREEN (rgb->g);
  out[2] = FROMSCREEN (rgb->b);
}


static TWorld *
tworld_create (gint          width,
               gint          height,
               SkyValues    *values)
{
  TMatrix  temp1, temp2;
  TWorld  *world;
  gfloat   amplitudes[] = { 1.0, 0.5, 0.25, 0.125, 0.0625,
                            0.03125, 0.05, 0.05, 0.04, 0.0300 };

  world = g_new0 (TWorld, 1);
  world->width = width;
  world->height = height;
  world->clouds = t_perlin3d_new_full (10, 16.0, amplitudes, values->seed);
  world->planet_radius = EARTH_RADIUS;
  world->cloud_height = EARTH_CLOUD_HEIGHT;
  world->camera_distance = width * 0.5 /
                           tan (DEGRAD (LENS_ANGLE) * 0.5);
  world->sun_x = RINT ((width - 1) * values->sun_x);
  world->sun_y = RINT ((height - 1) * values->sun_y);
  world->sun_show = values->sun_show;
  world->time = values->time;
  from_screen (world->horizon_color, &values->horizon_color);
  from_screen (world->sky_color, &values->sky_color);
  from_screen (world->sun_color, &values->sun_color);
  from_screen (world->cloud_color, &values->cloud_color);
  from_screen (world->shadow_color, &values->shadow_color);

  t_matrix_rotate (&temp1, values->tilt, 1);
  t_matrix_rotate (&temp2, values->rotation, 2);
  t_matrix_combine (&temp2, &temp1, &world->transform);
  t_vector_set (&world->camera_location, 0.0, EARTH_RADIUS + 0.2, 0.0);

  return world;
}


static void
gimp_rgb_set (GimpRGB *rgb, gdouble r, gdouble g, gdouble b)
{
  rgb->r = r;
  rgb->g = g;
  rgb->b = b;
  rgb->a = 0.0;
}

void
sky_values_init (SkyValues *values)
{
  values->tilt = 0.0;
  values->rotation = 0.0;
  values->seed = -1;
  values->sun_show = TRUE;
  values->sun_x = 0.2;
  values->sun_y = 0.2;
  values->time = 0.0;
  gimp_rgb_set (&values->horizon_color,
    TOSCREEN (HAZE_RED), TOSCREEN (HAZE_GREEN), TOSCREEN (HAZE_BLUE));
  gimp_rgb_set (&values->sky_color,
    TOSCREEN (SKY_RED), TOSCREEN (SKY_GREEN), TOSCREEN (SKY_BLUE));
  gimp_rgb_set (&values->sun_color,
    TOSCREEN (SUN_RED), TOSCREEN (SUN_GREEN), TOSCREEN (SUN_BLUE));
  gimp_rgb_set (&values->cloud_color,
    TOSCREEN (CLOUD_RED), TOSCREEN (CLOUD_GREEN), TOSCREEN (CLOUD_BLUE));
  gimp_rgb_set (&values->shadow_color,
    TOSCREEN (SHADOW_RED), TOSCREEN (SHADOW_GREEN), TOSCREEN (SHADOW_BLUE));
}


void
sky_render (GimpDrawable *drawable,
            SkyValues    *values)
{
  TWorld       *world;
  GimpPixelRgn  region;
  gpointer      ptr;
  gint          progress, max_progress;

  world = tworld_create (drawable->width, drawable->height, values);

  gimp_pixel_rgn_init (&region, drawable,
                       0, 0,
                       drawable->width, drawable->height,
                       TRUE, TRUE);

  gimp_progress_init (N_("Sky"));
  progress = 0;
  max_progress = drawable->width * drawable->height;
  for (ptr = gimp_pixel_rgns_register (1, &region);
       ptr != NULL;
       ptr = gimp_pixel_rgns_process (ptr))
    {
      world_render_sky (world, &region);
      progress += region.w * region.h;
      gimp_progress_update (((gfloat) progress) / max_progress);
    }

  t_world_free (world);
}


void
sky_preview (guint8    *data,
             guint      width,
             guint      height,
             SkyValues *values)
{
  TWorld       *world;
  GimpPixelRgn  region;

  world = tworld_create (width, height, values);

  region.data = data;
  region.drawable = NULL;
  region.bpp = 3;
  region.rowstride = width * region.bpp;
  region.x = 0;
  region.y = 0;
  region.w = width;
  region.h = height;
  region.dirty = FALSE;
  region.shadow = FALSE;
  region.process_count = 0;

  world_render_sky (world, &region);

  t_world_free (world);
}
