/*
    The Resynthesizer - A GIMP plug-in for resynthesizing textures
    Copyright (C) 2000  Paul Francis Harrison

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

//Number of patches in leaf node of kd-tree
#define LEAF_SIZE 16

/* Utility classes ===================================================== */

//Pixel channel type
typedef guint8 Pixelel;

typedef int bool;
const bool true = 1, false = 0;

//Intersperse the bits from several bytes (used to sort colors
//  allowing a fairly primitive quantization)
guint entwine(Pixelel *data, int n_el) {
  int i,j;
  guint result = 0;
  for(i=0;i<8;i++)
    for(j=0;j<n_el;j++)
      if ((data[j]>>i) & 1)
        result += 1 << (i*n_el+j);
  return result;
}

//Structure to hold an "entwine"ified pixels  
typedef struct {
  int x,y;
  guint twine;
} Twine_list_item;

int twine_list_item_sorter(const void *a, const void *b) {
  if ( ((Twine_list_item*)a)->twine < ((Twine_list_item*)b)->twine)
    return -1;
  else if ( ((Twine_list_item*)a)->twine > ((Twine_list_item*)b)->twine)
    return 1;
  else
    return 0;
}

typedef struct {
  int x,y;
} Coordinates;

//Structure+functions used to calculate the number of bits required to
//encode some normally distributed numbers
typedef struct {
  guint *x;
  guint *x_squared;
  int n_els, n;
} Entropy_accumulator;

Entropy_accumulator *entropy_accumulator_new(int n_els) {
  Entropy_accumulator *neo = g_new(Entropy_accumulator,1);
  neo->n_els     = n_els;
  neo->x         = g_new(guint,n_els);
  neo->x_squared = g_new(guint,n_els);

  return neo;
}

void entropy_accumulator_free(Entropy_accumulator *e) {
  g_free(e->x);
  g_free(e->x_squared);
  g_free(e);
}

void entropy_accumulator_calculate(Entropy_accumulator *e,
                                   double *bits,double *entropy) {
  int i;

  if (e->n <= 1) {
    *entropy = (double)(e->n_els)*sizeof(Pixelel)*8.0;
    *bits = e->n*(*entropy);
    return;
  }

  *entropy = *bits = 0.0;
    
  for(i=0;i<e->n_els;i++) {
    double mu    = (double)(e->x[i]) / e->n;
    double sigma, L;
    
    if (e->x_squared[i]-e->x[i]*mu >= 0.0) {
      sigma = sqrt( (e->x_squared[i]-e->x[i]*mu) / (e->n-1) ); 
      if (sigma < 1.0)
        sigma = 1.0;
    } else
      sigma = 1.0;
      
    L =
      e->n/2.0*log(2*3.1415926535) + e->n/2.0*log(sigma*sigma) +
      1.0/(2.0*sigma*sigma)*(e->x_squared[i]-e->x[i]*mu);

    *bits += log(256)+log(sigma)+L+
             0.5*log((double)(e->n)*e->n/(2.0*pow(sigma,6.0)))-0.5*log(12)+0.5;

    *entropy += 0.5*log(2*3.14159265)+0.5+log(sigma);
  }

  //Convert from nits to bits
  *bits /= log(2);
  *entropy /= log(2);
}

void entropy_accumulator_clear(Entropy_accumulator *e) {
  int i;
  for(i=0;i<e->n_els;i++) {
    e->x[i]         = 0;
    e->x_squared[i] = 0;
  }
  e->n = 0;
}

void entropy_accumulator_add(Entropy_accumulator *e, Pixelel *data) {
  int i;
  for(i=0;i<e->n_els;i++) {
    e->x[i]         += data[i];
    e->x_squared[i] += (guint)(data[i])*(guint)(data[i]);
  }
  e->n++;
}

//Bitmap class with three dimensions (width, height, number of channels)
typedef struct {
  int width, height, depth, element_size;
  void *data;
} Bitmap;

Bitmap *bitmap_new(int element_size) { 
  Bitmap *neo = g_new(Bitmap,1);
  neo->element_size = element_size;
  neo->data = 0;

  return neo;
}

void bitmap_free(Bitmap *b) {
  g_free(b->data);
  g_free(b);
}

void bitmap_size(Bitmap *b, int w,int h,int d) {
  b->width = w; 
  b->height = h; 
  b->depth = d;
    
  g_free(b->data);
  b->data = g_malloc(w*h*d*b->element_size);
}

#define bitmap_at(b,type,x,y) \
  ( (type*)((b)->data)+(x)*(b)->depth+(y)*(b)->width*(b)->depth )

void bitmap_to(Bitmap *b, GimpDrawable *drawable, int x1,int y1, int src_layer) {
  int i,j;
  GimpPixelRgn region;
  guchar *img;

  gimp_pixel_rgn_init(&region, drawable, x1,y1,b->width,b->height, TRUE,TRUE);

  img = g_new(guchar,b->width*b->height*b->depth);
  for(i=0;i<b->width*b->height;i++)
    for(j=0;j<drawable->bpp;j++)
      img[i*drawable->bpp+j] = ((Pixelel*)(b->data))[i*b->depth+src_layer+j];

  gimp_pixel_rgn_set_rect(&region, img, x1,y1,b->width,b->height);

  g_free(img);
}

void bitmap_from(Bitmap *b, GimpDrawable *drawable, 
                 int x1,int y1, int dest_layer) {
  int i,j;
  GimpPixelRgn region;
  guchar *img;

  gimp_pixel_rgn_init(&region, drawable, x1,y1,b->width,b->height, FALSE,FALSE);

  img = g_new(guchar,b->width*b->height*b->depth);
  gimp_pixel_rgn_get_rect(&region, img, x1,y1,b->width,b->height);

  for(i=0;i<b->width*b->height;i++)
    for(j=0;j<(int)(drawable->bpp);j++)
      ((Pixelel*)(b->data))[i*b->depth+dest_layer+j] = img[i*drawable->bpp+j];

  g_free(img);
}

//Get a drawable and possibly its selection mask from the GIMP
void fetch_image_and_mask(GimpDrawable *drawable, Bitmap *image, int bytes, 
                          Bitmap *mask, Pixelel default_mask_value) {
  int x,y,x1,y1,x2,y2;
  Bitmap *temp_mask;
  int sel_id;
  int xoff,yoff;
  int has_selection;
  GimpDrawable *mask_drawable;

  bitmap_size(image,drawable->width,drawable->height,bytes);
  bitmap_size(mask,drawable->width,drawable->height,1);
  
  bitmap_from(image,drawable,0,0,0);
  
  has_selection = gimp_drawable_mask_bounds(drawable->id,&x1,&y1,&x2,&y2);
  
  if (!has_selection) {
    memset(mask->data, default_mask_value, mask->width*mask->height*sizeof(Pixelel));
    return;
  }

  temp_mask = bitmap_new(sizeof(Pixelel));
  bitmap_size(temp_mask,x2-x1,y2-y1,1);

  sel_id = gimp_image_get_selection(gimp_drawable_image_id(drawable->id));
  mask_drawable = gimp_drawable_get(sel_id);

  gimp_drawable_offsets(drawable->id, &xoff, &yoff);

  bitmap_from(temp_mask, mask_drawable, x1+xoff,y1+yoff, 0);

  gimp_drawable_detach(mask_drawable);

  memset(mask->data, 0, mask->width*mask->height*sizeof(Pixelel));
  for(y=0;y<temp_mask->height;y++)
    for(x=0;x<temp_mask->width;x++)
      bitmap_at(mask,Pixelel,x+x1,y+y1)[0] = 
          bitmap_at(temp_mask,Pixelel,x,y)[0];

  bitmap_free(temp_mask);
}

//Structure used to hold information about pixels in the output
typedef struct {
  bool use;
  bool done;
  int priority;

  int heap_index;
} Status;

//Heap used to prioritize the selection of pixels in the output
typedef struct {
  Status **data;
  int n_els;
} Heap;

void heap_put(Heap *h, Status *item,int index) {
  h->data[index] = item;
  if (item)
    item->heap_index = index;
}

Heap *heap_new(Status *items,int n_items) {
  int i;
  Heap *h = g_new(Heap,1);
  h->n_els = n_items;
  h->data  = g_new(Status*,n_items);
  
  for(i=0;i<n_items;i++) 
    heap_put(h,items+i,i);

  return h;
}

void heap_free(Heap *h) {
  g_free(h->data);
  g_free(h);
}

int heap_parent(int i) { return (i+1)/2-1; }
int heap_left_child(int i) { return (i+1)*2-1; }
int heap_right_child(int i) { return (i+1)*2; }

void heap_promote(Heap *h, Status *item) {
  int i = item->heap_index;
  while(i > 0 && 
          (!h->data[heap_parent(i)] || 
	   h->data[heap_parent(i)]->priority < item->priority)) {
    heap_put(h, h->data[heap_parent(i)],i);
    i = heap_parent(i);
  }
  heap_put(h, item, i);
}

void heap_become_heapish(Heap *h) {
  int i;
  for(i=0;i<h->n_els;i++)
    if (h->data[i]->done)
      h->data[i] = 0;
    else
      heap_promote(h, h->data[i]);
}

Status *heap_pop_top(Heap *h) {
  Status *top = h->data[0];
    
  int i=0;
  for(;;) {
    int next = heap_left_child(i);
    if (next >= h->n_els) {
      heap_put(h, 0,i);
      break;
    }

    if (heap_right_child(i) < h->n_els &&
        h->data[heap_right_child(i)] != NULL &&
        (!h->data[next] || 
	 h->data[next]->priority < h->data[heap_right_child(i)]->priority))
      next = heap_right_child(i);
      
    heap_put(h,h->data[next],i);
    i = next;
      
    if (!h->data[i])
      break;
  }

  return top;
}

/* k-d tree  ======================================== */

// k-d trees are a way to do fast(ish) closest vector
// searches

typedef struct {
  int offset_no;
  Pixelel value;
} Pattern;

typedef struct {
  int x,y,el, offset;
  guint weight;

  //So we can sort from largest to smallest weight 
  // (provides a good speedup as distance calculations can often be aborted earlier)
  //bool operator < (const Offset &other) const { return weight > other.weight; }
} Offset;

int offset_sorter(const void *a, const void *b) {
  if ( ((Offset*)a)->weight > ((Offset*)b)->weight)
    return -1;
  else if ( ((Offset*)a)->weight < ((Offset*)b)->weight)
    return 1;
  else
    return 0;
}

struct _Node;

typedef struct {
  struct _Node *node;
  int best_dist;

  //bool operator < (const Queue_item &other) const { 
  //  return best_dist > other.best_dist; 
  //}
} Queue_item;


// Globals to avoid passing around these things incessantly 

// Array of Offset
GArray *Offsets;

// Array of Pattern
GArray *Patterns;

//More globals... array of Queue_item
GArray *node_queue;
int node_queue_count;

int cutoff_dist;
Pixelel *best;

void set_up_tree_globals() {
  Offsets    = g_array_new(FALSE,FALSE,sizeof(Offset));
  Patterns   = g_array_new(FALSE,FALSE,sizeof(Pattern));
  node_queue = g_array_new(FALSE,FALSE,sizeof(Queue_item)); 
}

void clean_up_tree_globals() {
  g_array_free(Offsets,TRUE);
  g_array_free(Patterns,TRUE);
  g_array_free(node_queue,TRUE);
}

void node_queue_upheap(int i) {
  Queue_item item = g_array_index(node_queue,Queue_item,i);

  for(;;) {
    int parent;
    
    if (i == 0)
      break;
      
    parent = heap_parent(i);
    if (g_array_index(node_queue,Queue_item,parent).best_dist <
        item.best_dist)
      break;

    g_array_index(node_queue,Queue_item,i) =
      g_array_index(node_queue,Queue_item,parent);

    i = parent;
  }
    
  g_array_index(node_queue,Queue_item,i) = item;
}

void node_queue_enqueue(Queue_item item) {
  if (node_queue_count == node_queue->len)
    g_array_set_size(node_queue,(node_queue->len?node_queue->len*2:1));

  g_array_index(node_queue,Queue_item,node_queue_count++) = item;
  node_queue_upheap(node_queue_count-1);
}

void node_queue_pop_top() {
  int i = 0;
  for(;;) {
    int child,right_child;

    child = heap_left_child(i);
    if (child >= node_queue_count)
      break;

    right_child = heap_right_child(i);
    if (right_child < node_queue_count &&
        g_array_index(node_queue,Queue_item,right_child).best_dist <
        g_array_index(node_queue,Queue_item,child).best_dist)
      child = right_child;

    g_array_index(node_queue,Queue_item,i) =
      g_array_index(node_queue,Queue_item,child);

    i = child;
  }

  g_array_index(node_queue,Queue_item,i) =
    g_array_index(node_queue,Queue_item,--node_queue_count);

  node_queue_upheap(i);
}

typedef struct {
  Pixelel min, max;
} Bound;

struct _Node {
  gboolean is_leaf;
  Bound *bounds;

  union {
    struct {
      Pixelel **patches;
      int n_patches;
    } leaf;
    struct {
      struct _Node *left_child, *right_child;
    } inner;
  } data;
};

typedef struct _Node Node;

void node_free(Node *node) {
  g_free(node->bounds);

  if (node->is_leaf)
    g_free(node->data.leaf.patches);
  else {
    node_free(node->data.inner.left_child);
    node_free(node->data.inner.right_child);
  }

  g_free(node);
}

void node_add_to_queue(Node *n) {
  Queue_item item;
  int i, dist = 0;
  for(i=0;i < Patterns->len;i++) {
    int offset_no = g_array_index(Patterns,Pattern,i).offset_no;
      
    if (g_array_index(Patterns,Pattern,i).value < n->bounds[offset_no].min) {
      dist += g_array_index(Offsets,Offset,offset_no).weight * 
        (n->bounds[offset_no].min-g_array_index(Patterns,Pattern,i).value);
      if (dist >= cutoff_dist)
	return;
    } else if (g_array_index(Patterns,Pattern,i).value > 
               n->bounds[offset_no].max) {
      dist += g_array_index(Offsets,Offset,offset_no).weight * 
        (g_array_index(Patterns,Pattern,i).value-n->bounds[offset_no].max);
      if (dist >= cutoff_dist)
        return;
    }
  }

  item.node = n;
  item.best_dist = dist;
  node_queue_enqueue(item);
}

void node_do_it(Node *n) {
  if (n->is_leaf) {
    int i,j;
  
    for(i=0;i<n->data.leaf.n_patches;i++) {
      int dist = 0;
      for(j=0;j<Patterns->len;j++) {
	int offset_no = g_array_index(Patterns,Pattern,j).offset_no;
	
        dist += g_array_index(Offsets,Offset,offset_no).weight * 
	  abs((int)(n->data.leaf.patches
	          [i][g_array_index(Offsets,Offset,offset_no).offset]) - 
	      (int)(g_array_index(Patterns,Pattern,j).value));

	if (dist > cutoff_dist)
	  goto skip;
      }

      best = n->data.leaf.patches[i];
      cutoff_dist = dist;

      skip: ;
    }
  
  } else {
    
    node_add_to_queue(n->data.inner.left_child);
    node_add_to_queue(n->data.inner.right_child);
  
  }
}

Pixelel *query_tree(Node *tree) {
  //Clear the pq
  node_queue_count = 0;

  cutoff_dist = 1<<30;

  node_do_it(tree);

  while(node_queue_count) {
    Queue_item item = g_array_index(node_queue,Queue_item,0);

    if (item.best_dist > cutoff_dist)
      break;

    node_queue_pop_top();
    node_do_it(item.node);
  }

  return best;
}

Node *build_tree(Pixelel **items,int n_items) {
  int i,j;
  Node *node = g_new(Node,1);
  
  node->is_leaf = n_items <= LEAF_SIZE; 

  if (node->is_leaf) {
    node->data.leaf.patches = g_new(Pixelel*,n_items);
    node->data.leaf.n_patches = n_items;
    for(i=0;i<n_items;i++)
      node->data.leaf.patches[i] = items[i];
  }

  node->bounds = g_new(Bound,Offsets->len);

  for(i=0;i<Offsets->len;i++) {
    node->bounds[i].min = 255;
    node->bounds[i].max = 0;
  }

  for(i=0;i<n_items;i++)
    for(j=0;j<Offsets->len;j++) {
      int offset = g_array_index(Offsets,Offset,j).offset;

      if (items[i][offset] < node->bounds[j].min)
        node->bounds[j].min = items[i][offset];
      if (items[i][offset] > node->bounds[j].max)
        node->bounds[j].max = items[i][offset];
    }

  if (!node->is_leaf) {
    int best = 0;
    int best_offset, lower, upper, start_upper, pivot;
    Pixelel *temp;

    for(i=1;i<Offsets->len;i++)
      if (g_array_index(Offsets,Offset,i).weight*
            (node->bounds[i].max-node->bounds[i].min) >
          g_array_index(Offsets,Offset,best).weight*
	    (node->bounds[best].max-node->bounds[best].min))
	best = i;

    best_offset = g_array_index(Offsets,Offset,best).offset;

    lower=0; 
    upper=n_items-1;
    for(;;) {
      Pixelel p_val = items[lower][best_offset];
      start_upper = lower+1;

      for(i=lower+1;i<=upper;i++)
        if (items[i][best_offset] < p_val) {
	  Pixelel *temp = items[i];
	  items[i] = items[start_upper];
	  items[start_upper] = temp;
	  start_upper++;
	}

      pivot = start_upper-1;
      temp = items[lower];
      items[lower] = items[pivot];
      items[pivot] = temp;

      if (pivot == n_items/2)
        break;

      if (pivot < n_items/2)
        lower = pivot+1;
      else
        upper = pivot-1;
    }

    node->data.inner.left_child = 
      build_tree(items,n_items/2);
    node->data.inner.right_child =
      build_tree(items+n_items/2,n_items-n_items/2);
  }

  return node;
}

/* Parameters ========================================= */

typedef struct {
  bool h_tile, v_tile;
  bool use_border;
  
  gint32 corpus_id, input_map_id, output_map_id;

  double random_weight, map_weight;
} Parameters;

bool wrap_or_clip(Parameters *parameters, Bitmap *image, int *xx,int *yy) {
  if (*xx < 0) {
    if (parameters->h_tile)
      *xx += image->width;
    else
      return false;
  }
  
  if (*xx >= image->width) {
    if (parameters->h_tile)
      *xx -= image->width;
    else
      return false;
  }
  
  if (*yy < 0) {
    if (parameters->v_tile)
      *yy += image->height;
    else
      return false;
  }
  
  if (*yy >= image->height) {
    if (parameters->v_tile)
      *yy -= image->height;
    else
      return false;
  }

  return true;
}

/* Interface ======================================== */

void click_func(GtkWidget *widget, gpointer data) {
  if (data)
    *((bool*)data) = true;
}

gint delete_func(GtkWidget *widget, GdkEvent *event, gpointer data) {
  if (data)
    *((bool*)data) = true;
  return TRUE;
}

typedef struct {
  GtkWidget *window;
  GtkWidget *preview;
  bool gray;
  int width, height;

  gboolean cancelled;

  Bitmap *bitmap;
} Progress_monitor;

Progress_monitor *progress_monitor_new(Bitmap *bitmap, bool gray) {
  double factor;
  GtkWidget *button_box, *cancel;
  Progress_monitor *p = g_new(Progress_monitor,1);

  p->gray = gray;
  p->bitmap = bitmap;

  factor = sqrt(9000.0 / (bitmap->width * bitmap->height));
  p->width = (int)(bitmap->width * factor);
  p->height = (int)(bitmap->height * factor);

  p->cancelled = false;
   
  p->window = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(p->window), "Resynthesize");
  gtk_signal_connect(GTK_OBJECT(p->window), "delete_event",
    GTK_SIGNAL_FUNC(delete_func), &p->cancelled);

  button_box = gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(button_box), 4);
  gtk_box_pack_end(GTK_BOX(GTK_DIALOG(p->window)->action_area),button_box, 
    FALSE,FALSE,0);

  cancel = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(cancel), "clicked",
    GTK_SIGNAL_FUNC(click_func), &p->cancelled);
  gtk_box_pack_start(GTK_BOX(button_box), cancel, FALSE,FALSE,0);

  gtk_preview_set_gamma(gimp_gamma());
    
  p->preview = 
    gtk_preview_new(gray ? GTK_PREVIEW_GRAYSCALE : GTK_PREVIEW_COLOR);
  gtk_preview_size(GTK_PREVIEW(p->preview), p->width, p->height);
  gtk_preview_set_expand(GTK_PREVIEW(p->preview), FALSE);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(p->window)->vbox), p->preview, 
    TRUE, TRUE, 4);

  gtk_widget_show_all(p->window);

  return p;
}

void progress_monitor_free(Progress_monitor *p) {
  gtk_widget_destroy(p->window);
  gdk_flush();

  g_free(p);
}

gboolean progress_monitor_aborted(Progress_monitor *p) {
  return p->cancelled;
}

void progress_monitor_update(Progress_monitor *p) {
  int x,y;
  guchar *row = g_new(guchar,p->gray ? p->width : p->width*3);

  if (p->gray)
    for(y=0;y<p->height;y++) {
      for(x=0;x<p->width;x++) 
        row[x] = 
          bitmap_at(p->bitmap,Pixelel,
	    x * p->bitmap->width / p->width,
	    y * p->bitmap->height / p->height)[0];
      gtk_preview_draw_row(GTK_PREVIEW(p->preview), &(row[0]), 0,y,p->width);
    }
  else
    for(y=0;y<p->height;y++) {
      for(x=0;x<p->width;x++) {
	guchar *temp = 
	  bitmap_at(p->bitmap,Pixelel,
	    x * p->bitmap->width / p->width,y * p->bitmap->height / p->height);
	row[x*3+0] = temp[0];
	row[x*3+1] = temp[1];
	row[x*3+2] = temp[2];
      }
      gtk_preview_draw_row(GTK_PREVIEW(p->preview), &(row[0]), 0,y,p->width);
    }

  gtk_widget_draw(p->preview,0);

  g_free(row);
}

void click_func_quit(GtkWidget *widget, gpointer data) {
  if (data)
    *((bool*)data) = true;
  gtk_main_quit();
}

gint delete_func_quit(GtkWidget *widget, GdkEvent *event, gpointer data) {
  if (data)
    *((bool*)data) = true;
  gtk_main_quit();
  return TRUE;
}

gint desired_corpus_bpp;
gint corpus_menu_constrain(gint32 image_id, gint32 drawable_id, gpointer data) {
  return
    drawable_id == -1 || (
      (gimp_drawable_is_rgb(drawable_id) || gimp_drawable_is_gray(drawable_id))
      && gimp_drawable_bpp(drawable_id) == desired_corpus_bpp);
}

gint map_menu_constrain(gint32 image_id, gint32 drawable_id, gpointer data) {
  return
    drawable_id == -1 || 
    gimp_drawable_is_rgb(drawable_id) || 
    gimp_drawable_is_gray(drawable_id);
}

void menu_callback(gint32 id, gpointer data) {
  *((gint32*)data) = id;
}

void checkbutton_callback(GtkWidget *widget, gpointer data) {
  *((bool*)data) = GTK_TOGGLE_BUTTON(widget)->active;
}

GtkWidget *map_input, *map_output, *map_slider;

void map_checkbutton_callback(GtkWidget *widget, gpointer data) {
  *((bool*)data) = GTK_TOGGLE_BUTTON(widget)->active;
  gtk_widget_set_sensitive(map_input,  *((bool*)data));
  gtk_widget_set_sensitive(map_output, *((bool*)data));
  gtk_widget_set_sensitive(map_slider, *((bool*)data));
}

GtkWidget *make_image_menu(char *text, GimpConstraintFunc constraint,
    gint32 *value) {
  GtkWidget *box, *label, *menu, *option_menu;

  box = gtk_hbox_new(FALSE,0);
  gtk_container_set_border_width(GTK_CONTAINER(box), 4);
  
  label = gtk_label_new(text);
  gtk_box_pack_start(GTK_BOX(box), label, FALSE,FALSE,0);

  menu = gimp_drawable_menu_new(constraint,
    menu_callback, value, *value);

  option_menu = gtk_option_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(option_menu), menu);
  gtk_box_pack_end(GTK_BOX(box), option_menu, FALSE,FALSE,0);

  return box;
}

GtkWidget *make_checkbutton(char *label, bool *state) {
  GtkWidget *checkbutton = gtk_check_button_new_with_label(label);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton), *state);
  gtk_signal_connect(GTK_OBJECT(checkbutton), "toggled",
    GTK_SIGNAL_FUNC(checkbutton_callback), state);

  return checkbutton;
}

GtkWidget *make_slider(char *label, double *variable) {
  GtkObject *adjustment;
  GtkWidget *table;

  table = gtk_table_new(1,3,FALSE);
  gtk_table_set_col_spacings(GTK_TABLE (table), 4);
  gtk_container_border_width(GTK_CONTAINER (table), 4);
  
  adjustment = 
    gimp_scale_entry_new(GTK_TABLE(table), 0,0, label, 125,50, 
                         *variable, 0.0,1.0,0.01,0.1,2,TRUE,0.0,1.0, NULL,NULL);

  gtk_signal_connect (GTK_OBJECT(adjustment), "value_changed",
                      GTK_SIGNAL_FUNC(gimp_double_adjustment_update),
		      variable);
  
  return table;				      
}

/* Open a dialog, get user to select options */
bool get_parameters_by_asking(Parameters *param, int default_drawable) {
  GtkWidget *window, *button_box, *ok, *cancel, *main_box, *input_frame,
            *corpus, *output_frame, *output_box, *h_tile_button,
	    *v_tile_button, *border_button, *randomness_slider, 
	    *map_frame, *map_box, *map_button;
  bool map;
  bool cancelled = false;
 
  window = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(window), "Resynthesize");
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
    GTK_SIGNAL_FUNC(delete_func_quit), &cancelled);

  button_box = gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(button_box), 4);
  gtk_box_pack_end(GTK_BOX(GTK_DIALOG(window)->action_area),button_box, 
    FALSE,FALSE,0);

  ok = gtk_button_new_with_label("OK");
  GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(ok), "clicked",
    GTK_SIGNAL_FUNC(click_func_quit), 0);
  gtk_box_pack_start(GTK_BOX(button_box), ok, FALSE,FALSE,0);
  gtk_widget_grab_default(ok);

  cancel = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(cancel), "clicked",
    GTK_SIGNAL_FUNC(click_func_quit), &cancelled);
  gtk_box_pack_start(GTK_BOX(button_box), cancel, FALSE,FALSE,0);

  main_box = GTK_DIALOG(window)->vbox;

  input_frame = gtk_frame_new("Input");
  gtk_frame_set_shadow_type(GTK_FRAME(input_frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(input_frame), 6);
  gtk_box_pack_start(GTK_BOX(main_box), input_frame, TRUE, TRUE, 0);
    
  desired_corpus_bpp = gimp_drawable_bpp(default_drawable);
  corpus = make_image_menu("Texture source: ", corpus_menu_constrain,
    &param->corpus_id);
  gtk_container_add(GTK_CONTAINER(input_frame), corpus);
  
  output_frame = gtk_frame_new("Output");
  gtk_frame_set_shadow_type(GTK_FRAME(output_frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(output_frame), 6);
  gtk_box_pack_start(GTK_BOX(main_box), output_frame, TRUE, TRUE, 0);

  output_box = gtk_vbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(output_frame), output_box);

  h_tile_button = make_checkbutton("Make horizontally tileable",
    &param->h_tile);
  gtk_box_pack_start_defaults(GTK_BOX(output_box), h_tile_button);

  v_tile_button = make_checkbutton("Make vertically tileable",
    &param->v_tile);
  gtk_box_pack_start_defaults(GTK_BOX(output_box), v_tile_button);

  border_button = make_checkbutton("Fit output to bordering pixels",
    &param->use_border);
  gtk_box_pack_start_defaults(GTK_BOX(output_box), border_button);

  randomness_slider = make_slider("Randomness:",
    &param->random_weight);
  gtk_box_pack_start_defaults(GTK_BOX(output_box), randomness_slider);
 
  map_frame = gtk_frame_new("Texture property maps");
  gtk_frame_set_shadow_type(GTK_FRAME(map_frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(map_frame), 6);
  gtk_box_pack_start(GTK_BOX(main_box), map_frame, TRUE, TRUE, 0);

  map_box = gtk_vbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(map_frame), map_box);
 
  map = (param->input_map_id != -1 && param->output_map_id != -1);
  map_button = gtk_check_button_new_with_label("Use texture property maps");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(map_button), map);
  gtk_signal_connect(GTK_OBJECT(map_button), "toggled",
    GTK_SIGNAL_FUNC(map_checkbutton_callback), &map);
  gtk_box_pack_start_defaults(GTK_BOX(map_box), map_button);
    
  map_input = make_image_menu("Input map: ", map_menu_constrain,
    &param->input_map_id);
  gtk_widget_set_sensitive(map_input, map);
  gtk_box_pack_start_defaults(GTK_BOX(map_box), map_input);

  map_output = make_image_menu("Output map: ", map_menu_constrain,
    &param->output_map_id);
  gtk_widget_set_sensitive(map_output, map);
  gtk_box_pack_start_defaults(GTK_BOX(map_box), map_output);

  map_slider = make_slider("Map importance:",
    &param->map_weight);
  gtk_widget_set_sensitive(map_slider, map);
  gtk_box_pack_start_defaults(GTK_BOX(map_box), map_slider);
  
  gtk_widget_show_all(window);

  gtk_main();

  gtk_widget_destroy(window);

  gdk_flush();

  if (!map) {
    param->input_map_id = -1;
    param->output_map_id = -1;
  }
  
  if (cancelled)
    return false;

  return true;
}

/* Restore the last parameters used from deep in the bowels of
   the main GIMP app */
bool get_last_parameters(Parameters *param, int default_drawable) {
  /* Defaults in case this is our first run */
  param->corpus_id = -1;
  param->input_map_id = -1;
  param->output_map_id = -1;
  param->v_tile = true;
  param->h_tile = true;
  param->use_border = false;
  param->random_weight = 0.25;
  param->map_weight    = 0.75;
 
  gimp_get_data("plug_in_resynthesizer", param);

  /* If image was resynthesized from itself last time, resythesize this 
     image from itself too */
  if (param->corpus_id == -1)
    param->corpus_id = default_drawable;

  return true;
}

/* Convert argument list into parameters */
bool get_parameters_from_list(Parameters *param, int n_args, GimpParam *args) {
  if (n_args != 11)
    return false;

  param->v_tile        = args[3].data.d_int32;
  param->h_tile        = args[4].data.d_int32;
  param->use_border    = args[5].data.d_int32;
  param->corpus_id     = args[6].data.d_int32;
  param->input_map_id  = args[7].data.d_int32;
  param->output_map_id = args[8].data.d_int32;
  param->random_weight = args[9].data.d_float;
  param->map_weight    = args[10].data.d_float;

  return true;
}

/* Talking to GIMP stuff */

static void query();
static void run(gchar *name,
                gint nparams,
		GimpParam *param,
		gint *nreturn_vals,
		GimpParam **return_vals);

GimpPlugInInfo PLUG_IN_INFO = {
  NULL, /* init_proc */
  NULL, /* quit_proc */
  query, /* query_proc */
  run, /* run_proc */
};

/* Macro to define the usual plugin main function */
MAIN()

/* Add capabilities of this plugin to Procedural DataBase */
static void query() {
  static GimpParamDef args[] = {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE, "image", "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },

    { GIMP_PDB_INT32, "vtile", "Make tilable vertically" },
    { GIMP_PDB_INT32, "htile", "Make tilable horizontally" },
    { GIMP_PDB_INT32, "noborder", "Dont change border pixels" },
    { GIMP_PDB_INT32, "corpus", "Layer to use as corpus" },
    { GIMP_PDB_INT32, "inmask", "Layer to use as input mask, -1 for none" },
    { GIMP_PDB_INT32, "outmask", "Layer to use as output mask, -1 for none" },
    { GIMP_PDB_FLOAT, "random_weight", "Amount of randomness in output" },
    { GIMP_PDB_FLOAT, "map_weight", "Weight to give to map, is map is used" }
  };

  GimpParamDef *return_vals = NULL;
  gint nargs = sizeof(args)/sizeof(args[0]);
  gint nreturn_vals = 0;

  gimp_install_procedure("plug-in-resynthesizer",
                         "Make tiles,"
			 "apply themes to images, "
			 "remove unwanted features, "
			 "melt your CPU, etc",
			 "",
			 "Paul Francis Harrison",
			 "Paul Francis Harrison",
			 "2000",
			 "<Image>/Filters/Map/Resynthesize...",
			 "RGB*, GRAY*",
			 GIMP_PLUGIN,
			 nargs, nreturn_vals,
			 args, return_vals);
}

void init_gtk() {
  int argc = 1;
  char **argv = g_new(char*,2);
  argv[0] = "resynthesizer";
  argv[1] = 0;
  gtk_init(&argc, &argv);
  gtk_rc_parse(gimp_gtkrc());
}

static void run(gchar *name,
                gint nparams,
		GimpParam *param,
		gint *nreturn_vals,
		GimpParam **return_vals) {
  static GimpParam values[1];
  int i,j,k,l,x,y;
  Parameters parameters;
  GimpDrawable *drawable, *corpus_drawable, *map_in_drawable, *map_out_drawable;
  bool ok, with_map;
  int input_bytes, map_bytes, map_pos, random_bytes, random_pos, bytes,
      n_pixels, n_corpus_pixels, patch_radius;    
  Bitmap *data, *status, *corpus, *corpus_mask, *weights;
  GArray *neighbours;
  Heap *heap;
  int weight_sum_len;
  double *weight_sum;
  Node *tree;

  srand(time(0));

  /* Unless anything goes wrong, result is success */
  *nreturn_vals = 1;
  *return_vals = values;
  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = GIMP_PDB_SUCCESS;

  init_gtk();

  /* Get drawable */
  drawable = gimp_drawable_get(param[2].data.d_drawable);

  if (!gimp_drawable_is_rgb(drawable->id) &&
      !gimp_drawable_is_gray(drawable->id)) {
    gimp_drawable_detach(drawable);
    values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    return;
  }

  /* Deal with run mode (parameter zero) */
  ok = false;
  switch(param[0].data.d_int32) {
    case GIMP_RUN_INTERACTIVE :
      get_last_parameters(&parameters,drawable->id); 
      ok = get_parameters_by_asking(&parameters,drawable->id); 
      break;
    case GIMP_RUN_NONINTERACTIVE :
      ok = get_parameters_from_list(&parameters, nparams, param); 
      break;
    case GIMP_RUN_WITH_LAST_VALS :
      ok = get_last_parameters(&parameters,drawable->id); 
      break;
  }

  if (!ok) {
    gimp_drawable_detach(drawable);
    values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    return;
  }


  corpus_drawable = gimp_drawable_get(parameters.corpus_id);
  if (corpus_drawable->bpp != drawable->bpp) {
    gimp_message("The input texture and output image must have the same number of color channels.");
    gimp_drawable_detach(drawable);
    gimp_drawable_detach(corpus_drawable);
    values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    return;
  }
  
  with_map = (parameters.input_map_id != -1 && parameters.output_map_id != -1);
  map_in_drawable=0;
  map_out_drawable=0;

  if (with_map) {
    map_in_drawable = gimp_drawable_get(parameters.input_map_id);
    map_out_drawable = gimp_drawable_get(parameters.output_map_id);

    if (map_in_drawable->bpp != map_out_drawable->bpp) {
      gimp_message("The input and output maps must have the same number of color channels");
      ok = false;
    } else if (map_in_drawable->width != corpus_drawable->width || 
               map_in_drawable->height != corpus_drawable->height) {
      gimp_message("The input map should be the same size as the input texture image");
      ok = false;
    } else if (map_out_drawable->width != drawable->width || 
               map_out_drawable->height != drawable->height) {
      gimp_message("The output map should be the same size as the output image");
      ok = false;
    }

    if (!ok) {
      gimp_drawable_detach(drawable);
      gimp_drawable_detach(corpus_drawable);
      gimp_drawable_detach(map_in_drawable);
      gimp_drawable_detach(map_out_drawable);
      values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
      return;
    }
  }

  if (parameters.corpus_id == -1)
    parameters.corpus_id = drawable->id;

  /* ************************************************************************** */
  /* Data layout (actual image and mask may vary)

     Actual corpus  R
                    G
		    B
		    A
     Mask           R
                    G
		    B
		    A
     Random         R
  */

  input_bytes  = drawable->bpp;
  map_bytes   = (with_map ? map_in_drawable->bpp : 0);
  map_pos     = input_bytes;
  random_bytes = 1;
  random_pos   = map_pos + map_bytes;
  bytes        = random_pos + random_bytes;

  data = bitmap_new(sizeof(Pixelel));            
                        //Actual output (+ noise + mask)
  //int n_pixels;        
                        //Number of pixels to be generated in output
  status = bitmap_new(sizeof(Status));          
                        //Have done this output pixel yet?, pixel priority
  //Heap *heap;            
                        //Heap of pointers to status, for priority queueing
  
  corpus = bitmap_new(sizeof(Pixelel));          
                        //Input image (+ noise + mask)
  corpus_mask = bitmap_new(sizeof(Pixelel));     
                        //Which pixels to use from corpus
  //int n_corpus_pixels; 
                        //How many useable pixels in the corpus
  weights = bitmap_new(sizeof(int));         
                        //Weighting given to neighbours when a pixel is added 

  //Arbitrarily set this to three
  patch_radius = 3;
  
  neighbours = g_array_new(FALSE,FALSE,sizeof(Coordinates));

  /* Choose which neighbours to use */

  { 
    int radius;
    for(radius=1;radius<=patch_radius;radius++) {
      for(i=0;i<radius*2;i++) {
        Coordinates c;
        c.x = -radius+i;
        c.y = radius;
        g_array_append_val(neighbours,c);
        c.x = radius;
        c.y = radius-i;
        g_array_append_val(neighbours,c);
        c.x = radius-i;
        c.y = -radius;
        g_array_append_val(neighbours,c);
        c.x = -radius;
        c.y = -radius+i;
        g_array_append_val(neighbours,c);
      }
    }
  }

  /* Allocate memory for output and status tracking */

  {
    Bitmap *mask = bitmap_new(sizeof(Pixelel));

    fetch_image_and_mask(drawable, data,bytes, mask, 255);

    if (with_map) {
      bitmap_from(data, map_out_drawable, 0, 0, map_pos);
      gimp_drawable_detach(map_out_drawable);
    }

    bitmap_size(status,data->width,data->height,1);
    heap = heap_new(status->data,status->width*status->height);

    for(i=0;i<status->width*status->height;i++) {
      if (parameters.use_border) {
        ((Status*)status->data)[i].use  = true;
        ((Status*)status->data)[i].done = (((Pixelel*)mask->data)[i] != 255);
      } else {
	((Status*)status->data)[i].use  = (((Pixelel*)mask->data)[i] != 0);
        ((Status*)status->data)[i].done = (((Pixelel*)mask->data)[i] == 0);
      }
      ((Status*)status->data)[i].priority = 0;
    }

    bitmap_free(mask);
  }
    
  n_pixels = 0;
 
  {
    Pixelel *ptr = data->data;
    for(i=0;i<data->width*data->height;i++,ptr+=data->depth) {
      if (!((Status*)status->data)[i].done) {
        n_pixels++;
        for(j=0;j<input_bytes;j++)
          ptr[j] = 128;
      }
      ptr[random_pos] = rand();
    }
  }

  /* Fetch the corpus */
 
  fetch_image_and_mask(corpus_drawable, corpus,bytes, corpus_mask, 0);
      
  gimp_drawable_detach(corpus_drawable);

  if (with_map) {
    bitmap_from(corpus, map_in_drawable, 0, 0, map_pos);
    gimp_drawable_detach(map_in_drawable);
  }

  {
    Pixelel *ptr = ((Pixelel*)corpus->data) + random_pos;
    for(i=0;i<corpus->width*corpus->height;i++,ptr+=corpus->depth)
      *ptr = rand();
  }

  /* Work out which pixels from the corpus to use */

  {
    Pixelel *ptr;
    ptr = corpus_mask->data;
    for(i=0;i<corpus_mask->width*corpus_mask->height;i++,ptr++)
      *ptr = (*ptr ? 0 : 1);
  }
  
  for(y=patch_radius;y<corpus->height-patch_radius;y++)
    for(x=patch_radius;x<corpus->width-patch_radius;x++)
      if (bitmap_at(corpus_mask,Pixelel,x,y)[0])
        for(i=0;i<neighbours->len;i++) {
	  Coordinates *n = &g_array_index(neighbours,Coordinates,i);
          if (bitmap_at(corpus_mask,Pixelel,x+n->x,y+n->y)[0])
            bitmap_at(corpus_mask,Pixelel,x+n->x,y+n->y)[0]++;
	}

  n_corpus_pixels = 0;
  
  {
    Pixelel *ptr;
    ptr = corpus_mask->data;
    for(i=0;i<corpus_mask->width*corpus_mask->height;i++,ptr++)
      if (*ptr == neighbours->len+1) {
        *ptr = (*ptr == neighbours->len+1 ? 1 : 0);
        n_corpus_pixels++;
      } else
        *ptr = 0;
  }

  /* Sanity check */

  if (!n_corpus_pixels || !n_pixels) {
    if (!n_corpus_pixels)
      gimp_message("The input texture is too small.");
    else
      gimp_message("The output image is too small.");

    values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    return;
  }


  /* Examine neighbouring pixel interactions */

  bitmap_size(weights, corpus->width,corpus->height,neighbours->len);

  weight_sum_len = neighbours->len + (with_map?neighbours->len+1:0);
  weight_sum = g_new(double,weight_sum_len);
  for(i=0;i<weight_sum_len;i++)
    weight_sum[i] = 0.0;
    
  gimp_progress_init("Calculating weights");

  {
    double total_bits, total_entropy;
    Twine_list_item *sorted_list;

    guint mask;
    int start;

    Entropy_accumulator *accumulator = entropy_accumulator_new(input_bytes);

    entropy_accumulator_clear(accumulator);
    for(i=0;i<corpus->width*corpus->height;i++)
      entropy_accumulator_add(accumulator,(Pixelel*)corpus->data+i*corpus->depth);

    entropy_accumulator_calculate(accumulator, &total_bits, &total_entropy);
    
    sorted_list = g_new(Twine_list_item,n_corpus_pixels);
    
    for(i=0,y=0;y<corpus->height;y++)
      for(x=0;x<corpus->width;x++) 
        if (bitmap_at(corpus_mask,Pixelel,x,y)[0]) {
	  sorted_list[i].x = x;
	  sorted_list[i].y = y;
	  sorted_list[i].twine = entwine(bitmap_at(corpus,Pixelel,x,y), input_bytes);
	  i++;
        }

    qsort(sorted_list,n_corpus_pixels,sizeof(Twine_list_item),
	  twine_list_item_sorter);

    for(i=0;i<neighbours->len;i++) {
      /* Find best quantization level */

      int best_level = 0;
      double best_bits = 1e30;

      for(j=0;j<=(int)(sizeof(Pixelel)*8);j++) {
	double this_bits;

        mask = (guint)(-1);
	mask <<= j*input_bytes;
	
	entropy_accumulator_clear(accumulator);
        start = 0;
	this_bits = 0.0;
        for(k=0;;k++) { 
	  if (k == n_corpus_pixels ||
	      (sorted_list[start].twine&mask) != (sorted_list[k].twine&mask)) {
	    double bits, entropy;
            entropy_accumulator_calculate(accumulator, &bits, &entropy);
	    this_bits += bits;
            entropy_accumulator_clear(accumulator);
	    start = k;
	  }

	  if (k == n_corpus_pixels)
	    break;
	    
	  entropy_accumulator_add(accumulator,
	    bitmap_at(corpus,Pixelel,
	      sorted_list[k].x+g_array_index(neighbours,Coordinates,i).x,
	      sorted_list[k].y+g_array_index(neighbours,Coordinates,i).y));
	}

        //Kinda dodgy, its possible the length won't be a... whats the
	// technical term... one hump function
	if (this_bits > best_bits)
	  break;

	best_level = j;
	best_bits = this_bits;
      }
      
#ifdef DEBUG
      printf("Best level %d\n",best_level);
#endif

      /* Calculate weights */

      mask = (guint)(-1);
      mask <<= best_level*input_bytes;
	
      entropy_accumulator_clear(accumulator);
      start = 0;
      for(k=0;;k++) { 
        if (k == n_corpus_pixels ||
	    (sorted_list[start].twine&mask) != (sorted_list[k].twine&mask)) {
	  int weight;
	  double bits, entropy;
          entropy_accumulator_calculate(accumulator, &bits, &entropy);

	  weight = (int)( (total_entropy - entropy) * 65536.0 );
	  if (weight < 0)
	    weight = 0;
	  
	  weight_sum[i] += weight * (k-start);

          for(l=start;l<k;l++)
	    bitmap_at(weights,int,sorted_list[l].x,sorted_list[l].y)[i] = 
	      weight;
	  
	  entropy_accumulator_clear(accumulator);
	  start = k;
	}
	
	if (k == n_corpus_pixels)
	  break;
	    
	entropy_accumulator_add(accumulator,
	  bitmap_at(corpus,Pixelel,
	    sorted_list[k].x+g_array_index(neighbours,Coordinates,i).x,
	    sorted_list[k].y+g_array_index(neighbours,Coordinates,i).y));
      }
      
      gimp_progress_update((double)(i+1) / weight_sum_len);
    }


    /* Do map weights if necessary */

    if (with_map) {
      for(i=0,y=0;y<corpus->height;y++)
        for(x=0;x<corpus->width;x++) 
	  if (bitmap_at(corpus_mask,Pixelel,x,y)[0]) {
	    sorted_list[i].x = x;
	    sorted_list[i].y = y;
	    sorted_list[i].twine = entwine(bitmap_at(corpus,Pixelel,x,y)+map_pos, map_bytes);
	    i++;
          }

      qsort(sorted_list,n_corpus_pixels,sizeof(Twine_list_item),
  	    twine_list_item_sorter);

      for(i=0;i<=neighbours->len;i++) {
        double best_bits, best_sum, this_bits, this_sum;
        int offset_x, offset_y;
	if (i < neighbours->len) {
	  offset_x = g_array_index(neighbours,Coordinates,i).x;
	  offset_y = g_array_index(neighbours,Coordinates,i).y;
	} else
	  offset_x = offset_y = 0;

        best_bits = 1e30;
	best_sum = 0.0;
	for(j=0;j<=(int)(sizeof(Pixelel)*8);j++) {
          mask = (guint)(-1);
	  mask <<= j*map_bytes;
	
	  entropy_accumulator_clear(accumulator);
          start = 0;
	  this_bits = 0.0;
	  this_sum = 0.0;
          for(k=0;;k++) { 
	    if (k == n_corpus_pixels ||
	        (sorted_list[start].twine&mask) != (sorted_list[k].twine&mask)) {
	      int weight;
	      double bits, entropy;
              entropy_accumulator_calculate(accumulator, &bits, &entropy);
	      this_bits += bits;

	      weight = (int)( (total_entropy - entropy) * 65536.0 );
	      if (weight < 0)
	        weight = 0;
	  
	      this_sum += weight * (k-start);
              entropy_accumulator_clear(accumulator);
	      start = k;
	    }

	    if (k == n_corpus_pixels)
	      break;
	    
	    entropy_accumulator_add(accumulator,
	      bitmap_at(corpus,Pixelel,
	        sorted_list[k].x+offset_x,
	        sorted_list[k].y+offset_y));
	  }

	  if (this_bits > best_bits)
	    break;

	  best_bits = this_bits;
	  best_sum = this_sum;
	}

	weight_sum[i+neighbours->len] = best_sum;
      
        gimp_progress_update((double)(i+1+neighbours->len) / weight_sum_len);
      }
    }

    g_free(sorted_list);
    entropy_accumulator_free(accumulator);
  }


  /* Normalize weights, formulate distance measure */

  set_up_tree_globals();

  {
    double sum, map_sum;
    Offset offset;

    for(y=0;y<corpus->height;y++)
      for(x=0;x<corpus->width;x++) 
        if (bitmap_at(corpus_mask,Pixelel,x,y)[0]) {
          int *weight = bitmap_at(weights,int,x,y);

	  int sum = 0;
	  for(i=0;i<neighbours->len;i++) {
	    sum += weight[i];
	  }

	  if (sum == 0)
	    for(i=0;i<neighbours->len;i++)
	      weight[i] = 65536 / neighbours->len;
	  else
	    for(i=0;i<neighbours->len;i++)
	      weight[i] = (int)(weight[i] * 65536.0 / sum);
        }

    for(i=0;i<weight_sum_len;i++) {
      weight_sum[i] = pow(2.0,weight_sum[i]
        /n_corpus_pixels
	/65536.0);
    }
    
    sum = 0.0;
    for(i=0;i<neighbours->len;i++)
      sum += weight_sum[i];
    
    map_sum = 0.0;
    if (with_map)
      for(i=0;i<=neighbours->len;i++)
        map_sum += weight_sum[i+neighbours->len];

#ifdef DEBUG
    for(y=-patch_radius;y<=patch_radius;y++) {
      for(x=-patch_radius;x<=patch_radius;x++) {
        int v = 0;
	for(i=0;i<neighbours->len;i++)
	  if (g_array_index(neighbours,Coordinates,i).x == x && 
	      g_array_index(neighbours,Coordinates,i).y == y)
            v = int(weight_sum[i]*65536.0 / sum);
	printf("%6i",v);
      }
      printf("\n");
    }

    if (with_map) {
      printf("\n");
      for(y=-patch_radius;y<=patch_radius;y++) {
        for(x=-patch_radius;x<=patch_radius;x++) {
          int v = 0;
	  for(i=0;i<neighbours->len;i++)
	    if (g_array_index(neighbours,Coordinates,i).x == x && 
	        g_array_index(neighbours,Coordinates,i).y == y)
              v = int(weight_sum[i+neighbours->len]*65536.0 / map_sum);
	  printf("%6i",v);
        }
        printf("\n");
      }
    }
#endif

    for(i=0;i<neighbours->len;i++) {
      offset.weight = (int)(weight_sum[i]*65536.0 / sum);
      offset.x  = -g_array_index(neighbours,Coordinates,i).x;
      offset.y  = -g_array_index(neighbours,Coordinates,i).y;

      for(j=0;j<input_bytes;j++) {
        offset.el = j;
        offset.offset = j
	  - g_array_index(neighbours,Coordinates,i).x * corpus->depth 
	  - g_array_index(neighbours,Coordinates,i).y * corpus->depth * corpus->width;
	g_array_append_val(Offsets,offset);
      }
    }

    if (with_map) {
      double mult = pow(2.0,16.0*parameters.map_weight) / map_sum;
      for(i=0;i<=neighbours->len;i++) {
        offset.weight = (int)(weight_sum[i+neighbours->len]*mult);

	if (i < neighbours->len) {
          offset.x  = -g_array_index(neighbours,Coordinates,i).x;
          offset.y  = -g_array_index(neighbours,Coordinates,i).y;
	} else
	  offset.x = offset.y = 0;
	
        for(j=0;j<map_bytes;j++) {
          offset.el = map_pos+j;
          offset.offset = map_pos+j
	    + offset.x * corpus->depth 
	    + offset.y * corpus->depth * corpus->width;
	  g_array_append_val(Offsets,offset);
        }
      }
    }

    offset.x  = 0;
    offset.y  = 0;
    offset.el = random_pos;
    offset.offset = random_pos;
    offset.weight = (int)(pow(2,16.0*parameters.random_weight));
    g_array_append_val(Offsets,offset);

    for(y=0;y<data->height;y++)
      for(x=0;x<data->width;x++)
        if (!bitmap_at(status,Status,x,y)[0].done)
	  for(i=0;i<neighbours->len;i++) {
	    int xx = x - g_array_index(neighbours,Coordinates,i).x;
	    int yy = y - g_array_index(neighbours,Coordinates,i).y;

	    bool use;
	    if (!wrap_or_clip(&parameters,data,&xx,&yy))
	      use = !parameters.use_border; 
	    else
	      use = bitmap_at(status,Status,xx,yy)[0].done;
	    
	    if (use)
	      bitmap_at(status,Status,x,y)[0].priority += (int)(weight_sum[i]*65536.0/sum); 
	  }
  }

  g_free(weight_sum);


  /* Sort offsets from largest to smallest weight */

  qsort(& (g_array_index(Offsets,Offset,0)), Offsets->len, sizeof(Offset),
        offset_sorter);

  /* Build kd-tree */
  
  {
    Pixelel **items = g_new(Pixelel*,n_corpus_pixels);

    for(i=0,y=0;y<corpus->height;y++)
      for(x=0;x<corpus->width;x++)
        if (bitmap_at(corpus_mask,Pixelel,x,y)[0])
          items[i++] = bitmap_at(corpus,Pixelel,x,y);

    tree = build_tree(items,n_corpus_pixels);  

    g_free(items);
  }

  /* Grow image */

  {
    Status *next;
    Progress_monitor *monitor;
    int n_pixels_done = 0;

    heap_become_heapish(heap);
  
    gimp_progress_init("Resynthesizing");

    monitor = progress_monitor_new(data, input_bytes < 3);
    progress_monitor_update(monitor);
    while(gtk_events_pending())
      gtk_main_iteration();
  
    while(!progress_monitor_aborted(monitor) && 
          (next = heap_pop_top(heap))) {
      Pixelel *result;
      int *weight;
      int pos = next-((Status*)status->data);
      int x = pos % status->width;
      int y = pos / status->width;

      g_array_set_size(Patterns,0);
      for(i=0;i<Offsets->len;i++) {
        Offset *o = &g_array_index(Offsets,Offset,i);
        int yy  = y + o->y;
        int xx  = x + o->x;
        int el = o->el;
      
        if (wrap_or_clip(&parameters,data,&xx,&yy) &&
	    bitmap_at(status,Status,xx,yy)[0].use &&
            (el >= input_bytes || bitmap_at(status,Status,xx,yy)[0].done)) {
          Pattern pattern;
          pattern.offset_no = i;
          pattern.value = bitmap_at(data,Pixelel,xx,yy)[el];
	  g_array_append_val(Patterns,pattern);
        }
      }
      
      bitmap_at(status,Status,x,y)[0].done = true;

      result = query_tree(tree);
      for(i=0;i<input_bytes;i++)
        bitmap_at(data,Pixelel,x,y)[i] = result[i];

      weight = (result - ((Pixelel*)corpus->data))/corpus->depth*weights->depth + weights->data;

      for(i=0;i<neighbours->len;i++) {
        int xx = x + g_array_index(neighbours,Coordinates,i).x;
        int yy = y + g_array_index(neighbours,Coordinates,i).y;

        if (wrap_or_clip(&parameters,data,&xx,&yy) && 
	    bitmap_at(status,Status,xx,yy)[0].use &&
	    !bitmap_at(status,Status,xx,yy)[0].done) {
	  bitmap_at(status,Status,xx,yy)[0].priority += weight[i];
	  heap_promote(heap,bitmap_at(status,Status,xx,yy));
        }
      }

      n_pixels_done++;
      if (!(n_pixels_done&255)) {
        gimp_progress_update((double)(n_pixels_done) / n_pixels);
        progress_monitor_update(monitor);
      }

      if (!(n_pixels_done&15)) 
        while(gtk_events_pending())
          gtk_main_iteration();
    }

    ok = !progress_monitor_aborted(monitor);
    progress_monitor_free(monitor);
  }

  node_free(tree);

  clean_up_tree_globals();

  /* Write result back to the GIMP, clean up */

  if (ok) {
    /* Write result to region */
    bitmap_to(data,drawable, 0,0, 0);

    /* Voodoo to update actual image */
    gimp_drawable_flush(drawable);
    gimp_drawable_merge_shadow(drawable->id,TRUE);
    gimp_drawable_update(drawable->id,0,0,data->width,data->height);

    /* Store parameters deep in the bowels of the GIMP */
    if (parameters.corpus_id == drawable->id)
      parameters.corpus_id = -1;
    gimp_set_data("plug_in_resynthesizer", &parameters, sizeof(parameters));
  }

  bitmap_free(data);
  bitmap_free(status);
  bitmap_free(corpus);
  bitmap_free(corpus_mask);
  bitmap_free(weights);

  heap_free(heap);

  g_array_free(neighbours,TRUE);
  
  gimp_drawable_detach(drawable);
  
  gimp_displays_flush();
}

